@@ -41,6 +41,7 @@
 namespace io = process::io;
 
 using process::Clock;
+using process::Future;
 using process::MAX_REAP_INTERVAL;
 using process::Subprocess;
 using process::subprocess;
@@ -77,13 +78,10 @@ void run_subprocess(const lambda::function<Try<Subprocess>()>& createSubprocess)
 // a file descriptor for a file, rather than a socket).
 TEST_F(SubprocessTest, PipeOutputToFileDescriptor)
 {
-  Try<string> testdir = os::mkdtemp();
-  ASSERT_SOME(testdir);
-
   // Create temporary files to pipe `stdin` to, and open it. We will pipe
   // output into this file.
   const string outfile_name = "out.txt";
-  const string outfile = path::join(testdir.get(), outfile_name);
+  const string outfile = path::join(sandbox.get(), outfile_name);
   ASSERT_SOME(os::touch(outfile));
 
   Try<int_fd> outfile_fd = os::open(outfile, O_RDWR);
@@ -92,7 +90,7 @@ TEST_F(SubprocessTest, PipeOutputToFileDescriptor)
   // Create temporary files to pipe `stderr` to, and open it. We will pipe
   // error into this file.
   const string errorfile_name = "error.txt";
-  const string errorfile = path::join(testdir.get(), errorfile_name);
+  const string errorfile = path::join(sandbox.get(), errorfile_name);
   ASSERT_SOME(os::touch(errorfile));
 
   Try<int_fd> errorfile_fd = os::open(errorfile, O_RDWR);
@@ -142,15 +140,12 @@ TEST_F(SubprocessTest, PipeOutputToFileDescriptor)
 
 TEST_F(SubprocessTest, PipeOutputToPath)
 {
-  Try<string> testdir = os::mkdtemp();
-  ASSERT_SOME(testdir);
-
   // Name the files to pipe output and error to.
   const string outfile_name = "out.txt";
-  const string outfile = path::join(testdir.get(), outfile_name);
+  const string outfile = path::join(sandbox.get(), outfile_name);
 
   const string errorfile_name = "error.txt";
-  const string errorfile = path::join(testdir.get(), errorfile_name);
+  const string errorfile = path::join(sandbox.get(), errorfile_name);
 
   // Pipe simple string to output file.
   run_subprocess(
@@ -189,12 +184,9 @@ TEST_F(SubprocessTest, PipeOutputToPath)
 
 TEST_F(SubprocessTest, EnvironmentEcho)
 {
-  Try<string> testdir = os::mkdtemp();
-  ASSERT_SOME(testdir);
-
   // Name the file to pipe output to.
   const string outfile_name = "out.txt";
-  const string outfile = path::join(testdir.get(), outfile_name);
+  const string outfile = path::join(sandbox.get(), outfile_name);
 
   // Pipe simple string to output file.
   run_subprocess(
@@ -1057,5 +1049,81 @@ TEST_F(SubprocessTest, EnvironmentOverride)
 }
 
 
+#ifdef __linux__
+// This test verifies:
+//   1. The subprocess will have the stdio file descriptors.
+//   2. The whitelisted file descriptors will be successfully
+//      inherited by the subprocess.
+//   3. The non-whitelisted file descriptors will be not be
+//      inherited by the subprocess.
+TEST_F(SubprocessTest, WhiteListFds)
+{
+  Try<int_fd> fd1 = os::open(
+      path::join(os::getcwd(), id::UUID::random().toString()),
+      O_CREAT | O_EXCL | O_RDONLY | O_CLOEXEC);
+
+  Try<int_fd> fd2 = os::open(
+      path::join(os::getcwd(), id::UUID::random().toString()),
+      O_CREAT | O_EXCL | O_RDONLY);
+
+  ASSERT_SOME(fd1);
+  ASSERT_SOME(fd2);
+
+  Try<Subprocess> s = subprocess(
+      "ls /dev/fd",
+      Subprocess::FD(STDIN_FILENO),
+      Subprocess::PIPE(),
+      Subprocess::FD(STDERR_FILENO),
+      None(),
+      None(),
+      {},
+      {},
+      {fd1.get()});
+
+  ASSERT_SOME(s);
+  ASSERT_SOME(s->out());
+
+  Future<string> output = io::read(s->out().get());
+  AWAIT_READY(output);
+
+  hashset<int_fd> fds;
+
+  vector<string> tokens = strings::tokenize(output.get(), "\n");
+  foreach (const string& fdString, tokens) {
+    Try<int_fd> fd = numify<int_fd>(fdString);
+    ASSERT_SOME(fd);
+
+    fds.insert(fd.get());
+  }
+
+  // The subprocess should always have the stdio file descriptors.
+  EXPECT_TRUE(fds.contains(STDIN_FILENO));
+  EXPECT_TRUE(fds.contains(STDOUT_FILENO));
+  EXPECT_TRUE(fds.contains(STDERR_FILENO));
+
+  // `fd1` should be inherited by the subprocess since it is whitelisted even
+  // it has `O_CLOEXEC` set initially.
+  EXPECT_TRUE(fds.contains(fd1.get()));
+
+  // `fd2` should not be inherited by the subprocess since it is not whitelisted
+  // even it has no `O_CLOEXEC` set initially.
+  EXPECT_FALSE(fds.contains(fd2.get()));
+
+  ASSERT_SOME(os::close(fd1.get()));
+  ASSERT_SOME(os::close(fd2.get()));
+
+  // Advance time until the internal reaper reaps the subprocess.
+  Clock::pause();
+  while (s->status().isPending()) {
+    Clock::advance(MAX_REAP_INTERVAL());
+    Clock::settle();
+  }
+  Clock::resume();
+
+  AWAIT_EXPECT_WEXITSTATUS_EQ(0, s->status());
+}
+#endif // __linux__
+
+
 // TODO(joerg84): Consider adding tests for setsid, working_directory,
 // and supervisor childHooks.
