@@ -620,6 +620,8 @@ TEST_F(FsTest, Close)
 
 
 #if defined(__linux__) || defined(__APPLE__)
+// NOTE: This test is otherwise disabled since it uses `os::setxattr`
+// and `os::getxattr` which are not available elsewhere.
 TEST_F(FsTest, Xattr)
 {
   const string file = path::join(os::getcwd(), id::UUID::random().toString());
@@ -655,6 +657,7 @@ TEST_F(FsTest, Xattr)
 }
 #endif // __linux__ || __APPLE__
 
+
 #ifdef __WINDOWS__
 // Check if the overlapped field is set properly on Windows.
 TEST_F(FsTest, Overlapped)
@@ -800,5 +803,40 @@ TEST_F(FsTest, ReadWriteAsyncLargeBuffer)
   EXPECT_SOME(os::close(pipes.get()[0]));
   EXPECT_SOME(os::close(pipes.get()[1]));
 }
+#endif // __WINDOWS__
 
+
+#ifndef __WINDOWS__
+TEST_F(FsTest, Used)
+{
+  Try<Bytes> used = fs::used(".");
+  ASSERT_SOME(used);
+
+  Try<Bytes> size = fs::size(".");
+  ASSERT_SOME(size);
+
+  // We unfortunately can't easily verify the used value since
+  // the disk usage can change at any point.
+
+  EXPECT_GT(used.get(), 0u);
+  EXPECT_LT(used.get(), size.get());
+}
+
+
+// This test verifies that the file descriptors returned by `os::lsof()`
+// are all open file descriptors and contains stdin, stdout and stderr.
+TEST_F(FsTest, Lsof)
+{
+  Try<std::vector<int_fd>> fds = os::lsof();
+  ASSERT_SOME(fds);
+
+  // Verify each `fd` is an open file descriptor.
+  foreach (int_fd fd, fds.get()) {
+    EXPECT_NE(-1, ::fcntl(fd, F_GETFD));
+  }
+
+  EXPECT_NE(std::find(fds->begin(), fds->end(), 0), fds->end());
+  EXPECT_NE(std::find(fds->begin(), fds->end(), 1), fds->end());
+  EXPECT_NE(std::find(fds->begin(), fds->end(), 2), fds->end());
+}
 #endif // __WINDOWS__
