@@ -46,6 +46,10 @@
 #include <stout/try.hpp>
 #include <stout/uuid.hpp>
 
+#if defined(__WINDOWS__)
+#include <stout/windows.hpp>
+#endif
+
 #include <stout/os/environment.hpp>
 #include <stout/os/int_fd.hpp>
 #include <stout/os/kill.hpp>
@@ -60,12 +64,16 @@
 #include <stout/os/sysctl.hpp>
 #endif
 
+#if defined(__WINDOWS__)
+#include <stout/os/shell.hpp>
+#endif
+
 #include <stout/tests/utils.hpp>
 
-#ifndef __WINDOWS__
+#if !defined(__WINDOWS__)
 using os::Exec;
-#endif // __WINDOWS__
 using os::Fork;
+#endif // __WINDOWS__
 using os::Process;
 using os::ProcessTree;
 
@@ -76,33 +84,80 @@ using std::set;
 using std::string;
 using std::vector;
 
+#if defined(__WINDOWS__)
+// NOTE: These are used to partially implement tests which otherwise
+// relied on `os::Exec` and `os::Fork`.
+using ::internal::windows::ProcessData;
+
+ProcessData windows_fork()
+{
+  // This will unfortunately generate some output.
+  Try<ProcessData> process_data = ::internal::windows::create_process(
+      "", {"ping.exe", "127.0.0.1", "-n", "10"}, None());
+  CHECK_SOME(process_data);
+  return process_data.get();
+}
+
+void windows_kill(const ProcessData& p)
+{
+  CHECK_EQ(TRUE, ::TerminateProcess(p.process_handle.get_handle(), 0));
+  CHECK_EQ(
+      WAIT_OBJECT_0, ::WaitForSingleObject(p.process_handle.get_handle(), 500));
+}
+#endif
 
 class OsTest : public TemporaryDirectoryTest {};
 
 
-#ifndef __WINDOWS__
 TEST_F(OsTest, Environment)
 {
   // Make sure the environment has some entries with '=' in the value.
   os::setenv("SOME_SPECIAL_FLAG", "--flag=foobar");
 
-  char** environ = os::raw::environment();
+#ifndef __WINDOWS__
+  char** raw_environ = os::raw::environment();
+#else
+  std::vector<string> raw_environ;
+  const std::unique_ptr<wchar_t[], decltype(&::FreeEnvironmentStringsW)> env(
+      ::GetEnvironmentStringsW(), &::FreeEnvironmentStringsW);
+
+  for (size_t i = 0; env[i] != L'\0' && env[i + 1] != L'\0';
+       /* incremented below */) {
+    std::wstring entry(&env[i]);
+
+    // Increment past the current environment string and null terminator.
+    i = i + entry.size() + 1;
+
+    size_t position = entry.find_first_of(L'=');
+    if (position == std::string::npos) {
+      continue; // Skip malformed environment entries.
+    }
+
+    raw_environ.push_back(stringify(entry.substr(0)));
+  }
+#endif // __WINDOWS__
 
   hashmap<string, string> environment = os::environment();
 
-  for (size_t index = 0; environ[index] != nullptr; index++) {
-    string entry(environ[index]);
-    size_t position = entry.find_first_of('=');
+#ifndef __WINDOWS__
+  for (size_t index = 0; raw_environ[index] != nullptr; index++) {
+#else
+  for (size_t index = 0; index < raw_environ.size(); index++) {
+#endif // __WINDOWS__
+    string entry(raw_environ[index]);
+    // In practice, sometimes the key starts with `=` even though it
+    // is not supposed to. If we don't skip it, we get an empty key.
+    size_t position = entry.find_first_of('=', 1);
     if (position == string::npos) {
       continue; // Skip malformed environment entries.
     }
+
     const string key = entry.substr(0, position);
     const string value = entry.substr(position + 1);
     EXPECT_TRUE(environment.contains(key));
     EXPECT_EQ(value, environment[key]);
   }
 }
-#endif // __WINDOWS__
 
 
 TEST_F(OsTest, TrivialEnvironment)
@@ -379,8 +434,6 @@ TEST_F(OsTest, Sysctl)
 #endif // __APPLE__ || __FreeBSD__
 
 
-// TODO(hausdorff): Enable when we implement `Fork` and `Exec`. See MESOS-3638.
-#ifndef __WINDOWS__
 TEST_F(OsTest, Children)
 {
   Try<set<pid_t>> children = os::children(getpid());
@@ -388,6 +441,7 @@ TEST_F(OsTest, Children)
   ASSERT_SOME(children);
   EXPECT_TRUE(children->empty());
 
+#ifndef __WINDOWS__
   Try<ProcessTree> tree =
     Fork(None(),                   // Child.
          Fork(Exec(SLEEP_COMMAND(10))),   // Grandchild.
@@ -398,19 +452,33 @@ TEST_F(OsTest, Children)
 
   pid_t child = tree->process.pid;
   pid_t grandchild = tree->children.front().process.pid;
+#else
+  // NOTE: On Windows, while debugging, children will include, e.g.,
+  // `conhost.exe`, so we have to count those too.
+  const size_t existing_children = children.get().size();
+  ProcessData process_data = windows_fork();
+  pid_t child = process_data.pid;
+#endif // __WINDOWS__
 
   // Ensure the non-recursive children does not include the
   // grandchild.
   children = os::children(getpid(), false);
 
   ASSERT_SOME(children);
+
+#ifndef __WINDOWS__
   EXPECT_EQ(1u, children->size());
+#else
+  EXPECT_EQ(existing_children + 1u, children->size());
+#endif // __WINDOWS__
+
   EXPECT_EQ(1u, children->count(child));
 
   children = os::children(getpid());
 
   ASSERT_SOME(children);
 
+#ifndef __WINDOWS__
   // Depending on whether or not the shell has fork/exec'ed in each
   // above 'Exec', we could have 2 or 4 children. That is, some shells
   // might simply for exec the command above (i.e., 'sleep 10') while
@@ -428,19 +496,31 @@ TEST_F(OsTest, Children)
 
   // We have to reap the child for running the tests in repetition.
   ASSERT_EQ(child, waitpid(child, nullptr, 0));
+#else
+  // Cleanup by killing the descendant processes.
+  windows_kill(process_data);
+
+  children = os::children(getpid(), false);
+  ASSERT_SOME(children);
+  EXPECT_EQ(existing_children, children->size());
+#endif
 }
 
-
+#ifndef __WINDOWS__
 void dosetsid()
 {
   if (::setsid() == -1) {
     ABORT(string("Failed to setsid: ") + os::strerror(errno));
   }
 }
+#endif // __WINDOWS__
 
 
 TEST_F(OsTest, Killtree)
 {
+  // NOTE: This test is written differently for Windows since there is
+  // no implementation of `fork` and `exec` on Windows.
+#ifndef __WINDOWS__
   Try<ProcessTree> tree =
     Fork(&dosetsid,                        // Child.
          Fork(None(),                      // Grandchild.
@@ -559,9 +639,37 @@ TEST_F(OsTest, Killtree)
 
   // We have to reap the child for running the tests in repetition.
   ASSERT_EQ(child, waitpid(child, nullptr, 0));
+#else
+  Try<set<pid_t>> children = os::children(getpid());
+  ASSERT_SOME(children);
+
+  const size_t existing_children = children->size();
+  ProcessData process_data = windows_fork();
+  pid_t child = process_data.pid;
+
+  Try<std::wstring> name = os::name_job(child);
+  ASSERT_SOME(name);
+  Try<SharedHandle> handle = os::create_job(name.get());
+  ASSERT_SOME(handle);
+  ASSERT_SOME(os::assign_job(handle.get(), child));
+
+  children = os::children(getpid());
+  ASSERT_SOME(children);
+  ASSERT_EQ(existing_children + 1u, children->size());
+
+  Try<list<ProcessTree>> trees = os::killtree(child, SIGKILL, true, true);
+  ASSERT_SOME(trees);
+
+  children = os::children(getpid());
+  ASSERT_SOME(children);
+  ASSERT_EQ(existing_children, children->size());
+#endif // __WINDOWS__
 }
 
 
+#ifndef __WINDOWS__
+// NOTE: This test is disabled for Windows since there is
+// no implementation of `fork` and `exec` on Windows.
 TEST_F(OsTest, KilltreeNoRoot)
 {
   Try<ProcessTree> tree =
@@ -677,6 +785,7 @@ TEST_F(OsTest, KilltreeNoRoot)
   EXPECT_NONE(os::process(grandchild));
   EXPECT_NONE(os::process(greatGrandchild));
 }
+#endif // __WINDOWS__
 
 
 TEST_F(OsTest, ProcessExists)
@@ -684,6 +793,10 @@ TEST_F(OsTest, ProcessExists)
   // Check we exist.
   EXPECT_TRUE(os::exists(::getpid()));
 
+  // NOTE: Some of this test is disabled for Windows since there is no
+  // implementation of `fork` and `exec` on Windows.
+#if !defined(__WINDOWS__)
+
   // In a FreeBSD jail, pid 1 may not exist.
 #if !defined(__FreeBSD__)
   // Check init/launchd/systemd exists.
@@ -735,9 +848,26 @@ TEST_F(OsTest, ProcessExists)
   EXPECT_WTERMSIG_EQ(SIGKILL, status);
 
   EXPECT_FALSE(os::exists(pid));
+#else
+  pid_t child;
+  // Scope the `ProcessData` so the handles are destructed (and thus
+  // closed) before the last test.
+  {
+    ProcessData process_data = windows_fork();
+    child = process_data.pid;
+    EXPECT_TRUE(os::exists(child));
+    windows_kill(process_data);
+  }
+
+  os::sleep(Milliseconds(500));
+  EXPECT_FALSE(os::exists(child));
+#endif // __WINDOWS__
 }
 
 
+#ifndef __WINDOWS__
+// NOTE: Enable this test when there is an implementation of
+// `os::getuid` and `os::chown` for Windows.
 TEST_F(OsTest, User)
 {
   Try<string> user_ = os::shell("id -un");
@@ -801,8 +931,13 @@ TEST_F(OsTest, User)
   EXPECT_SOME(os::setgroups(gids.get(), uid.get()));
   EXPECT_SOME(os::setuid(uid.get()));
 }
+#endif // __WINDOWS__
 
 
+#ifndef __WINDOWS__
+// NOTE: Enable this test if there are ever implementations of
+// `os::getuid`, `os::getgid`, `os::chmod` and `os::chown` for
+// Windows.
 TEST_F(OsTest, SYMLINK_Chown)
 {
   Result<uid_t> uid = os::getuid();
@@ -871,8 +1006,12 @@ TEST_F(OsTest, SYMLINK_Chown)
       os::stat::uid("chown/one/file",
                     FollowSymlink::DO_NOT_FOLLOW_SYMLINK));
 }
+#endif // __WINDOWS__
 
 
+#ifndef __WINDOWS__
+// NOTE: Enable this test when there is an implementation of
+// `os::getuid`, `os::getgid`, `os::chmod` and `os::chown` for Windows.
 TEST_F(OsTest, ChownNoAccess)
 {
   Result<uid_t> uid = os::getuid();
@@ -924,11 +1063,11 @@ TEST_F(OsTest, TrivialUser)
 }
 
 
+#ifndef __WINDOWS__
 // Test setting/resetting/appending to LD_LIBRARY_PATH environment
 // variable (DYLD_LIBRARY_PATH on OS X).
 //
 // NOTE: This will never be enabled on Windows as there is no equivalent.
-#ifndef __WINDOWS__
 TEST_F(OsTest, Libraries)
 {
   const string path1 = "/tmp/path1";
@@ -1040,8 +1179,8 @@ TEST_F(OsTest, Shell)
 }
 
 
-// NOTE: Disabled on Windows because `mknod` does not exist.
 #ifndef __WINDOWS__
+// NOTE: Disabled on Windows because `mknod` does not exist.
 TEST_F(OsTest, Mknod)
 {
 #ifdef __FreeBSD__
