@@ -38,6 +38,7 @@
 #include <stout/try.hpp>
 #include <stout/unreachable.hpp>
 
+#include "common/authorization.hpp"
 #include "common/http.hpp"
 #include "common/parse.hpp"
 #include "common/protobuf_utils.hpp"
@@ -412,8 +413,9 @@ public:
         case authorization::STOP_MAINTENANCE:
         case authorization::UPDATE_MAINTENANCE_SCHEDULE:
         case authorization::MODIFY_RESOURCE_PROVIDER_CONFIG:
-        case authorization::PRUNE_IMAGES:
+        case authorization::MARK_RESOURCE_PROVIDER_GONE:
         case authorization::VIEW_RESOURCE_PROVIDER:
+        case authorization::PRUNE_IMAGES:
           aclObject.set_type(ACL::Entity::ANY);
 
           break;
@@ -429,6 +431,7 @@ public:
         case authorization::DESTROY_BLOCK_DISK:
         case authorization::CREATE_MOUNT_DISK:
         case authorization::DESTROY_MOUNT_DISK:
+        case authorization::DESTROY_RAW_DISK:
           return Error("Authorization for action " + stringify(action_) +
                        " requires a specialized approver object.");
         case authorization::UNKNOWN:
@@ -605,7 +608,8 @@ public:
         case authorization::CREATE_BLOCK_DISK:
         case authorization::DESTROY_BLOCK_DISK:
         case authorization::CREATE_MOUNT_DISK:
-        case authorization::DESTROY_MOUNT_DISK: {
+        case authorization::DESTROY_MOUNT_DISK:
+        case authorization::DESTROY_RAW_DISK: {
           entityObject.set_type(ACL::Entity::SOME);
           if (object->resource) {
             if (object->resource->reservations_size() > 0) {
@@ -732,6 +736,7 @@ public:
         case authorization::WAIT_NESTED_CONTAINER:
         case authorization::WAIT_STANDALONE_CONTAINER:
         case authorization::MODIFY_RESOURCE_PROVIDER_CONFIG:
+        case authorization::MARK_RESOURCE_PROVIDER_GONE:
         case authorization::VIEW_RESOURCE_PROVIDER:
         case authorization::UNKNOWN:
           UNREACHABLE();
@@ -942,6 +947,11 @@ public:
           createHierarchicalRoleACLs(acls.destroy_mount_disks());
         break;
       }
+      case authorization::DESTROY_RAW_DISK: {
+        hierarchicalRoleACLs =
+          createHierarchicalRoleACLs(acls.destroy_raw_disks());
+        break;
+      }
       case authorization::ACCESS_MESOS_LOG:
       case authorization::ACCESS_SANDBOX:
       case authorization::ATTACH_CONTAINER_INPUT:
@@ -977,6 +987,7 @@ public:
       case authorization::WAIT_NESTED_CONTAINER:
       case authorization::WAIT_STANDALONE_CONTAINER:
       case authorization::MODIFY_RESOURCE_PROVIDER_CONFIG:
+      case authorization::MARK_RESOURCE_PROVIDER_GONE:
       case authorization::VIEW_RESOURCE_PROVIDER:
         UNREACHABLE();
     }
@@ -1163,7 +1174,8 @@ public:
       case authorization::CREATE_BLOCK_DISK:
       case authorization::DESTROY_BLOCK_DISK:
       case authorization::CREATE_MOUNT_DISK:
-      case authorization::DESTROY_MOUNT_DISK: {
+      case authorization::DESTROY_MOUNT_DISK:
+      case authorization::DESTROY_RAW_DISK: {
         return getHierarchicalRoleApprover(subject, action);
       }
       case authorization::ACCESS_MESOS_LOG:
@@ -1198,6 +1210,7 @@ public:
       case authorization::WAIT_NESTED_CONTAINER:
       case authorization::WAIT_STANDALONE_CONTAINER:
       case authorization::MODIFY_RESOURCE_PROVIDER_CONFIG:
+      case authorization::MARK_RESOURCE_PROVIDER_GONE:
       case authorization::VIEW_RESOURCE_PROVIDER:
       case authorization::UNKNOWN: {
         Result<vector<GenericACL>> genericACLs =
@@ -1554,11 +1567,12 @@ private:
         }
 
         return acls_;
-      case authorization::PRUNE_IMAGES:
-        foreach (const ACL::PruneImages& acl, acls.prune_images()) {
+      case authorization::MARK_RESOURCE_PROVIDER_GONE:
+        foreach (const ACL::MarkResourceProvidersGone& acl,
+                 acls.mark_resource_providers_gone()) {
           GenericACL acl_;
           acl_.subjects = acl.principals();
-          acl_.objects = acl.images();
+          acl_.objects = acl.resource_providers();
 
           acls_.push_back(acl_);
         }
@@ -1575,6 +1589,16 @@ private:
           acls_.push_back(acl_);
         }
 
+        return acls_;
+      case authorization::PRUNE_IMAGES:
+        foreach (const ACL::PruneImages& acl, acls.prune_images()) {
+          GenericACL acl_;
+          acl_.subjects = acl.principals();
+          acl_.objects = acl.images();
+
+          acls_.push_back(acl_);
+        }
+
         return acls_;
       case authorization::REGISTER_FRAMEWORK:
       case authorization::CREATE_VOLUME:
@@ -1590,6 +1614,7 @@ private:
       case authorization::DESTROY_BLOCK_DISK:
       case authorization::CREATE_MOUNT_DISK:
       case authorization::DESTROY_MOUNT_DISK:
+      case authorization::DESTROY_RAW_DISK:
         return Error("Extracting ACLs for " + stringify(action) + " requires "
                      "a specialized function");
       case authorization::UNKNOWN:
@@ -1662,7 +1687,7 @@ Option<Error> LocalAuthorizer::validate(const ACLs& acls)
   foreach (const ACL::GetEndpoint& acl, acls.get_endpoints()) {
     if (acl.paths().type() == ACL::Entity::SOME) {
       foreach (const string& path, acl.paths().values()) {
-        if (!AUTHORIZABLE_ENDPOINTS.contains(path)) {
+        if (!authorization::AUTHORIZABLE_ENDPOINTS.contains(path)) {
           return Error("Path: '" + path + "' is not an authorizable path");
         }
       }
@@ -1751,6 +1776,21 @@ Option<Error> LocalAuthorizer::validate(const ACLs& acls)
     }
   }
 
+  foreach (const ACL::MarkResourceProvidersGone& acl,
+           acls.mark_resource_providers_gone()) {
+    if (acl.resource_providers().type() == ACL::Entity::SOME) {
+      return Error(
+          "ACL.MarkResourceProvidersGone type must be either NONE or ANY");
+    }
+  }
+
+  foreach (const ACL::ViewResourceProvider& acl,
+           acls.view_resource_providers()) {
+    if (acl.resource_providers().type() == ACL::Entity::SOME) {
+      return Error("ACL.ViewResourceProvider type must be either NONE or ANY");
+    }
+  }
+
   foreach (const ACL::ModifyResourceProviderConfig& acl,
            acls.modify_resource_provider_configs()) {
     if (acl.resource_providers().type() == ACL::Entity::SOME) {
@@ -1765,13 +1805,6 @@ Option<Error> LocalAuthorizer::validate(const ACLs& acls)
     }
   }
 
-  foreach (const ACL::ViewResourceProvider& acl,
-           acls.view_resource_providers()) {
-    if (acl.resource_providers().type() == ACL::Entity::SOME) {
-      return Error("ACL.ViewResourceProvider type must be either NONE or ANY");
-    }
-  }
-
   // TODO(alexr): Consider validating not only protobuf, but also the original
   // JSON in order to spot misspelled names. A misspelled action may affect
   // authorization result and hence lead to a security issue (e.g. when there
