@@ -32,6 +32,7 @@
 #include <mesos/quota/quota.hpp>
 
 #include <process/authenticator.hpp>
+#include <process/clock.hpp>
 #include <process/collect.hpp>
 #include <process/dispatch.hpp>
 #include <process/future.hpp>
@@ -47,6 +48,7 @@
 
 #include <stout/os/permissions.hpp>
 
+#include "common/authorization.hpp"
 #include "common/http.hpp"
 
 #include "messages/messages.hpp"
@@ -94,16 +96,6 @@ ostream& operator<<(ostream& stream, ContentType contentType)
 
 namespace internal {
 
-// Set of endpoint whose access is protected with the authorization
-// action `GET_ENDPOINTS_WITH_PATH`.
-hashset<string> AUTHORIZABLE_ENDPOINTS{
-    "/containers",
-    "/files/debug",
-    "/logging/toggle",
-    "/metrics/snapshot",
-    "/monitor/statistics"};
-
-
 string serialize(
     ContentType contentType,
     const google::protobuf::Message& message)
@@ -583,6 +575,39 @@ static void json(JSON::ObjectWriter* writer, const ContainerStatus& status)
 }
 
 
+static void json(
+    JSON::ObjectWriter* writer,
+    const DomainInfo::FaultDomain::RegionInfo& regionInfo)
+{
+  writer->field("name", regionInfo.name());
+}
+
+
+static void json(
+    JSON::ObjectWriter* writer,
+    const DomainInfo::FaultDomain::ZoneInfo& zoneInfo)
+{
+  writer->field("name", zoneInfo.name());
+}
+
+
+static void json(
+    JSON::ObjectWriter* writer,
+    const DomainInfo::FaultDomain& faultDomain)
+{
+    writer->field("region", faultDomain.region());
+    writer->field("zone", faultDomain.zone());
+}
+
+
+void json(JSON::ObjectWriter* writer, const DomainInfo& domainInfo)
+{
+  if (domainInfo.has_fault_domain()) {
+    writer->field("fault_domain", domainInfo.fault_domain());
+  }
+}
+
+
 void json(JSON::ObjectWriter* writer, const ExecutorInfo& executorInfo)
 {
   writer->field("executor_id", executorInfo.executor_id().value());
@@ -610,6 +635,14 @@ void json(JSON::ObjectWriter* writer, const ExecutorInfo& executorInfo)
 }
 
 
+void json(
+    JSON::StringWriter* writer,
+    const FrameworkInfo::Capability& capability)
+{
+  writer->set(FrameworkInfo::Capability::Type_Name(capability.type()));
+}
+
+
 void json(JSON::ArrayWriter* writer, const Labels& labels)
 {
   foreach (const Label& label, labels.labels()) {
@@ -618,6 +651,27 @@ void json(JSON::ArrayWriter* writer, const Labels& labels)
 }
 
 
+void json(JSON::ObjectWriter* writer, const MasterInfo& info)
+{
+  writer->field("id", info.id());
+  writer->field("pid", info.pid());
+  writer->field("port", info.port());
+  writer->field("hostname", info.hostname());
+
+  if (info.has_domain()) {
+    writer->field("domain", info.domain());
+  }
+}
+
+
+void json(
+    JSON::StringWriter* writer,
+    const MasterInfo::Capability& capability)
+{
+  writer->set(MasterInfo::Capability::Type_Name(capability.type()));
+}
+
+
 static void json(JSON::ObjectWriter* writer, const NetworkInfo& info)
 {
   if (info.groups().size() > 0) {
@@ -651,6 +705,16 @@ static void json(JSON::ObjectWriter* writer, const NetworkInfo& info)
 }
 
 
+void json(JSON::ObjectWriter* writer, const Offer& offer)
+{
+  writer->field("id", offer.id().value());
+  writer->field("framework_id", offer.framework_id().value());
+  writer->field("allocation_info", JSON::Protobuf(offer.allocation_info()));
+  writer->field("slave_id", offer.slave_id().value());
+  writer->field("resources", offer.resources());
+}
+
+
 template <typename ResourceIterable>
 void json(
     JSON::ObjectWriter* writer,
@@ -702,6 +766,27 @@ void json(
 }
 
 
+void json(JSON::ObjectWriter* writer, const SlaveInfo& slaveInfo)
+{
+  writer->field("id", slaveInfo.id().value());
+  writer->field("hostname", slaveInfo.hostname());
+  writer->field("port", slaveInfo.port());
+  writer->field("attributes", Attributes(slaveInfo.attributes()));
+
+  if (slaveInfo.has_domain()) {
+    writer->field("domain", slaveInfo.domain());
+  }
+}
+
+
+void json(
+    JSON::StringWriter* writer,
+    const SlaveInfo::Capability& capability)
+{
+  writer->set(SlaveInfo::Capability::Type_Name(capability.type()));
+}
+
+
 void json(JSON::ObjectWriter* writer, const Task& task)
 {
   writer->field("id", task.task_id().value());
@@ -733,6 +818,10 @@ void json(JSON::ObjectWriter* writer, const Task& task)
   if (task.has_container()) {
     writer->field("container", JSON::Protobuf(task.container()));
   }
+
+  if (task.has_health_check()) {
+    writer->field("health_check", JSON::Protobuf(task.health_check()));
+  }
 }
 
 
@@ -755,39 +844,6 @@ void json(JSON::ObjectWriter* writer, const TaskStatus& status)
 }
 
 
-static void json(
-    JSON::ObjectWriter* writer,
-    const DomainInfo::FaultDomain::RegionInfo& regionInfo)
-{
-  writer->field("name", regionInfo.name());
-}
-
-
-static void json(
-    JSON::ObjectWriter* writer,
-    const DomainInfo::FaultDomain::ZoneInfo& zoneInfo)
-{
-  writer->field("name", zoneInfo.name());
-}
-
-
-static void json(
-    JSON::ObjectWriter* writer,
-    const DomainInfo::FaultDomain& faultDomain)
-{
-    writer->field("region", faultDomain.region());
-    writer->field("zone", faultDomain.zone());
-}
-
-
-void json(JSON::ObjectWriter* writer, const DomainInfo& domainInfo)
-{
-  if (domainInfo.has_fault_domain()) {
-    writer->field("fault_domain", domainInfo.fault_domain());
-  }
-}
-
-
 static void json(JSON::NumberWriter* writer, const Value::Scalar& scalar)
 {
   writer->set(scalar.value());
@@ -811,74 +867,6 @@ static void json(JSON::StringWriter* writer, const Value::Text& text)
   writer->set(text.value());
 }
 
-namespace authorization {
-
-const Option<authorization::Subject> createSubject(
-    const Option<Principal>& principal)
-{
-  if (principal.isSome()) {
-    authorization::Subject subject;
-
-    if (principal->value.isSome()) {
-      subject.set_value(principal->value.get());
-    }
-
-    foreachpair (const string& key, const string& value, principal->claims) {
-      Label* claim = subject.mutable_claims()->mutable_labels()->Add();
-      claim->set_key(key);
-      claim->set_value(value);
-    }
-
-    return subject;
-  }
-
-  return None();
-}
-
-} // namespace authorization {
-
-const AuthorizationCallbacks createAuthorizationCallbacks(
-    Authorizer* authorizer)
-{
-  typedef lambda::function<process::Future<bool>(
-      const process::http::Request& httpRequest,
-      const Option<Principal>& principal)> Callback;
-
-  AuthorizationCallbacks callbacks;
-
-  Callback getEndpoint = [authorizer](
-      const process::http::Request& httpRequest,
-      const Option<Principal>& principal) -> process::Future<bool> {
-        const string path = httpRequest.url.path;
-
-        if (!internal::AUTHORIZABLE_ENDPOINTS.contains(path)) {
-          return Failure(
-              "Endpoint '" + path + "' is not an authorizable endpoint.");
-        }
-
-        authorization::Request authRequest;
-        authRequest.set_action(mesos::authorization::GET_ENDPOINT_WITH_PATH);
-
-        Option<authorization::Subject> subject =
-          authorization::createSubject(principal);
-        if (subject.isSome()) {
-          authRequest.mutable_subject()->CopyFrom(subject.get());
-        }
-
-        authRequest.mutable_object()->set_value(path);
-
-        LOG(INFO) << "Authorizing principal '"
-                  << (principal.isSome() ? stringify(principal.get()) : "ANY")
-                  << "' to GET the endpoint '" << path << "'";
-
-        return authorizer->authorized(authRequest);
-      };
-
-  callbacks.insert(std::make_pair("/logging/toggle", getEndpoint));
-  callbacks.insert(std::make_pair("/metrics/snapshot", getEndpoint));
-
-  return callbacks;
-}
 
 Future<Owned<ObjectApprovers>> ObjectApprovers::create(
     const Option<Authorizer*>& authorizer,
@@ -938,7 +926,7 @@ process::Future<bool> authorizeEndpoint(
     return Failure("Unexpected request method '" + method + "'");
   }
 
-  if (!internal::AUTHORIZABLE_ENDPOINTS.contains(endpoint)) {
+  if (!authorization::AUTHORIZABLE_ENDPOINTS.contains(endpoint)) {
     return Failure(
         "Endpoint '" + endpoint + "' is not an authorizable endpoint.");
   }
@@ -1126,4 +1114,18 @@ void logRequest(const process::http::Request& request)
                 : "");
 }
 
+
+void logResponse(
+    const process::http::Request& request,
+    const process::http::Response& response)
+{
+  LOG(INFO) << "HTTP " << request.method << " for " << request.url
+            << (request.client.isSome()
+                ? " from " + stringify(request.client.get())
+                : "")
+            << ": '" << response.status << "'"
+            << " after " << (process::Clock::now() - request.received).ms()
+            << Milliseconds::units();
+}
+
 }  // namespace mesos {
