@@ -14,6 +14,8 @@
 // See the License for the specific language governing permissions and
 // limitations under the License.
 
+#include "common/protobuf_utils.hpp"
+
 #ifdef __WINDOWS__
 #include <stout/internal/windows/grp.hpp>
 #include <stout/internal/windows/pwd.hpp>
@@ -47,7 +49,6 @@
 #endif // __WINDOWS__
 
 #include "common/http.hpp"
-#include "common/protobuf_utils.hpp"
 #include "common/resources_utils.hpp"
 
 #include "master/master.hpp"
@@ -66,6 +67,7 @@ using google::protobuf::RepeatedPtrField;
 using mesos::authorization::VIEW_ROLE;
 
 using mesos::slave::ContainerLimitation;
+using mesos::slave::ContainerMountInfo;
 using mesos::slave::ContainerState;
 
 using process::Owned;
@@ -75,6 +77,62 @@ namespace mesos {
 namespace internal {
 namespace protobuf {
 
+// If `descriptor` is not a descriptor of a protobuf union,
+// this constructor will abort the process.
+UnionValidator::UnionValidator(const google::protobuf::Descriptor* descriptor)
+{
+  const auto* typeFieldDescriptor = descriptor->FindFieldByName("type");
+  CHECK_NOTNULL(typeFieldDescriptor);
+
+  typeDescriptor_ = typeFieldDescriptor->enum_type();
+  CHECK_NOTNULL(typeDescriptor_);
+
+  const auto* unknownTypeValueDescriptor =
+    typeDescriptor_->FindValueByNumber(0);
+  if (unknownTypeValueDescriptor != nullptr) {
+    CHECK_EQ(unknownTypeValueDescriptor->name(), "UNKNOWN");
+  }
+
+  for (int index = 0; index < typeDescriptor_->value_count(); index++) {
+    const auto* typeValueDescriptor = typeDescriptor_->value(index);
+    if (typeValueDescriptor->number() == 0) {
+      // We are skipping the "UNKNOWN" value of the enum.
+      continue;
+    }
+
+    const auto* fieldDescriptor =
+      descriptor->FindFieldByName(strings::lower(typeValueDescriptor->name()));
+
+    CHECK_NOTNULL(fieldDescriptor);
+    unionFieldDescriptors_.emplace_back(
+        typeValueDescriptor->number(), fieldDescriptor);
+  }
+}
+
+
+Option<Error> UnionValidator::validate(
+  const int messageTypeNumber, const google::protobuf::Message& message) const
+{
+  const auto* reflection = message.GetReflection();
+  for (const auto& item : unionFieldDescriptors_) {
+    const auto typeNumber = item.first;
+    const auto* fieldDescriptor = item.second;
+    if (
+        messageTypeNumber != typeNumber &&
+        reflection->HasField(message, fieldDescriptor)) {
+      const auto* descr = typeDescriptor_->FindValueByNumber(messageTypeNumber);
+      return Error(
+          "Protobuf union `" + message.GetDescriptor()->full_name() +
+          "` with `Type == " +
+          (descr == nullptr ? string("<UNKNOWN>") : descr->name()) +
+          "` should not have the field `" +
+          fieldDescriptor->name() + "` set.");
+    }
+  }
+  return None();
+}
+
+
 bool frameworkHasCapability(
     const FrameworkInfo& framework,
     FrameworkInfo::Capability::Type capability)
@@ -92,14 +150,26 @@ bool frameworkHasCapability(
 
 bool isTerminalState(const TaskState& state)
 {
-  return (state == TASK_FINISHED ||
-          state == TASK_FAILED ||
-          state == TASK_KILLED ||
-          state == TASK_LOST ||
-          state == TASK_ERROR ||
-          state == TASK_DROPPED ||
-          state == TASK_GONE ||
-          state == TASK_GONE_BY_OPERATOR);
+  switch (state) {
+    case TASK_FINISHED:
+    case TASK_FAILED:
+    case TASK_KILLED:
+    case TASK_LOST:
+    case TASK_ERROR:
+    case TASK_DROPPED:
+    case TASK_GONE:
+    case TASK_GONE_BY_OPERATOR:
+      return true;
+    case TASK_KILLING:
+    case TASK_STAGING:
+    case TASK_STARTING:
+    case TASK_RUNNING:
+    case TASK_UNREACHABLE:
+    case TASK_UNKNOWN:
+      return false;
+  }
+
+  UNREACHABLE();
 }
 
 
@@ -341,6 +411,10 @@ Task createTask(
     t.mutable_container()->CopyFrom(task.container());
   }
 
+  if (task.has_health_check()) {
+    t.mutable_health_check()->CopyFrom(task.health_check());
+  }
+
   // Copy `user` if set.
   if (task.has_command() && task.command().has_user()) {
     t.set_user(task.command().user());
@@ -403,17 +477,17 @@ Option<ContainerStatus> getTaskContainerStatus(const Task& task)
 bool isTerminalState(const OperationState& state)
 {
   switch (state) {
-    case OPERATION_FINISHED:
-    case OPERATION_FAILED:
-    case OPERATION_ERROR:
     case OPERATION_DROPPED:
-      return true;
-    case OPERATION_UNSUPPORTED:
-    case OPERATION_PENDING:
-    case OPERATION_UNREACHABLE:
+    case OPERATION_ERROR:
+    case OPERATION_FAILED:
+    case OPERATION_FINISHED:
     case OPERATION_GONE_BY_OPERATOR:
+      return true;
+    case OPERATION_PENDING:
     case OPERATION_RECOVERING:
     case OPERATION_UNKNOWN:
+    case OPERATION_UNREACHABLE:
+    case OPERATION_UNSUPPORTED:
       return false;
   }
 
@@ -426,7 +500,9 @@ OperationStatus createOperationStatus(
     const Option<OperationID>& operationId,
     const Option<string>& message,
     const Option<Resources>& convertedResources,
-    const Option<id::UUID>& uuid)
+    const Option<id::UUID>& uuid,
+    const Option<SlaveID>& slaveId,
+    const Option<ResourceProviderID>& resourceProviderId)
 {
   OperationStatus status;
   status.set_state(state);
@@ -447,6 +523,14 @@ OperationStatus createOperationStatus(
     status.mutable_uuid()->CopyFrom(protobuf::createUUID(uuid.get()));
   }
 
+  if (slaveId.isSome()) {
+    status.mutable_slave_id()->CopyFrom(slaveId.get());
+  }
+
+  if (resourceProviderId.isSome()) {
+    status.mutable_resource_provider_id()->CopyFrom(resourceProviderId.get());
+  }
+
   return status;
 }
 
@@ -1049,7 +1133,8 @@ bool operator==(const Capabilities& left, const Capabilities& right)
          left.hierarchicalRole == right.hierarchicalRole &&
          left.reservationRefinement == right.reservationRefinement &&
          left.resourceProvider == right.resourceProvider &&
-         left.resizeVolume == right.resizeVolume;
+         left.resizeVolume == right.resizeVolume &&
+         left.agentOperationFeedback == right.agentOperationFeedback;
 }
 
 
@@ -1105,6 +1190,57 @@ ContainerState createContainerState(
   return state;
 }
 
+
+ContainerMountInfo createContainerMount(
+    const string& source,
+    const string& target,
+    unsigned long flags)
+{
+  ContainerMountInfo mnt;
+
+  mnt.set_source(source);
+  mnt.set_target(target);
+  mnt.set_flags(flags);
+
+  return mnt;
+}
+
+
+ContainerMountInfo createContainerMount(
+    const string& source,
+    const string& target,
+    const string& type,
+    unsigned long flags)
+{
+  ContainerMountInfo mnt;
+
+  mnt.set_source(source);
+  mnt.set_target(target);
+  mnt.set_type(type);
+  mnt.set_flags(flags);
+
+  return mnt;
+}
+
+
+ContainerMountInfo createContainerMount(
+    const string& source,
+    const string& target,
+    const string& type,
+    const string& options,
+    unsigned long flags)
+{
+  ContainerMountInfo mnt;
+
+  mnt.set_source(source);
+  mnt.set_target(target);
+  mnt.set_type(type);
+  mnt.set_options(options);
+  mnt.set_flags(flags);
+
+  return mnt;
+}
+
 } // namespace slave {
 
 namespace maintenance {
