@@ -80,9 +80,13 @@ bool operator==(const CommandInfo& left, const CommandInfo& right)
 
 bool operator==(const CommandInfo::URI& left, const CommandInfo::URI& right)
 {
+  // NOTE: We purposefully do not compare the value of the `cache` field
+  // because a URI downloaded from source or from the fetcher cache should
+  // be considered identical.
   return left.value() == right.value() &&
     left.executable() == right.executable() &&
-    left.extract() == right.extract();
+    left.extract() == right.extract() &&
+    left.output_file() == right.output_file();
 }
 
 
@@ -488,6 +492,23 @@ bool operator==(const OperationStatus& left, const OperationStatus& right)
     return false;
   }
 
+  if (left.has_slave_id() != right.has_slave_id()) {
+    return false;
+  }
+
+  if (left.has_slave_id() && left.slave_id() != right.slave_id()) {
+    return false;
+  }
+
+  if (left.has_resource_provider_id() != right.has_resource_provider_id()) {
+    return false;
+  }
+
+  if (left.has_resource_provider_id() &&
+      left.resource_provider_id() != right.resource_provider_id()) {
+    return false;
+  }
+
   return true;
 }
 
@@ -747,6 +768,35 @@ ostream& operator<<(ostream& stream, const OperationState& state)
 }
 
 
+ostream& operator<<(ostream& stream, const Operation& operation)
+{
+  stream << operation.uuid() << " (";
+
+  stream << operation.info().type();
+
+  if (operation.has_framework_id()) {
+    stream << " for framework " << operation.framework_id();
+  }
+
+  if (operation.info().has_id()) {
+    stream << ", ID: " << operation.info().id();
+  }
+
+  const OperationStatus& latestStatus(operation.latest_status());
+
+  if (latestStatus.has_resource_provider_id()) {
+    stream << ", affecting resource provider "
+           << latestStatus.resource_provider_id();
+  }
+
+  stream << ", latest state: " << latestStatus.state();
+
+  stream << ")";
+
+  return stream;
+}
+
+
 ostream& operator<<(ostream& stream, const RateLimits& limits)
 {
   return stream << limits.DebugString();
