@@ -15,6 +15,7 @@
 // limitations under the License.
 
 #include "common/values.hpp"
+#include "common/validation.hpp"
 
 #include <algorithm>
 #include <cmath>
@@ -98,12 +99,30 @@ bool operator==(const Value::Scalar& left, const Value::Scalar& right)
 }
 
 
+bool operator<(const Value::Scalar& left, const Value::Scalar& right)
+{
+  return convertToFixed(left.value()) < convertToFixed(right.value());
+}
+
+
 bool operator<=(const Value::Scalar& left, const Value::Scalar& right)
 {
   return convertToFixed(left.value()) <= convertToFixed(right.value());
 }
 
 
+bool operator>(const Value::Scalar& left, const Value::Scalar& right)
+{
+  return convertToFixed(left.value()) > convertToFixed(right.value());
+}
+
+
+bool operator>=(const Value::Scalar& left, const Value::Scalar& right)
+{
+  return convertToFixed(left.value()) >= convertToFixed(right.value());
+}
+
+
 Value::Scalar operator+(const Value::Scalar& left, const Value::Scalar& right)
 {
   Value::Scalar result = left;
@@ -707,6 +726,12 @@ Try<Value> parse(const string& text)
     } else if (index == string::npos) {
       Try<double> value_ = numify<double>(temp);
       if (!value_.isError()) {
+        Option<Error> error =
+          common::validation::validateInputScalarValue(*value_);
+        if (error.isSome()) {
+          return Error("Invalid scalar value '" + temp + "':" + error->message);
+        }
+
         // This is a scalar.
         Value::Scalar* scalar = value.mutable_scalar();
         value.set_type(Value::SCALAR);
