@@ -82,6 +82,10 @@
 #include "internal/devolve.hpp"
 #include "internal/evolve.hpp"
 
+#ifdef ENABLE_LAUNCHER_SEALING
+#include "linux/memfd.hpp"
+#endif // ENABLE_LAUNCHER_SEALING
+
 #include "logging/logging.hpp"
 
 #include "messages/messages.hpp"
@@ -133,7 +137,9 @@ public:
       const Option<Environment>& _taskEnvironment,
       const Option<CapabilityInfo>& _effectiveCapabilities,
       const Option<CapabilityInfo>& _boundingCapabilities,
+      const Option<string>& _ttySlavePath,
       const Option<ContainerLaunchInfo>& _taskLaunchInfo,
+      const Option<vector<gid_t>> _taskSupplementaryGroups,
       const FrameworkID& _frameworkId,
       const ExecutorID& _executorId,
       const Duration& _shutdownGracePeriod)
@@ -158,7 +164,9 @@ public:
       taskEnvironment(_taskEnvironment),
       effectiveCapabilities(_effectiveCapabilities),
       boundingCapabilities(_boundingCapabilities),
+      ttySlavePath(_ttySlavePath),
       taskLaunchInfo(_taskLaunchInfo),
+      taskSupplementaryGroups(_taskSupplementaryGroups),
       frameworkId(_frameworkId),
       executorId(_executorId),
       lastTaskStatus(None()) {}
@@ -256,6 +264,10 @@ public:
         break;
       }
 
+      case Event::HEARTBEAT: {
+        break;
+      }
+
       case Event::UNKNOWN: {
         LOG(WARNING) << "Received an UNKNOWN event and ignored";
         break;
@@ -411,7 +423,9 @@ protected:
       const Option<string>& workingDirectory,
       const Option<CapabilityInfo>& effectiveCapabilities,
       const Option<CapabilityInfo>& boundingCapabilities,
-      const Option<ContainerLaunchInfo>& taskLaunchInfo)
+      const Option<string>& ttySlavePath,
+      const Option<ContainerLaunchInfo>& taskLaunchInfo,
+      const Option<vector<gid_t>>& taskSupplementaryGroups)
   {
     // Prepare the flags to pass to the launch process.
     slave::MesosContainerizerLaunch::Flags launchFlags;
@@ -487,16 +501,28 @@ protected:
           taskLaunchInfo->clone_namespaces());
     }
 
+    if (taskSupplementaryGroups.isSome()) {
+      foreach (gid_t gid, taskSupplementaryGroups.get()) {
+        launchInfo.add_supplementary_groups(gid);
+      }
+    }
+
     launchFlags.launch_info = JSON::protobuf(launchInfo);
 
-    // TODO(tillt): Consider using a flag allowing / disallowing the
-    // log output of possibly sensitive data. See MESOS-7292.
-    string commandString = strings::format(
-        "%s %s <POSSIBLY-SENSITIVE-DATA>",
-        path::join(launcherDir, MESOS_CONTAINERIZER),
-        MesosContainerizerLaunch::NAME).get();
+    // Determine the mesos containerizer binary depends on whether we
+    // need to clone and seal it on linux.
+    string initPath = path::join(launcherDir, MESOS_CONTAINERIZER);
+#ifdef ENABLE_LAUNCHER_SEALING
+    // Clone the launcher binary in memory for security concerns.
+    Try<int_fd> memFd = memfd::cloneSealedFile(initPath);
+    if (memFd.isError()) {
+      ABORT(
+          "Failed to clone a sealed file '" + initPath + "' in memory: " +
+          memFd.error());
+    }
 
-    LOG(INFO) << "Running '" << commandString << "'";
+    initPath = "/proc/self/fd/" + stringify(memFd.get());
+#endif // ENABLE_LAUNCHER_SEALING
 
     // Fork the child using launcher.
     vector<string> argv(2);
@@ -513,8 +539,13 @@ protected:
         [](pid_t pid) { return os::set_job_kill_on_close_limit(pid); }));
 #endif // __WINDOWS__
 
+    vector<process::Subprocess::ChildHook> childHooks;
+    if (ttySlavePath.isNone()) {
+      childHooks.emplace_back(Subprocess::ChildHook::SETSID());
+    }
+
     Try<Subprocess> s = subprocess(
-        path::join(launcherDir, MESOS_CONTAINERIZER),
+        initPath,
         argv,
         Subprocess::FD(STDIN_FILENO),
         Subprocess::FD(STDOUT_FILENO),
@@ -523,10 +554,10 @@ protected:
         None(),
         None(),
         parentHooks,
-        {Subprocess::ChildHook::SETSID()});
+        childHooks);
 
     if (s.isError()) {
-      ABORT("Failed to launch '" + commandString + "': " + s.error());
+      ABORT("Failed to launch task subprocess: " + s.error());
     }
 
     return s->pid();
@@ -688,7 +719,9 @@ protected:
         workingDirectory,
         effectiveCapabilities,
         boundingCapabilities,
-        taskLaunchInfo);
+        ttySlavePath,
+        taskLaunchInfo,
+        taskSupplementaryGroups);
 
     LOG(INFO) << "Forked command at " << pid.get();
 
@@ -1218,7 +1251,9 @@ private:
   Option<Environment> taskEnvironment;
   Option<CapabilityInfo> effectiveCapabilities;
   Option<CapabilityInfo> boundingCapabilities;
+  Option<string> ttySlavePath;
   Option<ContainerLaunchInfo> taskLaunchInfo;
+  Option<vector<gid_t>> taskSupplementaryGroups;
   const FrameworkID frameworkId;
   const ExecutorID executorId;
   Owned<MesosBase> mesos;
@@ -1282,6 +1317,14 @@ public:
         "task_launch_info",
         "The launch info to run the task.");
 
+    add(&Flags::tty_slave_path,
+        "tty_slave_path",
+        "A path to the slave end of the attached TTY if there is one.");
+
+    add(&Flags::task_supplementary_groups,
+        "task_supplementary_groups",
+        "Comma-separated list of supplementary groups to run the task with.");
+
     add(&Flags::launcher_dir,
         "launcher_dir",
         "Directory path of Mesos binaries.",
@@ -1299,7 +1342,9 @@ public:
   Option<Environment> task_environment;
   Option<mesos::CapabilityInfo> effective_capabilities;
   Option<mesos::CapabilityInfo> bounding_capabilities;
+  Option<string> tty_slave_path;
   Option<JSON::Object> task_launch_info;
+  Option<vector<gid_t>> task_supplementary_groups;
   string launcher_dir;
 };
 
@@ -1390,7 +1435,9 @@ int main(int argc, char** argv)
           flags.task_environment,
           flags.effective_capabilities,
           flags.bounding_capabilities,
+          flags.tty_slave_path,
           task_launch_info,
+          flags.task_supplementary_groups,
           frameworkId,
           executorId,
           shutdownGracePeriod));
