@@ -76,6 +76,195 @@ class OfferFilter;
 class InverseOfferFilter;
 
 
+struct Framework
+{
+  Framework(
+      const FrameworkInfo& frameworkInfo,
+      const std::set<std::string>& suppressedRoles,
+      bool active,
+      bool publishPerFrameworkMetrics);
+
+  std::set<std::string> roles;
+
+  std::set<std::string> suppressedRoles;
+
+  protobuf::framework::Capabilities capabilities;
+
+  // Active offer and inverse offer filters for the framework.
+  // Offer filters are tied to the role the filtered resources
+  // were allocated to.
+  hashmap<std::string, hashmap<SlaveID, hashset<OfferFilter*>>> offerFilters;
+  hashmap<SlaveID, hashset<InverseOfferFilter*>> inverseOfferFilters;
+
+  bool active;
+
+  bool publishPerFrameworkMetrics;
+
+  process::Owned<FrameworkMetrics> metrics;
+
+  // TODO(bbannier): Consider documenting examples on how to use this setting.
+  hashmap<std::string, std::vector<ResourceQuantities>> minAllocatableResources;
+};
+
+
+class Slave
+{
+public:
+  Slave(
+      const SlaveInfo& _info,
+      const protobuf::slave::Capabilities& _capabilities,
+      bool _activated,
+      const Resources& _total,
+      const Resources& _allocated)
+    : info(_info),
+      capabilities(_capabilities),
+      activated(_activated),
+      total(_total),
+      allocated(_allocated),
+      shared(_total.shared()),
+      hasGpu_(_total.gpus().getOrElse(0) > 0)
+  {
+    updateAvailable();
+  }
+
+  const Resources& getTotal() const { return total; }
+
+  const Resources& getAllocated() const { return allocated; }
+
+  const Resources& getAvailable() const { return available; }
+
+  bool hasGpu() const { return hasGpu_; }
+
+  void updateTotal(const Resources& newTotal) {
+    total = newTotal;
+    shared = total.shared();
+    hasGpu_ = total.gpus().getOrElse(0) > 0;
+
+    updateAvailable();
+  }
+
+  void allocate(const Resources& toAllocate)
+  {
+    allocated += toAllocate;
+
+    updateAvailable();
+  }
+
+  void unallocate(const Resources& toUnallocate)
+  {
+    allocated -= toUnallocate;
+
+    updateAvailable();
+  }
+
+  // The `SlaveInfo` that was passed to the allocator when the slave was added
+  // or updated. Currently only two fields are used: `hostname` for host
+  // whitelisting and in log messages, and `domain` for region-aware
+  // scheduling.
+  SlaveInfo info;
+
+  protobuf::slave::Capabilities capabilities;
+
+  bool activated; // Whether to offer resources.
+
+  // Represents a scheduled unavailability due to maintenance for a specific
+  // slave, and the responses from frameworks as to whether they will be able
+  // to gracefully handle this unavailability.
+  //
+  // NOTE: We currently implement maintenance in the allocator to be able to
+  // leverage state and features such as the FrameworkSorter and OfferFilter.
+  struct Maintenance
+  {
+    Maintenance(const Unavailability& _unavailability)
+      : unavailability(_unavailability) {}
+
+    // The start time and optional duration of the event.
+    Unavailability unavailability;
+
+    // A mapping of frameworks to the inverse offer status associated with
+    // this unavailability.
+    //
+    // NOTE: We currently lose this information during a master fail over
+    // since it is not persisted or replicated. This is ok as the new master's
+    // allocator will send out new inverse offers and re-collect the
+    // information. This is similar to all the outstanding offers from an old
+    // master being invalidated, and new offers being sent out.
+    hashmap<FrameworkID, mesos::allocator::InverseOfferStatus> statuses;
+
+    // Represents the "unit of accounting" for maintenance. When a
+    // `FrameworkID` is present in the hashset it means an inverse offer has
+    // been sent out. When it is not present it means no offer is currently
+    // outstanding.
+    hashset<FrameworkID> offersOutstanding;
+  };
+
+  // When the `maintenance` is set the slave is scheduled to be unavailable at
+  // a given point in time, for an optional duration. This information is used
+  // to send out `InverseOffers`.
+  Option<Maintenance> maintenance;
+
+private:
+  void updateAvailable() {
+    // In order to subtract from the total,
+    // we strip the allocation information.
+    Resources allocated_ = allocated;
+    allocated_.unallocate();
+
+    // Calling `nonShared()` currently copies the underlying resources
+    // and is therefore rather expensive. We avoid it in the common
+    // case that there are no shared resources.
+    //
+    // TODO(mzhu): Ideally there would be a single logical path here.
+    // One solution is to have `Resources` be copy-on-write such that
+    // `nonShared()` performs no copying and instead points to a
+    // subset of the original `Resource` objects.
+    if (shared.empty()) {
+      available = total - allocated_;
+    } else {
+      // Since shared resources are offerable even when they are in use, we
+      // always include them as part of available resources.
+      available = (total.nonShared() - allocated_.nonShared()) + shared;
+    }
+  }
+
+  // Total amount of regular *and* oversubscribed resources.
+  Resources total;
+
+  // Regular *and* oversubscribed resources that are allocated.
+  //
+  // NOTE: We maintain multiple copies of each shared resource allocated
+  // to a slave, where the number of copies represents the number of times
+  // this shared resource has been allocated to (and has not been recovered
+  // from) a specific framework.
+  //
+  // NOTE: We keep track of the slave's allocated resources despite
+  // having that information in sorters. This is because the
+  // information in sorters is not accurate if some framework
+  // hasn't reregistered. See MESOS-2919 for details.
+  Resources allocated;
+
+  // We track the total and allocated resources on the slave to
+  // avoid calculating it in place every time.
+  //
+  // Note that `available` always contains all the shared resources on the
+  // agent regardless whether they have ever been allocated or not.
+  // NOTE, however, we currently only offer a shared resource only if it has
+  // not been offered in an allocation cycle to a framework. We do this mainly
+  // to preserve the normal offer behavior. This may change in the future
+  // depending on use cases.
+  //
+  // Note that it's possible for the slave to be over-allocated!
+  // In this case, allocated > total.
+  Resources available;
+
+  // We keep a copy of the shared resources to avoid unnecessary copying.
+  Resources shared;
+
+  // We cache whether the agent has gpus as an optimization.
+  bool hasGpu_;
+};
+
+
 // Implements the basic allocator algorithm - first pick a role by
 // some criteria, then pick one of their frameworks to allocate to.
 class HierarchicalAllocatorProcess : public MesosAllocatorProcess
@@ -101,7 +290,7 @@ public:
   }
 
   void initialize(
-      const Duration& allocationInterval,
+      const mesos::allocator::Options& options,
       const lambda::function<
           void(const FrameworkID&,
                const hashmap<std::string, hashmap<SlaveID, Resources>>&)>&
@@ -109,14 +298,7 @@ public:
       const lambda::function<
           void(const FrameworkID&,
                const hashmap<SlaveID, UnavailableResources>&)>&
-        inverseOfferCallback,
-      const Option<std::set<std::string>>&
-        fairnessExcludeResourceNames = None(),
-      bool filterGpuResources = true,
-      const Option<DomainInfo>& domain = None(),
-      const Option<std::vector<Resources>>&
-        minAllocatableResources = None(),
-      const size_t maxCompletedFrameworks = 0) override;
+        inverseOfferCallback) override;
 
   void recover(
       const int _expectedAgentCount,
@@ -229,15 +411,15 @@ public:
   void updateWeights(
       const std::vector<WeightInfo>& weightInfos) override;
 
+  void pause() override;
+
+  void resume() override;
+
 protected:
   // Useful typedefs for dispatch/delay/defer to self()/this.
   typedef HierarchicalAllocatorProcess Self;
   typedef HierarchicalAllocatorProcess This;
 
-  // Idempotent helpers for pausing and resuming allocation.
-  void pause();
-  void resume();
-
   // Allocate any allocatable resources from all known agents.
   process::Future<Nothing> allocate();
 
@@ -293,16 +475,19 @@ protected:
       const FrameworkID& frameworkID,
       const SlaveID& slaveID) const;
 
-  bool allocatable(const Resources& resources);
+  bool allocatable(
+      const Resources& resources,
+      const std::string& role,
+      const Framework& framework) const;
 
   bool initialized;
   bool paused;
 
+  mesos::allocator::Options options;
+
   // Recovery data.
   Option<int> expectedAgentCount;
 
-  Duration allocationInterval;
-
   lambda::function<
       void(const FrameworkID&,
            const hashmap<std::string, hashmap<SlaveID, Resources>>&)>
@@ -316,32 +501,6 @@ protected:
   friend Metrics;
   Metrics metrics;
 
-  // TODO(mzhu): Pull out the nested Framework struct for clearer
-  // logic division with the allocator.
-  struct Framework
-  {
-    Framework(
-        const FrameworkInfo& frameworkInfo,
-        const std::set<std::string>& suppressedRoles,
-        bool active);
-
-    std::set<std::string> roles;
-
-    std::set<std::string> suppressedRoles;
-
-    protobuf::framework::Capabilities capabilities;
-
-    // Active offer and inverse offer filters for the framework.
-    // Offer filters are tied to the role the filtered resources
-    // were allocated to.
-    hashmap<std::string, hashmap<SlaveID, hashset<OfferFilter*>>> offerFilters;
-    hashmap<SlaveID, hashset<InverseOfferFilter*>> inverseOfferFilters;
-
-    bool active;
-
-    process::Owned<FrameworkMetrics> metrics;
-  };
-
   double _event_queue_dispatches()
   {
     return static_cast<double>(eventCount<process::DispatchEvent>());
@@ -365,158 +524,6 @@ protected:
   BoundedHashMap<FrameworkID, process::Owned<FrameworkMetrics>>
     completedFrameworkMetrics;
 
-  // TODO(mzhu): Pull out the nested Slave class for clearer
-  // logic division with the allocator.
-  class Slave
-  {
-  public:
-    Slave(
-        const SlaveInfo& _info,
-        const protobuf::slave::Capabilities& _capabilities,
-        bool _activated,
-        const Resources& _total,
-        const Resources& _allocated)
-      : info(_info),
-        capabilities(_capabilities),
-        activated(_activated),
-        total(_total),
-        allocated(_allocated),
-        shared(_total.shared())
-    {
-      updateAvailable();
-    }
-
-    const Resources& getTotal() const { return total; }
-
-    const Resources& getAllocated() const { return allocated; }
-
-    const Resources& getAvailable() const { return available; }
-
-    void updateTotal(const Resources& newTotal) {
-      total = newTotal;
-      shared = total.shared();
-
-      updateAvailable();
-    }
-
-    void allocate(const Resources& toAllocate)
-    {
-      allocated += toAllocate;
-
-      updateAvailable();
-    }
-
-    void unallocate(const Resources& toUnallocate)
-    {
-      allocated -= toUnallocate;
-
-      updateAvailable();
-    }
-
-    // The `SlaveInfo` that was passed to the allocator when the slave was added
-    // or updated. Currently only two fields are used: `hostname` for host
-    // whitelisting and in log messages, and `domain` for region-aware
-    // scheduling.
-    SlaveInfo info;
-
-    protobuf::slave::Capabilities capabilities;
-
-    bool activated; // Whether to offer resources.
-
-    // Represents a scheduled unavailability due to maintenance for a specific
-    // slave, and the responses from frameworks as to whether they will be able
-    // to gracefully handle this unavailability.
-    //
-    // NOTE: We currently implement maintenance in the allocator to be able to
-    // leverage state and features such as the FrameworkSorter and OfferFilter.
-    struct Maintenance
-    {
-      Maintenance(const Unavailability& _unavailability)
-        : unavailability(_unavailability) {}
-
-      // The start time and optional duration of the event.
-      Unavailability unavailability;
-
-      // A mapping of frameworks to the inverse offer status associated with
-      // this unavailability.
-      //
-      // NOTE: We currently lose this information during a master fail over
-      // since it is not persisted or replicated. This is ok as the new master's
-      // allocator will send out new inverse offers and re-collect the
-      // information. This is similar to all the outstanding offers from an old
-      // master being invalidated, and new offers being sent out.
-      hashmap<FrameworkID, mesos::allocator::InverseOfferStatus> statuses;
-
-      // Represents the "unit of accounting" for maintenance. When a
-      // `FrameworkID` is present in the hashset it means an inverse offer has
-      // been sent out. When it is not present it means no offer is currently
-      // outstanding.
-      hashset<FrameworkID> offersOutstanding;
-    };
-
-    // When the `maintenance` is set the slave is scheduled to be unavailable at
-    // a given point in time, for an optional duration. This information is used
-    // to send out `InverseOffers`.
-    Option<Maintenance> maintenance;
-
-  private:
-    void updateAvailable() {
-      // In order to subtract from the total,
-      // we strip the allocation information.
-      Resources allocated_ = allocated;
-      allocated_.unallocate();
-
-      // Calling `nonShared()` currently copies the underlying resources
-      // and is therefore rather expensive. We avoid it in the common
-      // case that there are no shared resources.
-      //
-      // TODO(mzhu): Ideally there would be a single logical path here.
-      // One solution is to have `Resources` be copy-on-write such that
-      // `nonShared()` performs no copying and instead points to a
-      // subset of the original `Resource` objects.
-      if (shared.empty()) {
-        available = total - allocated_;
-      } else {
-        // Since shared resources are offerable even when they are in use, we
-        // always include them as part of available resources.
-        available = (total.nonShared() - allocated_.nonShared()) + shared;
-      }
-    }
-
-    // Total amount of regular *and* oversubscribed resources.
-    Resources total;
-
-    // Regular *and* oversubscribed resources that are allocated.
-    //
-    // NOTE: We maintain multiple copies of each shared resource allocated
-    // to a slave, where the number of copies represents the number of times
-    // this shared resource has been allocated to (and has not been recovered
-    // from) a specific framework.
-    //
-    // NOTE: We keep track of the slave's allocated resources despite
-    // having that information in sorters. This is because the
-    // information in sorters is not accurate if some framework
-    // hasn't reregistered. See MESOS-2919 for details.
-    Resources allocated;
-
-    // We track the total and allocated resources on the slave to
-    // avoid calculating it in place every time.
-    //
-    // Note that `available` always contains all the shared resources on the
-    // agent regardless whether they have ever been allocated or not.
-    // NOTE, however, we currently only offer a shared resource only if it has
-    // not been offered in an allocation cycle to a framework. We do this mainly
-    // to preserve the normal offer behavior. This may change in the future
-    // depending on use cases.
-    //
-    // Note that it's possible for the slave to be over-allocated!
-    // In this case, allocated > total.
-    Resources available;
-
-    // We keep a copy of the shared resources to avoid unnecessary copying.
-    Resources shared;
-  };
-
   hashmap<SlaveID, Slave> slaves;
 
   // A set of agents that are kept as allocation candidates. Events
@@ -532,34 +539,29 @@ protected:
   // Specifically, we keep track of the roles when a framework subscribes to
   // the role, and/or when there are resources allocated to the role
   // (e.g. some tasks and/or executors are consuming resources under the role).
+  //
+  // NOTE: There is currently not a role entry when there is a
+  // reservation but no allocation / framework!
+  //
+  // TODO(bmahler): Turn this into a `hashmap<string, Role>` that
+  // also tracks a role if it has reservations.
   hashmap<std::string, hashset<FrameworkID>> roles;
 
-  // Configured quota for each role, if any. If a role does not have
-  // an entry here it has the default quota of (no guarantee, no limit).
-  hashmap<std::string, Quota> quotas;
+  // Configured guaranteed resource quantities for each role, if any.
+  // If a role does not have an entry here it has (the default)
+  // no guarantee.
+  hashmap<std::string, ResourceQuantities> quotaGuarantees;
 
   // Aggregated resource reservations on all agents tied to a
-  // particular role, if any. These are stripped scalar quantities
-  // that contain no meta-data.
+  // particular role, if any.
   //
-  // Only roles with non-empty reservations will be stored in the map.
-  hashmap<std::string, Resources> reservationScalarQuantities;
+  // Only roles with non-empty scalar reservation quantities will
+  // be stored in the map.
+  hashmap<std::string, ResourceQuantities> reservationScalarQuantities;
 
   // Slaves to send offers for.
   Option<hashset<std::string>> whitelist;
 
-  // Resources (by name) that will be excluded from a role's fair share.
-  Option<std::set<std::string>> fairnessExcludeResourceNames;
-
-  // Filter GPU resources based on the `GPU_RESOURCES` framework capability.
-  bool filterGpuResources;
-
-  // The master's domain, if any.
-  Option<DomainInfo> domain;
-
-  // The minimum allocatable resources, if any.
-  Option<std::vector<Resources>> minAllocatableResources;
-
   // There are two stages of allocation:
   //
   //   Stage 1: Allocate to satisfy quota guarantees.
