@@ -16,6 +16,7 @@
 
 #include "master/allocator/sorter/drf/sorter.hpp"
 
+#include <iterator>
 #include <set>
 #include <string>
 #include <vector>
@@ -70,102 +71,91 @@ void DRFSorter::initialize(
 
 void DRFSorter::add(const string& clientPath)
 {
-  vector<string> pathElements = strings::tokenize(clientPath, "/");
-  CHECK(!pathElements.empty());
+  CHECK(!clients.contains(clientPath)) << clientPath;
 
-  Node* current = root;
-  Node* lastCreatedNode = nullptr;
+  // Adding a client is a two phase algorithm:
+  //
+  //            root
+  //          /  |  \       Three interesting cases:
+  //         a   e   w        Add a                     (i.e. phase 1(a))
+  //         |      / \       Add e/f, e/f/g, e/f/g/... (i.e. phase 1(b))
+  //         b     .   z      Add w/x, w/x/y, w/x/y/... (i.e. phase 1(c))
+  //
+  //   Phase 1: Walk down the tree until:
+  //     (a) we run out of tokens -> add "." node
+  //     (b) or, we reach a leaf -> transform the leaf into internal + "."
+  //     (c) or, we're at an internal node but can't find the next child
+  //
+  //   Phase 2: For any remaining tokens, walk down creating children:
+  //     (a) if last token of the client path -> create INACTIVE_LEAF
+  //     (b) else, create INTERNAL and keep going
+
+  vector<string> tokens = strings::split(clientPath, "/");
+  auto token = tokens.begin();
 
   // Traverse the tree to add new nodes for each element of the path,
   // if that node doesn't already exist (similar to `mkdir -p`).
-  foreach (const string& element, pathElements) {
-    Node* node = nullptr;
+  Node* current = root;
 
-    foreach (Node* child, current->children) {
-      if (child->name == element) {
-        node = child;
-        break;
-      }
+  // Phase 1:
+  while (true) {
+    // Case (a).
+    if (token == tokens.end()) {
+      Node* virt = new Node(".", Node::INACTIVE_LEAF, current);
+
+      current->addChild(virt);
+      current = virt;
+
+      break;
     }
 
-    if (node != nullptr) {
-      current = node;
-      continue;
-    }
-
-    // We didn't find `element`, so add a new child to `current`.
-    //
-    // If adding this child would result in turning `current` from a
-    // leaf node into an internal node, we need to create an
-    // additional child node: `current` must have been associated with
-    // a client and clients must always be associated with leaf nodes.
+    // Case (b).
     if (current->isLeaf()) {
-      Node* parent = CHECK_NOTNULL(current->parent);
+      Node::Kind oldKind = current->kind;
 
-      parent->removeChild(current);
+      current->parent->removeChild(current);
+      current->kind = Node::INTERNAL;
+      current->parent->addChild(current);
 
-      // Create a node under `parent`. This internal node will take
-      // the place of `current` in the tree.
-      Node* internal = new Node(current->name, Node::INTERNAL, parent);
-      parent->addChild(internal);
-      internal->allocation = current->allocation;
+      Node* virt = new Node(".", oldKind, current);
+      virt->allocation = current->allocation;
 
-      CHECK_EQ(current->path, internal->path);
+      current->addChild(virt);
+      clients[virt->clientPath()] = virt;
 
-      // Update `current` to become a virtual leaf node and a child of
-      // `internal`.
-      current->name = ".";
-      current->parent = internal;
-      current->path = strings::join("/", parent->path, current->name);
-
-      internal->addChild(current);
-
-      CHECK_EQ(internal->path, current->clientPath());
-
-      current = internal;
+      break;
     }
 
-    // Now actually add a new child to `current`.
-    Node* newChild = new Node(element, Node::INTERNAL, current);
-    current->addChild(newChild);
+    Option<Node*> child = [&]() -> Option<Node*> {
+      foreach (Node* c, current->children) {
+        if (c->name == *token) return c;
+      }
+      return None();
+    }();
 
-    current = newChild;
-    lastCreatedNode = newChild;
+    // Case (c).
+    if (child.isNone()) break;
+
+    current = *child;
+    ++token;
   }
 
-  CHECK(current->kind == Node::INTERNAL);
+  // Phase 2:
+  for (; token != tokens.end(); ++token) {
+    Node::Kind kind = (token == tokens.end() - 1)
+      ? Node::INACTIVE_LEAF : Node::INTERNAL;
 
-  // `current` is the node associated with the last element of the
-  // path. If we didn't add `current` to the tree above, create a leaf
-  // node now. For example, if the tree contains "a/b" and we add a
-  // new client "a", we want to create a new leaf node "a/." here.
-  if (current != lastCreatedNode) {
-    Node* newChild = new Node(".", Node::INACTIVE_LEAF, current);
-    current->addChild(newChild);
-    current = newChild;
-  } else {
-    // If we created `current` in the loop above, it was marked an
-    // `INTERNAL` node. It should actually be an inactive leaf node.
-    current->kind = Node::INACTIVE_LEAF;
+    Node* child = new Node(*token, kind, current);
 
-    // `current` has changed from an internal node to an inactive
-    // leaf, so remove and re-add it to its parent. This moves it to
-    // the end of the parent's list of children.
-    CHECK_NOTNULL(current->parent);
-
-    current->parent->removeChild(current);
-    current->parent->addChild(current);
+    current->addChild(child);
+    current = child;
   }
 
-  // `current` is the newly created node associated with the last
-  // element of the path. `current` should be an inactive leaf node.
   CHECK(current->children.empty());
   CHECK(current->kind == Node::INACTIVE_LEAF);
 
-  // Add a new entry to the lookup table. The full path of the newly
-  // added client should not already exist in `clients`.
   CHECK_EQ(clientPath, current->clientPath());
-  CHECK(!clients.contains(clientPath));
+  CHECK(!clients.contains(clientPath)) << clientPath;
 
   clients[clientPath] = current;
 
@@ -209,14 +199,11 @@ void DRFSorter::remove(const string& clientPath)
 
     // Update `parent` to reflect the fact that the resources in the
     // leaf node are no longer allocated to the subtree rooted at
-    // `parent`. We skip `root`, because we never update the
-    // allocation made to the root node.
-    if (parent != root) {
-      foreachpair (const SlaveID& slaveId,
-                   const Resources& resources,
-                   leafAllocation) {
-        parent->allocation.subtract(slaveId, resources);
-      }
+    // `parent`.
+    foreachpair (const SlaveID& slaveId,
+                 const Resources& resources,
+                 leafAllocation) {
+      parent->allocation.subtract(slaveId, resources);
     }
 
     if (current->children.empty()) {
@@ -312,6 +299,24 @@ void DRFSorter::updateWeight(const string& path, double weight)
 
   // TODO(neilc): Avoid dirtying the tree in some circumstances.
   dirty = true;
+
+  // Update the weight of the corresponding internal node,
+  // if it exists (this client may not exist despite there
+  // being a weight).
+  Node* node = find(path);
+
+  if (node == nullptr) {
+    return;
+  }
+
+  // If there is a virtual leaf, we need to move up one level.
+  if (node->name == ".") {
+    node = CHECK_NOTNULL(node->parent);
+  }
+
+  CHECK_EQ(path, node->path);
+
+  node->weight = weight;
 }
 
 
@@ -322,16 +327,47 @@ void DRFSorter::allocated(
 {
   Node* current = CHECK_NOTNULL(find(clientPath));
 
-  // NOTE: We don't currently update the `allocation` for the root
-  // node. This is debatable, but the current implementation doesn't
-  // require looking at the allocation of the root node.
-  while (current != root) {
+  // Walk up the tree adjusting allocations. If the tree is
+  // sorted, we keep it sorted (see below).
+  while (current != nullptr) {
     current->allocation.add(slaveId, resources);
-    current = CHECK_NOTNULL(current->parent);
-  }
 
-  // TODO(neilc): Avoid dirtying the tree in some circumstances.
-  dirty = true;
+    // Note that inactive leaves are not sorted, and are always
+    // stored in `children` after the active leaves and internal
+    // nodes. See the comment on `Node::children`.
+    if (current != root && !dirty && current->kind != Node::INACTIVE_LEAF) {
+      current->share = calculateShare(current);
+
+      vector<Node*>& children = current->parent->children;
+
+      // Locate the node position in the parent's children
+      // and shift it into its sorted position.
+      //
+      // TODO(bmahler): Consider storing the node's position
+      // in the parent's children to avoid scanning.
+      auto position = std::find(children.begin(), children.end(), current);
+      CHECK(position != children.end());
+
+      // Shift left until done (if needed).
+      while (position != children.begin() &&
+             DRFSorter::Node::compareDRF(current, *std::prev(position))) {
+        std::swap(*position, *std::prev(position));
+        --position;
+      }
+
+      // Or, shift right until done (if needed). Note that when
+      // shifting right, we need to stop once we reach the
+      // inactive leaves (see `Node::children`).
+      while (std::next(position) != children.end() &&
+             (*std::next(position))->kind != Node::INACTIVE_LEAF &&
+             DRFSorter::Node::compareDRF(*std::next(position), current)) {
+        std::swap(*position, *std::next(position));
+        ++position;
+      }
+    }
+
+    current = current->parent;
+  }
 }
 
 
@@ -347,12 +383,9 @@ void DRFSorter::update(
 
   Node* current = CHECK_NOTNULL(find(clientPath));
 
-  // NOTE: We don't currently update the `allocation` for the root
-  // node. This is debatable, but the current implementation doesn't
-  // require looking at the allocation of the root node.
-  while (current != root) {
+  while (current != nullptr) {
     current->allocation.update(slaveId, oldAllocation, newAllocation);
-    current = CHECK_NOTNULL(current->parent);
+    current = current->parent;
   }
 
   // Just assume the total has changed, per the TODO above.
@@ -367,15 +400,15 @@ void DRFSorter::unallocated(
 {
   Node* current = CHECK_NOTNULL(find(clientPath));
 
-  // NOTE: We don't currently update the `allocation` for the root
-  // node. This is debatable, but the current implementation doesn't
-  // require looking at the allocation of the root node.
-  while (current != root) {
+  while (current != nullptr) {
     current->allocation.subtract(slaveId, resources);
-    current = CHECK_NOTNULL(current->parent);
+    current = current->parent;
   }
 
-  // TODO(neilc): Avoid dirtying the tree in some circumstances.
+  // TODO(bmahler): Similar to `allocated()`, avoid dirtying the
+  // tree here in favor of adjusting the node positions. This
+  // hasn't been critical to do since we don't allocate upon
+  // resource recovery in the allocator.
   dirty = true;
 }
 
@@ -388,11 +421,17 @@ const hashmap<SlaveID, Resources>& DRFSorter::allocation(
 }
 
 
-const Resources& DRFSorter::allocationScalarQuantities(
+const ResourceQuantities& DRFSorter::allocationScalarQuantities(
     const string& clientPath) const
 {
   const Node* client = CHECK_NOTNULL(find(clientPath));
-  return client->allocation.scalarQuantities;
+  return client->allocation.totals;
+}
+
+
+const ResourceQuantities& DRFSorter::allocationScalarQuantities() const
+{
+  return root->allocation.totals;
 }
 
 
@@ -436,9 +475,9 @@ Resources DRFSorter::allocation(
 }
 
 
-const Resources& DRFSorter::totalScalarQuantities() const
+const ResourceQuantities& DRFSorter::totalScalarQuantities() const
 {
-  return total_.scalarQuantities;
+  return total_.totals;
 }
 
 
@@ -454,14 +493,11 @@ void DRFSorter::add(const SlaveID& slaveId, const Resources& resources)
 
     total_.resources[slaveId] += resources;
 
-    const Resources scalarQuantities =
-      (resources.nonShared() + newShared).createStrippedScalarQuantity();
+    const ResourceQuantities scalarQuantities =
+      ResourceQuantities::fromScalarResources(
+          (resources.nonShared() + newShared).scalars());
 
-    total_.scalarQuantities += scalarQuantities;
-
-    foreach (const Resource& resource, scalarQuantities) {
-      total_.totals[resource.name()] += resource.scalar();
-    }
+    total_.totals += scalarQuantities;
 
     // We have to recalculate all shares when the total resources
     // change, but we put it off until `sort` is called so that if
@@ -488,15 +524,12 @@ void DRFSorter::remove(const SlaveID& slaveId, const Resources& resources)
         return !total_.resources[slaveId].contains(resource);
       });
 
-    const Resources scalarQuantities =
-      (resources.nonShared() + absentShared).createStrippedScalarQuantity();
+    const ResourceQuantities scalarQuantities =
+      ResourceQuantities::fromScalarResources(
+          (resources.nonShared() + absentShared).scalars());
 
-    foreach (const Resource& resource, scalarQuantities) {
-      total_.totals[resource.name()] -= resource.scalar();
-    }
-
-    CHECK(total_.scalarQuantities.contains(scalarQuantities));
-    total_.scalarQuantities -= scalarQuantities;
+    CHECK(total_.totals.contains(scalarQuantities));
+    total_.totals -= scalarQuantities;
 
     if (total_.resources[slaveId].empty()) {
       total_.resources.erase(slaveId);
@@ -551,6 +584,10 @@ vector<string> DRFSorter::sort()
   // inactive leaves sorted after active leaves and internal nodes.
   vector<string> result;
 
+  // TODO(bmahler): This over-reserves where there are inactive
+  // clients, only reserve the number of active clients.
+  result.reserve(clients.size());
+
   std::function<void (const Node*)> listClients =
       [&listClients, &result](const Node* node) {
     foreach (const Node* child, node->children) {
@@ -606,28 +643,26 @@ double DRFSorter::calculateShare(const Node* node) const
       continue;
     }
 
-    if (scalar.value() > 0.0 &&
-        node->allocation.totals.contains(resourceName)) {
-      const double allocation =
-        node->allocation.totals.at(resourceName).value();
-
-      share = std::max(share, allocation / scalar.value());
+    if (scalar.value() <= 0.0) {
+      continue;
     }
+
+    Value::Scalar allocation = node->allocation.totals.get(resourceName);
+
+    share = std::max(share, allocation.value() / scalar.value());
   }
 
-  return share / findWeight(node);
+  return share / getWeight(node);
 }
 
 
-double DRFSorter::findWeight(const Node* node) const
+double DRFSorter::getWeight(const Node* node) const
 {
-  Option<double> weight = weights.get(node->path);
-
-  if (weight.isNone()) {
-    return 1.0;
+  if (node->weight.isNone()) {
+    node->weight = weights.get(node->path).getOrElse(1.0);
   }
 
-  return weight.get();
+  return CHECK_NOTNONE(node->weight);
 }
 
 
