@@ -25,14 +25,6 @@
 #include <string>
 #include <vector>
 
-// Using `boost::circular_buffer::debug_iterator` can lead to segfaults
-// because they are not thread-safe (see MESOS-9177), so we must ensure
-// they're disabled. Both versions of this macro are needed to account
-// for differences between older and newer Boost versions.
-#define BOOST_CB_DISABLE_DEBUG 1
-#define BOOST_CB_ENABLE_DEBUG 0
-#include <boost/circular_buffer.hpp>
-
 #include <mesos/mesos.hpp>
 #include <mesos/resources.hpp>
 #include <mesos/type_utils.hpp>
@@ -50,6 +42,7 @@
 
 #include <mesos/scheduler/scheduler.hpp>
 
+#include <process/collect.hpp>
 #include <process/future.hpp>
 #include <process/limiter.hpp>
 #include <process/http.hpp>
@@ -62,6 +55,7 @@
 
 #include <stout/boundedhashmap.hpp>
 #include <stout/cache.hpp>
+#include <stout/circular_buffer.hpp>
 #include <stout/foreach.hpp>
 #include <stout/hashmap.hpp>
 #include <stout/hashset.hpp>
@@ -73,6 +67,7 @@
 #include <stout/try.hpp>
 #include <stout/uuid.hpp>
 
+#include "common/heartbeater.hpp"
 #include "common/http.hpp"
 #include "common/resources_utils.hpp"
 
@@ -158,6 +153,14 @@ Slave(Master* const _master,
 
   void removeOperation(Operation* operation);
 
+  // Marks a non-speculative operation as an orphan when the originating
+  // framework is torn down by the master, or when an agent reregisters
+  // with operations from unknown frameworks. If the operation is
+  // non-terminal, this has the side effect of modifying the agent's
+  // total resources, and should therefore be followed by
+  // `allocator->updateSlave()`.
+  void markOperationAsOrphan(Operation* operation);
+
   Operation* getOperation(const UUID& uuid) const;
 
   void addOffer(Offer* offer);
@@ -253,6 +256,19 @@ Slave(Master* const _master,
   // unacknowledged status updates on this agent.
   hashmap<UUID, Operation*> operations;
 
+  // Pending operations whose originating framework is unknown.
+  // These operations could be pending, or terminal with unacknowledged
+  // status updates.
+  //
+  // This list can be populated whenever a framework is torn down in the
+  // lifetime of the master, or when an agent reregisters with an operation.
+  //
+  // If the originating framework is completed, the master will
+  // acknowledge any status updates instead of the framework.
+  // If an orphan does not belong to a completed framework, the master
+  // will only acknowledge status updates after a fixed delay.
+  hashset<UUID> orphanedOperations;
+
   // Active offers on this slave.
   hashset<Offer*> offers;
 
@@ -334,109 +350,6 @@ inline std::ostream& operator<<(std::ostream& stream, const Slave& slave)
 }
 
 
-// Represents the streaming HTTP connection to a framework or a client
-// subscribed to the '/api/vX' endpoint.
-struct HttpConnection
-{
-  HttpConnection(const process::http::Pipe::Writer& _writer,
-                 ContentType _contentType,
-                 id::UUID _streamId)
-    : writer(_writer),
-      contentType(_contentType),
-      streamId(_streamId) {}
-
-  // We need to evolve the internal old style message/unversioned event into a
-  // versioned event e.g., `v1::scheduler::Event` or `v1::master::Event`.
-  template <typename Message, typename Event = v1::scheduler::Event>
-  bool send(const Message& message)
-  {
-    ::recordio::Encoder<Event> encoder (lambda::bind(
-        serialize, contentType, lambda::_1));
-
-    return writer.write(encoder.encode(evolve(message)));
-  }
-
-  bool close()
-  {
-    return writer.close();
-  }
-
-  process::Future<Nothing> closed() const
-  {
-    return writer.readerClosed();
-  }
-
-  process::http::Pipe::Writer writer;
-  ContentType contentType;
-  id::UUID streamId;
-};
-
-
-// This process periodically sends heartbeats to a given HTTP connection.
-// The `Message` template parameter is the type of the heartbeat event passed
-// into the heartbeater during construction, while the `Event` template
-// parameter is the versioned event type which is sent to the client.
-// The optional delay parameter is used to specify the delay period before it
-// sends the first heartbeat.
-template <typename Message, typename Event>
-class Heartbeater : public process::Process<Heartbeater<Message, Event>>
-{
-public:
-  Heartbeater(const std::string& _logMessage,
-              const Message& _heartbeatMessage,
-              const HttpConnection& _http,
-              const Duration& _interval,
-              const Option<Duration>& _delay = None(),
-              const Option<lambda::function<void(const Message&)>>&
-                _callback = None())
-    : process::ProcessBase(process::ID::generate("heartbeater")),
-      logMessage(_logMessage),
-      heartbeatMessage(_heartbeatMessage),
-      http(_http),
-      interval(_interval),
-      delay(_delay),
-      callback(_callback) {}
-
-protected:
-  void initialize() override
-  {
-    if (delay.isSome()) {
-      process::delay(
-          delay.get(),
-          this,
-          &Heartbeater<Message, Event>::heartbeat);
-    } else {
-      heartbeat();
-    }
-  }
-
-private:
-  void heartbeat()
-  {
-    // Only send a heartbeat if the connection is not closed.
-    if (http.closed().isPending()) {
-      VLOG(2) << "Sending heartbeat to " << logMessage;
-
-      if (callback.isSome()) {
-        callback.get()(heartbeatMessage);
-      }
-
-      Message message(heartbeatMessage);
-      http.send<Message, Event>(message);
-    }
-
-    process::delay(interval, this, &Heartbeater<Message, Event>::heartbeat);
-  }
-
-  const std::string logMessage;
-  const Message heartbeatMessage;
-  HttpConnection http;
-  const Duration interval;
-  const Option<Duration> delay;
-  const Option<lambda::function<void(const Message&)>> callback;
-};
-
-
 class Master : public ProtobufProcess<Master>
 {
 public:
@@ -452,6 +365,11 @@ public:
 
   ~Master() override;
 
+  // Compare this master's capabilities with registry's minimum capability.
+  // Return the set of capabilities missing from this master.
+  static hashset<std::string> misingMinimumCapabilities(
+      const MasterInfo& masterInfo, const Registry& registry);
+
   // Message handlers.
   void submitScheduler(
       const std::string& name);
@@ -602,7 +520,10 @@ protected:
   void consume(process::ExitedEvent&& event) override;
 
   void exited(const process::UPID& pid) override;
-  void exited(const FrameworkID& frameworkId, const HttpConnection& http);
+  void exited(
+      const FrameworkID& frameworkId,
+      const StreamingHttpConnection<v1::scheduler::Event>& http);
+
   void _exited(Framework* framework);
 
   // Invoked upon noticing a subscriber disconnection.
@@ -709,7 +630,7 @@ protected:
       Framework* framework,
       const FrameworkInfo& frameworkInfo,
       const Option<process::UPID>& pid,
-      const Option<HttpConnection>& http,
+      const Option<StreamingHttpConnection<v1::scheduler::Event>>& http,
       const std::set<std::string>& suppressedRoles);
 
   // Replace the scheduler for a framework with a new process ID, in
@@ -718,7 +639,9 @@ protected:
 
   // Replace the scheduler for a framework with a new HTTP connection,
   // in the event of a scheduler failover.
-  void failoverFramework(Framework* framework, const HttpConnection& http);
+  void failoverFramework(
+      Framework* framework,
+      const StreamingHttpConnection<v1::scheduler::Event>& http);
 
   void _failoverFramework(Framework* framework);
 
@@ -763,12 +686,26 @@ protected:
       const std::string& message,
       Option<process::metrics::Counter> reason = None());
 
+  // Removes an agent from the master's state in the following cases:
+  // * When maintenance is started on an agent
+  // * When an agent registers with a new ID from a previously-known IP + port
+  // * When an agent unregisters itself with an `UnregisterSlaveMessage`
   void _removeSlave(
       Slave* slave,
       const process::Future<bool>& registrarResult,
       const std::string& removalCause,
       Option<process::metrics::Counter> reason = None());
 
+  // Removes an agent from the master's state in the following cases:
+  // * When marking an agent unreachable
+  // * When marking an agent gone
+  //
+  // NOTE that in spite of the name `__removeSlave()`, this function is NOT a
+  // continuation of `_removeSlave()`. Rather, these two functions perform
+  // similar logic for slightly different cases.
+  //
+  // TODO(greggomann): refactor `_removeSlave` and `__removeSlave` into a single
+  // common helper function. (See MESOS-9550)
   void __removeSlave(
       Slave* slave,
       const std::string& message,
@@ -1000,6 +937,12 @@ protected:
   // Remove the operation.
   void removeOperation(Operation* operation);
 
+  // Send operation update for all operations on the agent.
+  void sendBulkOperationFeedback(
+    Slave* slave,
+    OperationState operationState,
+    const std::string& message);
+
   // Attempts to update the allocator by applying the given operation.
   // If successful, updates the slave's resources, sends a
   // 'CheckpointResourcesMessage' to the slave with the updated
@@ -1050,7 +993,7 @@ private:
 
   void drop(
       const process::UPID& from,
-      const scheduler::Call& call,
+      const mesos::scheduler::Call& call,
       const std::string& message);
 
   void drop(
@@ -1060,30 +1003,30 @@ private:
 
   void drop(
       Framework* framework,
-      const scheduler::Call& call,
+      const mesos::scheduler::Call& call,
       const std::string& message);
 
   void drop(
       Framework* framework,
-      const scheduler::Call::Suppress& suppress,
+      const mesos::scheduler::Call::Suppress& suppress,
       const std::string& message);
 
   void drop(
       Framework* framework,
-      const scheduler::Call::Revive& revive,
+      const mesos::scheduler::Call::Revive& revive,
       const std::string& message);
 
   // Call handlers.
   void receive(
       const process::UPID& from,
-      scheduler::Call&& call);
+      mesos::scheduler::Call&& call);
 
   void subscribe(
-      HttpConnection http,
-      const scheduler::Call::Subscribe& subscribe);
+      StreamingHttpConnection<v1::scheduler::Event> http,
+      const mesos::scheduler::Call::Subscribe& subscribe);
 
   void _subscribe(
-      HttpConnection http,
+      StreamingHttpConnection<v1::scheduler::Event> http,
       const FrameworkInfo& frameworkInfo,
       bool force,
       const std::set<std::string>& suppressedRoles,
@@ -1091,7 +1034,7 @@ private:
 
   void subscribe(
       const process::UPID& from,
-      const scheduler::Call::Subscribe& subscribe);
+      const mesos::scheduler::Call::Subscribe& subscribe);
 
   void _subscribe(
       const process::UPID& from,
@@ -1102,74 +1045,74 @@ private:
 
   // Subscribes a client to the 'api/vX' endpoint.
   void subscribe(
-      const HttpConnection& http,
+      const StreamingHttpConnection<v1::master::Event>& http,
       const Option<process::http::authentication::Principal>& principal);
 
   void teardown(Framework* framework);
 
   void accept(
       Framework* framework,
-      scheduler::Call::Accept&& accept);
+      mesos::scheduler::Call::Accept&& accept);
 
   void _accept(
       const FrameworkID& frameworkId,
       const SlaveID& slaveId,
       const Resources& offeredResources,
-      scheduler::Call::Accept&& accept,
+      mesos::scheduler::Call::Accept&& accept,
       const process::Future<
           std::vector<process::Future<bool>>>& authorizations);
 
   void acceptInverseOffers(
       Framework* framework,
-      const scheduler::Call::AcceptInverseOffers& accept);
+      const mesos::scheduler::Call::AcceptInverseOffers& accept);
 
   void decline(
       Framework* framework,
-      scheduler::Call::Decline&& decline);
+      mesos::scheduler::Call::Decline&& decline);
 
   void declineInverseOffers(
       Framework* framework,
-      const scheduler::Call::DeclineInverseOffers& decline);
+      const mesos::scheduler::Call::DeclineInverseOffers& decline);
 
   void revive(
       Framework* framework,
-      const scheduler::Call::Revive& revive);
+      const mesos::scheduler::Call::Revive& revive);
 
   void kill(
       Framework* framework,
-      const scheduler::Call::Kill& kill);
+      const mesos::scheduler::Call::Kill& kill);
 
   void shutdown(
       Framework* framework,
-      const scheduler::Call::Shutdown& shutdown);
+      const mesos::scheduler::Call::Shutdown& shutdown);
 
   void acknowledge(
       Framework* framework,
-      scheduler::Call::Acknowledge&& acknowledge);
+      mesos::scheduler::Call::Acknowledge&& acknowledge);
 
   void acknowledgeOperationStatus(
       Framework* framework,
-      scheduler::Call::AcknowledgeOperationStatus&& acknowledge);
+      mesos::scheduler::Call::AcknowledgeOperationStatus&& acknowledge);
 
   void reconcile(
       Framework* framework,
-      scheduler::Call::Reconcile&& reconcile);
+      mesos::scheduler::Call::Reconcile&& reconcile);
 
-  scheduler::Response::ReconcileOperations reconcileOperations(
+  void reconcileOperations(
       Framework* framework,
-      const scheduler::Call::ReconcileOperations& reconcile);
+      mesos::scheduler::Call::ReconcileOperations&& reconcile);
 
   void message(
       Framework* framework,
-      scheduler::Call::Message&& message);
+      mesos::scheduler::Call::Message&& message);
 
   void request(
       Framework* framework,
-      const scheduler::Call::Request& request);
+      const mesos::scheduler::Call::Request& request);
 
   void suppress(
       Framework* framework,
-      const scheduler::Call::Suppress& suppress);
+      const mesos::scheduler::Call::Suppress& suppress);
 
   bool elected() const
   {
@@ -1185,8 +1128,8 @@ private:
       const hashset<SlaveID>& toRemoveGone,
       const process::Future<bool>& registrarResult);
 
-  process::Future<bool> authorizeLogAccess(
-      const Option<process::http::authentication::Principal>& principal);
+  std::vector<std::string> filterRoles(
+      const process::Owned<ObjectApprovers>& approvers) const;
 
   /**
    * Returns whether the given role is on the whitelist.
@@ -1223,6 +1166,11 @@ private:
         const Option<process::http::authentication::Principal>&
             principal) const;
 
+    process::Future<process::http::Response> update(
+        const mesos::master::Call& call,
+        const Option<process::http::authentication::Principal>& principal)
+      const;
+
     process::Future<process::http::Response> set(
         const mesos::master::Call& call,
         const Option<process::http::authentication::Principal>&
@@ -1244,27 +1192,26 @@ private:
             principal) const;
 
   private:
-    // Heuristically tries to determine whether a quota request could
-    // reasonably be satisfied given the current cluster capacity. The
-    // goal is to determine whether a user may accidentally request an
-    // amount of resources that would prevent frameworks without quota
-    // from getting any offers. A force flag will allow users to bypass
-    // this check.
+    // Returns an error if the total quota guarantees overcommits
+    // the cluster. This is not a quota satisfiability check: it's
+    // possible that quota is unsatisfiable even if the quota
+    // does not overcommit the cluster.
+
+    // Returns an error if the total quota guarantees overcommits
+    // the cluster. This is not a quota satisfiability check: it's
+    // possible that quota is unsatisfiable even if the quota
+    // does not overcommit the cluster. Specifically, we verify that
+    // the following inequality holds:
     //
-    // The heuristic tests whether the total quota, including the new
-    // request, does not exceed the sum of non-static cluster resources,
-    // i.e. the following inequality holds:
-    //   total - statically reserved >= total quota + quota request
+    //    total cluster capacity >= total quota w/ quota request applied
     //
-    // Please be advised that:
-    //   * It is up to an allocator how to satisfy quota (for example,
-    //     what resources to account towards quota, as well as which
-    //     resources to consider allocatable for quota).
-    //   * Even if there are enough resources at the moment of this check,
-    //     agents may terminate at any time, rendering the cluster under
-    //     quota.
-    Option<Error> capacityHeuristic(
-        const mesos::quota::QuotaInfo& request) const;
+    // Note, total cluster capacity accounts resources of all the
+    // registered agents, including resources from resource providers
+    // as well as reservations (both static and dynamic ones).
+    static Option<Error> overcommitCheck(
+        const std::vector<Resources>& agents,
+        const hashmap<std::string, Quota>& quotas,
+        const mesos::quota::QuotaInfo& request);
 
     // We always want to rescind offers after the capacity heuristic. The
     // reason for this is the race between the allocator and the master:
@@ -1390,12 +1337,66 @@ private:
     Master* master;
   };
 
+public:
+  // Inner class used to namespace HTTP handlers that do not change the
+  // underlying master object.
+  //
+  // Endpoints served by this handler are only permitted to depend on
+  // the request query parameters and the authorization filters to
+  // make caching of responses possible.
+  //
+  // NOTE: Most member functions of this class are not routed directly but
+  // dispatched from their corresponding handlers in the outer `Http` class.
+  // This is because deciding whether an incoming request is read-only often
+  // requires some inspection, e.g. distinguishing between "GET" and "POST"
+  // requests to the same endpoint.
+  class ReadOnlyHandler
+  {
+  public:
+    explicit ReadOnlyHandler(const Master* _master) : master(_master) {}
+
+    // /frameworks
+    process::http::Response frameworks(
+        const hashmap<std::string, std::string>& queryParameters,
+        const process::Owned<ObjectApprovers>& approvers) const;
+
+    // /roles
+    process::http::Response roles(
+        const hashmap<std::string, std::string>& queryParameters,
+        const process::Owned<ObjectApprovers>& approvers) const;
+
+    // /slaves
+    process::http::Response slaves(
+        const hashmap<std::string, std::string>& queryParameters,
+        const process::Owned<ObjectApprovers>& approvers) const;
+
+    // /state
+    process::http::Response state(
+        const hashmap<std::string, std::string>& queryParameters,
+        const process::Owned<ObjectApprovers>& approvers) const;
+
+    // /state-summary
+    process::http::Response stateSummary(
+        const hashmap<std::string, std::string>& queryParameters,
+        const process::Owned<ObjectApprovers>& approvers) const;
+
+    // /tasks
+    process::http::Response tasks(
+        const hashmap<std::string, std::string>& queryParameters,
+        const process::Owned<ObjectApprovers>& approvers) const;
+
+  private:
+    const Master* master;
+  };
+
+private:
   // Inner class used to namespace HTTP route handlers (see
   // master/http.cpp for implementations).
   class Http
   {
   public:
     explicit Http(Master* _master) : master(_master),
+                                     readonlyHandler(_master),
                                      quotaHandler(_master),
                                      weightsHandler(_master) {}
 
@@ -1430,6 +1431,8 @@ private:
             principal) const;
 
     // /master/frameworks
+    //
+    // NOTE: Requests to this endpoint are batched.
     process::Future<process::http::Response> frameworks(
         const process::http::Request& request,
         const Option<process::http::authentication::Principal>&
@@ -1450,6 +1453,8 @@ private:
             principal) const;
 
     // /master/roles
+    //
+    // NOTE: Requests to this endpoint are batched.
     process::Future<process::http::Response> roles(
         const process::http::Request& request,
         const Option<process::http::authentication::Principal>&
@@ -1462,6 +1467,8 @@ private:
             principal) const;
 
     // /master/slaves
+    //
+    // NOTE: Requests to this endpoint are batched.
     process::Future<process::http::Response> slaves(
         const process::http::Request& request,
         const Option<process::http::authentication::Principal>&
@@ -1473,15 +1480,19 @@ private:
     process::Future<process::http::Response> state(
         const process::http::Request& request,
         const Option<process::http::authentication::Principal>&
-            principal);
+            principal) const;
 
     // /master/state-summary
+    //
+    // NOTE: Requests to this endpoint are batched.
     process::Future<process::http::Response> stateSummary(
         const process::http::Request& request,
         const Option<process::http::authentication::Principal>&
             principal) const;
 
     // /master/tasks
+    //
+    // NOTE: Requests to this endpoint are batched.
     process::Future<process::http::Response> tasks(
         const process::http::Request& request,
         const Option<process::http::authentication::Principal>&
@@ -1561,17 +1572,6 @@ private:
         const Option<process::http::authentication::Principal>&
             principal) const;
 
-    // A continuation for `state()`. Schedules request processing in a batch
-    // of other '/state' requests.
-    process::Future<process::http::Response> deferStateRequest(
-        const process::http::Request& request,
-        const process::Owned<ObjectApprovers>& approvers);
-
-    // A helper that responds to batched, i.e., accumulated, '/state'
-    // requests in parallel, i.e., a continuation for `deferStateRequest()`.
-    // See also `BatchedStateRequest`.
-    void processStateRequestsBatch();
-
     process::Future<std::vector<const Task*>> _tasks(
         const size_t limit,
         const size_t offset,
@@ -1649,23 +1649,15 @@ private:
      *
      * @param slaveId The ID of the slave that the operation is
      *     updating.
-     * @param required The resources needed to satisfy the operation.
-     *     This is used for an optimization where we try to only
-     *     rescind offers that would contribute to satisfying the
-     *     operation.
      * @param operation The operation to be performed.
      *
-     * @return Returns 'OK' if successful, 'Conflict' otherwise.
+     * @return Returns 'OK' if successful, 'BadRequest' if the
+     *     operation is malformed, 'Conflict' otherwise.
      */
     process::Future<process::http::Response> _operation(
         const SlaveID& slaveId,
-        Resources required,
         const Offer::Operation& operation) const;
 
-    process::Future<std::vector<std::string>> _roles(
-        const Option<process::http::authentication::Principal>&
-            principal) const;
-
     // Master API handlers.
 
     process::Future<process::http::Response> getAgents(
@@ -1838,11 +1830,13 @@ private:
 
     process::Future<process::http::Response> reconcileOperations(
         Framework* framework,
-        const scheduler::Call::ReconcileOperations& call,
+        const mesos::scheduler::Call::ReconcileOperations& call,
         ContentType contentType) const;
 
     Master* master;
 
+    ReadOnlyHandler readonlyHandler;
+
     // NOTE: The quota specific pieces of the Operator API are factored
     // out into this separate class.
     QuotaHandler quotaHandler;
@@ -1851,16 +1845,38 @@ private:
     // out into this separate class.
     WeightsHandler weightsHandler;
 
-    // TODO(alexr): Consider adding a `type` or `handler` field to expand
-    // batching to other heavy read-only requests, e.g., '/state-summary'.
-    struct BatchedStateRequest
+    // Since the Master actor is one of the most loaded in a typical Mesos
+    // installation, we take some extra care to keep the backlog small.
+    // In particular, all read-only requests are batched and executed in
+    // parallel, instead of going through the master queue separately.
+
+    typedef process::http::Response
+      (Master::ReadOnlyHandler::*ReadOnlyRequestHandler)(
+          const hashmap<std::string, std::string>&,
+          const process::Owned<ObjectApprovers>&) const;
+
+    process::Future<process::http::Response> deferBatchedRequest(
+        ReadOnlyRequestHandler handler,
+        const Option<process::http::authentication::Principal>& principal,
+        const hashmap<std::string, std::string>& queryParameters,
+        const process::Owned<ObjectApprovers>& approvers) const;
+
+    void processRequestsBatch() const;
+
+    struct BatchedRequest
     {
-      process::http::Request request;
+      ReadOnlyRequestHandler handler;
+      hashmap<std::string, std::string> queryParameters;
+      Option<process::http::authentication::Principal> principal;
       process::Owned<ObjectApprovers> approvers;
+
+      // NOTE: The returned response should be either of type
+      // `BODY` or `PATH`, since `PIPE`-type responses would
+      // break the deduplication mechanism.
       process::Promise<process::http::Response> promise;
     };
 
-    std::vector<BatchedStateRequest> batchedStateRequests;
+    mutable std::vector<BatchedRequest> batchedRequests;
   };
 
   Master(const Master&);              // No copying.
@@ -2094,7 +2110,9 @@ private:
 
   struct Subscribers
   {
-    Subscribers(Master* _master) : master(_master) {};
+    Subscribers(Master* _master, size_t maxSubscribers)
+      : master(_master),
+        subscribed(maxSubscribers) {};
 
     // Represents a client subscribed to the 'api/vX' endpoint.
     //
@@ -2103,25 +2121,20 @@ private:
     struct Subscriber
     {
       Subscriber(
-          const HttpConnection& _http,
+          const StreamingHttpConnection<v1::master::Event>& _http,
           const Option<process::http::authentication::Principal> _principal)
         : http(_http),
-          principal(_principal)
-      {
-        mesos::master::Event event;
-        event.set_type(mesos::master::Event::HEARTBEAT);
-
-        heartbeater =
-          process::Owned<Heartbeater<mesos::master::Event, v1::master::Event>>(
-              new Heartbeater<mesos::master::Event, v1::master::Event>(
-                  "subscriber " + stringify(http.streamId),
-                  event,
-                  http,
-                  DEFAULT_HEARTBEAT_INTERVAL,
-                  DEFAULT_HEARTBEAT_INTERVAL));
-
-        process::spawn(heartbeater.get());
-      }
+          heartbeater(
+              "subscriber " + stringify(http.streamId),
+              []() {
+                mesos::master::Event event;
+                event.set_type(mesos::master::Event::HEARTBEAT);
+                return event;
+              }(),
+              http,
+              DEFAULT_HEARTBEAT_INTERVAL,
+              DEFAULT_HEARTBEAT_INTERVAL),
+          principal(_principal) {}
 
       // Not copyable, not assignable.
       Subscriber(const Subscriber&) = delete;
@@ -2142,14 +2155,10 @@ private:
         // after passing ownership to the `Subscriber` object. See MESOS-5843
         // for more details.
         http.close();
-
-        terminate(heartbeater.get());
-        wait(heartbeater.get());
       }
 
-      HttpConnection http;
-      process::Owned<Heartbeater<mesos::master::Event, v1::master::Event>>
-        heartbeater;
+      StreamingHttpConnection<v1::master::Event> http;
+      ResponseHeartbeater<mesos::master::Event, v1::master::Event> heartbeater;
       const Option<process::http::authentication::Principal> principal;
     };
 
@@ -2163,7 +2172,7 @@ private:
 
     // Active subscribers to the 'api/vX' endpoint keyed by the stream
     // identifier.
-    hashmap<id::UUID, process::Owned<Subscriber>> subscribed;
+    BoundedHashMap<id::UUID, process::Owned<Subscriber>> subscribed;
   };
 
   Subscribers subscribers;
@@ -2233,6 +2242,14 @@ private:
   double _slaves_inactive();
   double _slaves_unreachable();
 
+  // TODO(bevers): Remove these and make the above functions
+  // const instead after MESOS-4995 is resolved.
+  double _const_slaves_connected() const;
+  double _const_slaves_disconnected() const;
+  double _const_slaves_active() const;
+  double _const_slaves_inactive() const;
+  double _const_slaves_unreachable() const;
+
   double _frameworks_connected();
   double _frameworks_disconnected();
   double _frameworks_active();
@@ -2324,7 +2341,7 @@ struct Framework
   Framework(Master* const master,
             const Flags& masterFlags,
             const FrameworkInfo& info,
-            const HttpConnection& _http,
+            const StreamingHttpConnection<v1::scheduler::Event>& _http,
             const process::Time& time = process::Clock::now());
 
   Framework(Master* const master,
@@ -2393,7 +2410,8 @@ struct Framework
 
   void updateConnection(const process::UPID& newPid);
 
-  void updateConnection(const HttpConnection& newHttp);
+  void updateConnection(
+      const StreamingHttpConnection<v1::scheduler::Event>& newHttp);
 
   // Closes the HTTP connection and stops the heartbeat.
   //
@@ -2425,7 +2443,7 @@ struct Framework
   // (scheduler driver). At most one of `http` and `pid` will be set
   // according to the last connection made by the framework; neither
   // field will be set if the framework is in state `RECOVERED`.
-  Option<HttpConnection> http;
+  Option<StreamingHttpConnection<v1::scheduler::Event>> http;
   Option<process::UPID> pid;
 
   State state;
@@ -2445,9 +2463,9 @@ struct Framework
   // Tasks launched by this framework that have reached a terminal
   // state and have had all their updates acknowledged. We only keep a
   // fixed-size cache to avoid consuming too much memory. We use
-  // boost::circular_buffer rather than BoundedHashMap because there
+  // circular_buffer rather than BoundedHashMap because there
   // can be multiple completed tasks with the same task ID.
-  boost::circular_buffer<process::Owned<Task>> completedTasks;
+  circular_buffer<process::Owned<Task>> completedTasks;
 
   // When an agent is marked unreachable, tasks running on it are stored
   // here. We only keep a fixed-size cache to avoid consuming too much memory.
@@ -2506,10 +2524,10 @@ struct Framework
   hashmap<SlaveID, Resources> offeredResources;
 
   // This is only set for HTTP frameworks.
-  Option<process::Owned<Heartbeater<scheduler::Event, v1::scheduler::Event>>>
+  process::Owned<ResponseHeartbeater<scheduler::Event, v1::scheduler::Event>>
     heartbeater;
 
-  // This is used for per-framwork metrics.
+  // This is used for per-framework metrics.
   FrameworkMetrics metrics;
 
 private:
@@ -2528,21 +2546,35 @@ private:
 template <typename Message>
 void Framework::send(const Message& message)
 {
-  if (!connected()) {
-    LOG(WARNING) << "Master attempted to send message to disconnected"
-                 << " framework " << *this;
-  }
-
   metrics.incrementEvent(message);
 
+  if (!connected()) {
+    LOG(WARNING) << "Master attempting to send message to disconnected"
+                 << " framework " << *this;
+
+    // NOTE: We proceed here without returning to support the case where a
+    // "disconnected" framework is still talking to the master and the master
+    // wants to shut it down by sending a `FrameworkErrorMessage`. This can
+    // occur in a one-way network partition where the master -> framework link
+    // is broken but the framework -> master link remains intact. Note that we
+    // have no periodic heartbeats between the master and pid-based schedulers.
+    //
+    // TODO(chhsiao): Update the `FrameworkErrorMessage` call-sites that rely on
+    // the lack of a `return` here to directly call `process::send` so that this
+    // function doesn't need to deal with the special case. Then we can check
+    // that one of `http` or `pid` is set if the framework is connected.
+  }
+
   if (http.isSome()) {
     if (!http->send(message)) {
-      LOG(WARNING) << "Unable to send event to framework " << *this << ":"
+      LOG(WARNING) << "Unable to send message to framework " << *this << ":"
                    << " connection closed";
     }
-  } else {
-    CHECK_SOME(pid);
+  } else if (pid.isSome()) {
     master->send(pid.get(), message);
+  } else {
+    LOG(WARNING) << "Unable to send message to framework " << *this << ":"
+                 << " framework is recovered but has not reregistered";
   }
 }
 
