@@ -80,6 +80,10 @@ Metrics::Metrics(const Master& master)
     outstanding_offers(
         "master/outstanding_offers",
         defer(master, &Master::_outstanding_offers)),
+    operation_states(
+        "master/operations/"),
+    operator_event_stream_subscribers(
+        "master/operator_event_stream_subscribers"),
     tasks_staging(
         "master/tasks_staging",
         defer(master, &Master::_tasks_staging)),
@@ -113,6 +117,8 @@ Metrics::Metrics(const Master& master)
         "master/tasks_gone_by_operator"),
     dropped_messages(
         "master/dropped_messages"),
+    http_cache_hits(
+        "master/http_cache_hits"),
     messages_register_framework(
         "master/messages_register_framework"),
     messages_reregister_framework(
@@ -153,6 +159,8 @@ Metrics::Metrics(const Master& master)
         "master/messages_unregister_slave"),
     messages_status_update(
         "master/messages_status_update"),
+    messages_operation_status_update(
+        "master/messages_operation_status_update"),
     messages_exited_executor(
         "master/messages_exited_executor"),
     messages_update_slave(
@@ -175,6 +183,10 @@ Metrics::Metrics(const Master& master)
         "master/valid_status_update_acknowledgements"),
     invalid_status_update_acknowledgements(
         "master/invalid_status_update_acknowledgements"),
+    valid_operation_status_updates(
+        "master/valid_operation_status_updates"),
+    invalid_operation_status_updates(
+        "master/invalid_operation_status_updates"),
     valid_operation_status_update_acknowledgements(
         "master/valid_operation_status_update_acknowledgements"),
     invalid_operation_status_update_acknowledgements(
@@ -232,6 +244,8 @@ Metrics::Metrics(const Master& master)
 
   process::metrics::add(outstanding_offers);
 
+  process::metrics::add(operator_event_stream_subscribers);
+
   process::metrics::add(tasks_staging);
   process::metrics::add(tasks_starting);
   process::metrics::add(tasks_running);
@@ -247,6 +261,7 @@ Metrics::Metrics(const Master& master)
   process::metrics::add(tasks_gone_by_operator);
 
   process::metrics::add(dropped_messages);
+  process::metrics::add(http_cache_hits);
 
   // Messages from schedulers.
   process::metrics::add(messages_register_framework);
@@ -271,6 +286,7 @@ Metrics::Metrics(const Master& master)
   process::metrics::add(messages_reregister_slave);
   process::metrics::add(messages_unregister_slave);
   process::metrics::add(messages_status_update);
+  process::metrics::add(messages_operation_status_update);
   process::metrics::add(messages_exited_executor);
   process::metrics::add(messages_update_slave);
 
@@ -289,6 +305,9 @@ Metrics::Metrics(const Master& master)
   process::metrics::add(valid_status_update_acknowledgements);
   process::metrics::add(invalid_status_update_acknowledgements);
 
+  process::metrics::add(valid_operation_status_updates);
+  process::metrics::add(invalid_operation_status_updates);
+
   process::metrics::add(valid_operation_status_update_acknowledgements);
   process::metrics::add(invalid_operation_status_update_acknowledgements);
 
@@ -361,6 +380,30 @@ Metrics::Metrics(const Master& master)
     process::metrics::add(used);
     process::metrics::add(percent);
   }
+
+  for (int index = 0;
+       index < Offer::Operation::Type_descriptor()->value_count();
+       index++) {
+    const google::protobuf::EnumValueDescriptor* descriptor =
+      Offer::Operation::Type_descriptor()->value(index);
+
+    const Offer::Operation::Type type =
+      static_cast<Offer::Operation::Type>(descriptor->number());
+
+    // Skip `LAUNCH` and `LAUNCH_GROUP` because they are special-cased
+    // in the master and don't go through the usual offer operation feedback
+    // cycle.
+    if (type == Offer::Operation::UNKNOWN ||
+        type == Offer::Operation::LAUNCH ||
+        type == Offer::Operation::LAUNCH_GROUP) {
+      continue;
+    }
+
+    std::string prefix =
+      "master/operations/" + strings::lower(descriptor->name()) + "/";
+
+    operation_type_states.emplace(type, prefix);
+  }
 }
 
 
@@ -383,6 +426,8 @@ Metrics::~Metrics()
 
   process::metrics::remove(outstanding_offers);
 
+  process::metrics::remove(operator_event_stream_subscribers);
+
   process::metrics::remove(tasks_staging);
   process::metrics::remove(tasks_starting);
   process::metrics::remove(tasks_running);
@@ -398,6 +443,7 @@ Metrics::~Metrics()
   process::metrics::remove(tasks_gone_by_operator);
 
   process::metrics::remove(dropped_messages);
+  process::metrics::remove(http_cache_hits);
 
   // Messages from schedulers.
   process::metrics::remove(messages_register_framework);
@@ -422,6 +468,7 @@ Metrics::~Metrics()
   process::metrics::remove(messages_reregister_slave);
   process::metrics::remove(messages_unregister_slave);
   process::metrics::remove(messages_status_update);
+  process::metrics::remove(messages_operation_status_update);
   process::metrics::remove(messages_exited_executor);
   process::metrics::remove(messages_update_slave);
 
@@ -440,6 +487,9 @@ Metrics::~Metrics()
   process::metrics::remove(valid_status_update_acknowledgements);
   process::metrics::remove(invalid_status_update_acknowledgements);
 
+  process::metrics::remove(valid_operation_status_updates);
+  process::metrics::remove(invalid_operation_status_updates);
+
   process::metrics::remove(valid_operation_status_update_acknowledgements);
   process::metrics::remove(invalid_operation_status_update_acknowledgements);
 
@@ -552,8 +602,123 @@ void Metrics::incrementTasksStates(
 }
 
 
-FrameworkMetrics::FrameworkMetrics(const FrameworkInfo& _frameworkInfo)
+Metrics::OperationStates::OperationStates(const std::string& prefix)
+  : total(prefix + "total"),
+    pending(prefix + "pending"),
+    recovering(prefix + "recovering"),
+    unreachable(prefix + "unreachable"),
+    finished(prefix + "finished"),
+    failed(prefix + "failed"),
+    error(prefix + "error"),
+    dropped(prefix + "dropped"),
+    gone_by_operator(prefix + "gone_by_operator")
+{
+  process::metrics::add(total);
+  process::metrics::add(pending);
+  process::metrics::add(recovering);
+  process::metrics::add(finished);
+  process::metrics::add(failed);
+  process::metrics::add(error);
+  process::metrics::add(dropped);
+  process::metrics::add(unreachable);
+  process::metrics::add(gone_by_operator);
+}
+
+
+Metrics::OperationStates::~OperationStates()
+{
+  process::metrics::remove(total);
+  process::metrics::remove(pending);
+  process::metrics::remove(recovering);
+  process::metrics::remove(finished);
+  process::metrics::remove(failed);
+  process::metrics::remove(error);
+  process::metrics::remove(dropped);
+  process::metrics::remove(unreachable);
+  process::metrics::remove(gone_by_operator);
+}
+
+
+void Metrics::OperationStates::update(
+    const OperationState& operationState,
+    int delta)
+{
+  total += delta;
+
+  switch(operationState) {
+    case OPERATION_FINISHED:
+      finished += delta;
+      break;
+    case OPERATION_DROPPED:
+      dropped += delta;
+      break;
+    case OPERATION_ERROR:
+      error += delta;
+      break;
+    case OPERATION_FAILED:
+      failed += delta;
+      break;
+    case OPERATION_GONE_BY_OPERATOR:
+      gone_by_operator += delta;
+      break;
+    case OPERATION_PENDING:
+      pending += delta;
+      break;
+    case OPERATION_UNREACHABLE:
+      unreachable += delta;
+      break;
+    case OPERATION_RECOVERING:
+      recovering += delta;
+      break;
+    case OPERATION_UNSUPPORTED:
+    case OPERATION_UNKNOWN:
+      LOG(ERROR) << "Unexpected operation state: " << operationState;
+      break;
+  }
+}
+
+
+void Metrics::incrementOperationState(
+    Offer::Operation::Type type,
+    const OperationState& state)
+{
+  operation_states.update(state, 1);
+  if (operation_type_states.count(type)) {
+    operation_type_states.at(type).update(state, 1);
+  }
+}
+
+
+void Metrics::decrementOperationState(
+    Offer::Operation::Type type,
+    const OperationState& state)
+{
+  operation_states.update(state, -1);
+  if (operation_type_states.count(type)) {
+    operation_type_states.at(type).update(state, -1);
+  }
+}
+
+
+void Metrics::transitionOperationState(
+    Offer::Operation::Type type,
+    const OperationState& oldState,
+    const OperationState& newState)
+{
+  if (oldState == newState) {
+    return;  // Nothing to do.
+  }
+
+  decrementOperationState(type, oldState);
+  incrementOperationState(type, newState);
+}
+
+
+FrameworkMetrics::FrameworkMetrics(
+    const FrameworkInfo& _frameworkInfo,
+    bool _publishPerFrameworkMetrics)
   : frameworkInfo(_frameworkInfo),
+    publishPerFrameworkMetrics(_publishPerFrameworkMetrics),
     subscribed(
         getFrameworkMetricPrefix(frameworkInfo) + "subscribed"),
     calls(
@@ -571,15 +736,14 @@ FrameworkMetrics::FrameworkMetrics(const FrameworkInfo& _frameworkInfo)
     operations(
         getFrameworkMetricPrefix(frameworkInfo) + "operations")
 {
-  process::metrics::add(subscribed);
-
-  process::metrics::add(offers_sent);
-  process::metrics::add(offers_accepted);
-  process::metrics::add(offers_declined);
-  process::metrics::add(offers_rescinded);
+  addMetric(subscribed);
+  addMetric(offers_sent);
+  addMetric(offers_accepted);
+  addMetric(offers_declined);
+  addMetric(offers_rescinded);
 
   // Add metrics for scheduler calls.
-  process::metrics::add(calls);
+  addMetric(calls);
   for (int index = 0;
        index < scheduler::Call::Type_descriptor()->value_count();
        index++) {
@@ -598,11 +762,11 @@ FrameworkMetrics::FrameworkMetrics(const FrameworkInfo& _frameworkInfo)
         strings::lower(descriptor->name()));
 
     call_types.put(type, counter);
-    process::metrics::add(counter);
+    addMetric(counter);
   }
 
   // Add metrics for scheduler events.
-  process::metrics::add(events);
+  addMetric(events);
   for (int index = 0;
        index < scheduler::Event::Type_descriptor()->value_count();
        index++) {
@@ -621,7 +785,7 @@ FrameworkMetrics::FrameworkMetrics(const FrameworkInfo& _frameworkInfo)
         strings::lower(descriptor->name()));
 
     event_types.put(type, counter);
-    process::metrics::add(counter);
+    addMetric(counter);
   }
 
   // Add metrics for both active and terminal task states.
@@ -637,19 +801,19 @@ FrameworkMetrics::FrameworkMetrics(const FrameworkInfo& _frameworkInfo)
           strings::lower(descriptor->name()));
 
       terminal_task_states.put(state, counter);
-      process::metrics::add(counter);
+      addMetric(counter);
     } else {
       PushGauge gauge = PushGauge(
           getFrameworkMetricPrefix(frameworkInfo) + "tasks/active/" +
           strings::lower(TaskState_Name(state)));
 
       active_task_states.put(state, gauge);
-      process::metrics::add(gauge);
+      addMetric(gauge);
     }
   }
 
   // Add metrics for offer operations.
-  process::metrics::add(operations);
+  addMetric(operations);
   for (int index = 0;
        index < Offer::Operation::Type_descriptor()->value_count();
        index++) {
@@ -668,41 +832,41 @@ FrameworkMetrics::FrameworkMetrics(const FrameworkInfo& _frameworkInfo)
       "operations/" + strings::lower(descriptor->name()));
 
     operation_types.put(type, counter);
-    process::metrics::add(counter);
+    addMetric(counter);
   }
 }
 
 
 FrameworkMetrics::~FrameworkMetrics()
 {
-  process::metrics::remove(subscribed);
+  removeMetric(subscribed);
 
-  process::metrics::remove(calls);
+  removeMetric(calls);
   foreachvalue (const Counter& counter, call_types) {
-    process::metrics::remove(counter);
+    removeMetric(counter);
   }
 
   process::metrics::remove(events);
   foreachvalue (const Counter& counter, event_types) {
-    process::metrics::remove(counter);
+    removeMetric(counter);
   }
 
-  process::metrics::remove(offers_sent);
-  process::metrics::remove(offers_accepted);
-  process::metrics::remove(offers_declined);
-  process::metrics::remove(offers_rescinded);
+  removeMetric(offers_sent);
+  removeMetric(offers_accepted);
+  removeMetric(offers_declined);
+  removeMetric(offers_rescinded);
 
   foreachvalue (const Counter& counter, terminal_task_states) {
-    process::metrics::remove(counter);
+    removeMetric(counter);
   }
 
   foreachvalue (const PushGauge& gauge, active_task_states) {
-    process::metrics::remove(gauge);
+    removeMetric(gauge);
   }
 
   process::metrics::remove(operations);
   foreachvalue (const Counter& counter, operation_types) {
-    process::metrics::remove(counter);
+    removeMetric(counter);
   }
 }
 
@@ -753,6 +917,22 @@ string getFrameworkMetricPrefix(const FrameworkInfo& frameworkInfo)
 }
 
 
+template <typename T>
+void FrameworkMetrics::addMetric(const T& metric)  {
+  if (publishPerFrameworkMetrics) {
+    process::metrics::add(metric);
+  }
+}
+
+
+template <typename T>
+void FrameworkMetrics::removeMetric(const T& metric)  {
+  if (publishPerFrameworkMetrics) {
+    process::metrics::remove(metric);
+  }
+}
+
+
 void FrameworkMetrics::incrementEvent(const scheduler::Event& event)
 {
   ++CHECK_NOTNONE(event_types.get(event.type()));
