@@ -40,6 +40,7 @@
 #include "checks/health_checker.hpp"
 
 #include "common/protobuf_utils.hpp"
+#include "common/resource_quantities.hpp"
 #include "common/validation.hpp"
 
 #include "master/master.hpp"
@@ -60,9 +61,7 @@ namespace validation {
 namespace master {
 namespace call {
 
-Option<Error> validate(
-    const mesos::master::Call& call,
-    const Option<Principal>& principal)
+Option<Error> validate(const mesos::master::Call& call)
 {
   if (!call.IsInitialized()) {
     return Error("Not initialized: " + call.InitializationErrorString());
@@ -243,12 +242,22 @@ Option<Error> validate(
     case mesos::master::Call::GET_QUOTA:
       return None();
 
+    case mesos::master::Call::UPDATE_QUOTA:
+      if (!call.has_update_quota()) {
+        return Error("Expecting 'update_quota' to be present");
+      }
+      return None();
+
+    // TODO(bmahler): Add this to a deprecated call section
+    // at the bottom once deprecated by `UPDATE_QUOTA`.
     case mesos::master::Call::SET_QUOTA:
       if (!call.has_set_quota()) {
         return Error("Expecting 'set_quota' to be present");
       }
       return None();
 
+    // TODO(bmahler): Add this to a deprecated call section
+    // at the bottom once deprecated by `UPDATE_QUOTA`.
     case mesos::master::Call::REMOVE_QUOTA:
       if (!call.has_remove_quota()) {
         return Error("Expecting 'remove_quota' to be present");
@@ -500,11 +509,61 @@ Option<Error> validateRoles(const FrameworkInfo& frameworkInfo)
   return None();
 }
 
+
+Option<Error> validateFrameworkId(const FrameworkInfo& frameworkInfo)
+{
+  if (!frameworkInfo.has_id() || frameworkInfo.id().value().empty()) {
+    return None();
+  }
+
+  return common::validation::validateID(frameworkInfo.id().value());
+}
+
+
+Option<Error> validateOfferFilters(const FrameworkInfo& frameworkInfo)
+{
+  // Use `auto` in place of `protobuf::MapPair<string, Value::Scalar>`
+  // below since `foreach` is a macro and cannot contain angle brackets.
+  foreach (auto&& filter, frameworkInfo.offer_filters()) {
+    const OfferFilters& offerFilters = filter.second;
+
+    Option<Error> error =
+      common::validation::validateOfferFilters(offerFilters);
+
+    if (error.isSome()) {
+      return error;
+    }
+  }
+
+  return None();
+}
+
 } // namespace internal {
 
+
 Option<Error> validate(const mesos::FrameworkInfo& frameworkInfo)
 {
-  return internal::validateRoles(frameworkInfo);
+  // TODO(jay_guo): This currently only validates the role(s),
+  // framework ID and offer filters, validate more fields!
+  Option<Error> error = internal::validateRoles(frameworkInfo);
+
+  if (error.isSome()) {
+    return error;
+  }
+
+  error = internal::validateFrameworkId(frameworkInfo);
+
+  if (error.isSome()) {
+    return error;
+  }
+
+  error = internal::validateOfferFilters(frameworkInfo);
+
+  if (error.isSome()) {
+    return error;
+  }
+
+  return None();
 }
 
 } // namespace framework {
@@ -639,12 +698,6 @@ Option<Error> validate(
         return Error("Expecting 'agent_id' to be present");
       }
 
-      // TODO(gkleiman): Revisit this once agent supports sending status
-      // updates for operations affecting default resources (MESOS-8194).
-      if (!acknowledge.has_resource_provider_id()) {
-        return Error("Expecting 'resource_provider_id' to be present");
-      }
-
       return None();
     }
 
@@ -845,6 +898,23 @@ Option<Error> validateAllocatedToSingleRole(const Resources& resources)
   return None();
 }
 
+
+bool detectOverlappingSetAndRangeResources(
+    const vector<Resources>& resources)
+{
+  // If the sum of quantities of each `Resources` is not equal to the quantities
+  // of the sum of `Resources`, then there is some overlap in ranges or sets.
+  ResourceQuantities totalQuantities;
+  Resources totalResources;
+
+  foreach (const Resources& resources_, resources) {
+    totalQuantities += ResourceQuantities::fromResources(resources_);
+    totalResources += resources_;
+  }
+
+  return totalQuantities != ResourceQuantities::fromResources(totalResources);
+}
+
 namespace internal {
 
 // Validates that all the given resources are from the same resource
@@ -927,6 +997,19 @@ Option<Error> validateExecutorID(const ExecutorInfo& executor)
 
 Option<Error> validateType(const ExecutorInfo& executor)
 {
+  if (executor.has_container()) {
+    Option<Error> unionError =
+      protobuf::validateProtobufUnion(executor.container());
+    if (unionError.isSome()) {
+      LOG(WARNING)
+        << "Executor " << executor.executor_id()
+        // NOTE: Validation of FrameworkID is done after this validation.
+        << " of framework '"
+        << (executor.has_framework_id() ? executor.framework_id().value() : "")
+        << "' has an invalid protobuf union: "
+        << unionError.get();
+    }
+  }
   switch (executor.type()) {
     case ExecutorInfo::DEFAULT:
       if (executor.has_command()) {
@@ -1298,6 +1381,18 @@ Option<Error> validateTaskAndExecutorResources(const TaskInfo& task)
         "Task and its executor use invalid resources: " + error->message);
   }
 
+  // We perform the check for `has_executor()` again here because we needed to
+  // validate the total resources before checking for overlapping ranges/sets.
+  if (task.has_executor()) {
+    if (resource::detectOverlappingSetAndRangeResources({
+            task.resources(),
+            task.executor().resources()})) {
+      return Error("There are overlapping resources in the task resources " +
+                   stringify(task.resources()) + " and executor resources " +
+                   stringify(task.executor().resources()));
+    }
+  }
+
   error = resource::validateUniquePersistenceID(total);
   if (error.isSome()) {
     return Error("Task and its executor use duplicate persistence ID: " +
@@ -1564,7 +1659,10 @@ Option<Error> validateTaskGroupAndExecutorResources(
     const ExecutorInfo& executor)
 {
   Resources total = executor.resources();
+
+  vector<Resources> taskResources;
   foreach (const TaskInfo& task, taskGroup.tasks()) {
+    taskResources.push_back(task.resources());
     total += task.resources();
   }
 
@@ -1580,6 +1678,16 @@ Option<Error> validateTaskGroupAndExecutorResources(
                  " resources: " + error->message);
   }
 
+  vector<Resources> allResources(taskResources);
+  allResources.push_back(executor.resources());
+
+  if (resource::detectOverlappingSetAndRangeResources(allResources)) {
+    return Error("There are overlapping resources in the task group's task"
+                 " resources " + stringify(taskResources) +
+                 " and/or executor resources " +
+                 stringify(executor.resources()));
+  }
+
   return None();
 }
 
@@ -2520,6 +2628,12 @@ Option<Error> validate(const Offer::Operation::CreateDisk& createDisk)
     return Error("'target_type' is neither MOUNT or BLOCK");
   }
 
+  if (source.disk().source().has_profile() == createDisk.has_target_profile()) {
+    return createDisk.has_target_profile()
+      ? Error("'target_profile' must not be set when 'source' has a profile")
+      : Error("'target_profile' must be set when 'source' has no profile");
+  }
+
   return None();
 }
 
@@ -2538,8 +2652,20 @@ Option<Error> validate(const Offer::Operation::DestroyDisk& destroyDisk)
   }
 
   if (!Resources::isDisk(source, Resource::DiskInfo::Source::MOUNT) &&
-      !Resources::isDisk(source, Resource::DiskInfo::Source::BLOCK)) {
-    return Error("'source' is neither a MOUNT or BLOCK disk resource");
+      !Resources::isDisk(source, Resource::DiskInfo::Source::BLOCK) &&
+      !Resources::isDisk(source, Resource::DiskInfo::Source::RAW)) {
+    return Error("'source' is neither a MOUNT, BLOCK or RAW disk resource");
+  }
+
+  if (!source.disk().source().has_id()) {
+    return Error("'source' is not backed by a CSI volume");
+  }
+
+  if (Resources::isPersistentVolume(source)) {
+    return Error(
+        "A disk resource containing a persistent volume " + stringify(source) +
+        " cannot be destroyed directly. Please destroy the persistent volume"
+        " first then destroy the disk resource");
   }
 
   return None();
