@@ -33,8 +33,9 @@
 #include <process/id.hpp>
 #include <process/process.hpp>
 
-#include <process/metrics/pull_gauge.hpp>
+#include <process/metrics/counter.hpp>
 #include <process/metrics/metrics.hpp>
+#include <process/metrics/pull_gauge.hpp>
 
 #include <stout/hashmap.hpp>
 #include <stout/nothing.hpp>
@@ -63,6 +64,7 @@ using mesos::resource_provider::AdmitResourceProvider;
 using mesos::resource_provider::Call;
 using mesos::resource_provider::Event;
 using mesos::resource_provider::Registrar;
+using mesos::resource_provider::RemoveResourceProvider;
 
 using mesos::resource_provider::registry::Registry;
 
@@ -91,6 +93,7 @@ using process::http::UnsupportedMediaType;
 
 using process::http::authentication::Principal;
 
+using process::metrics::Counter;
 using process::metrics::PullGauge;
 
 namespace mesos {
@@ -170,7 +173,7 @@ struct ResourceProvider
 
   ResourceProviderInfo info;
   HttpConnection http;
-  hashmap<UUID, Owned<Promise<Nothing>>> publishes;
+  hashmap<id::UUID, Owned<Promise<Nothing>>> publishes;
 };
 
 
@@ -191,6 +194,9 @@ public:
 
   void reconcileOperations(const ReconcileOperationsMessage& message);
 
+  Future<Nothing> removeResourceProvider(
+      const ResourceProviderID& resourceProviderId);
+
   Future<Nothing> publishResources(const Resources& resources);
 
   Queue<ResourceProviderMessage> messages;
@@ -208,6 +214,12 @@ private:
       ResourceProvider* resourceProvider,
       const Call::UpdateOperationStatus& update);
 
+  ResourceProviderMessage createUpdateOperationStatus(
+      const OperationStatus& operationStatus,
+      const Option<id::UUID>& operationUuid,
+      const Option<FrameworkID>& frameworkId,
+      const Option<OperationStatus>& latestStatus);
+
   void updateState(
       ResourceProvider* resourceProvider,
       const Call::UpdateState& update);
@@ -240,6 +252,8 @@ private:
     ~Metrics();
 
     PullGauge subscribed;
+    Counter subscriptions;
+    Counter disconnections;
   };
 
   Owned<Registrar> registrar;
@@ -451,7 +465,9 @@ void ResourceProviderManagerProcess::applyOperation(
     const ApplyOperationMessage& message)
 {
   const Offer::Operation& operation = message.operation_info();
-  const FrameworkID& frameworkId = message.framework_id();
+  const Option<FrameworkID> frameworkId = message.has_framework_id()
+    ? message.framework_id()
+    : Option<FrameworkID>::none();
   const UUID& operationUUID = message.operation_uuid();
 
   Result<ResourceProviderID> resourceProviderId =
@@ -460,7 +476,11 @@ void ResourceProviderManagerProcess::applyOperation(
   if (!resourceProviderId.isSome()) {
     LOG(ERROR) << "Failed to get the resource provider ID of operation "
                << "'" << operation.id() << "' (uuid: " << operationUUID
-               << ") from framework " << frameworkId << ": "
+               << ") from "
+               << (frameworkId.isSome()
+                     ? "framework " + stringify(frameworkId.get())
+                     : "an operator API call")
+               << ": "
                << (resourceProviderId.isError() ? resourceProviderId.error()
                                                 : "Not found");
     return;
@@ -468,7 +488,10 @@ void ResourceProviderManagerProcess::applyOperation(
 
   if (!resourceProviders.subscribed.contains(resourceProviderId.get())) {
     LOG(WARNING) << "Dropping operation '" << operation.id() << "' (uuid: "
-                 << operationUUID << ") from framework " << frameworkId
+                 << operationUUID << ") from "
+                 << (frameworkId.isSome()
+                       ? "framework " + stringify(frameworkId.get())
+                       : "an operator API call")
                  << " because resource provider " << resourceProviderId.get()
                  << " is not subscribed";
     return;
@@ -488,8 +511,10 @@ void ResourceProviderManagerProcess::applyOperation(
 
   Event event;
   event.set_type(Event::APPLY_OPERATION);
-  event.mutable_apply_operation()
-    ->mutable_framework_id()->CopyFrom(frameworkId);
+  if (frameworkId.isSome()) {
+    event.mutable_apply_operation()
+      ->mutable_framework_id()->CopyFrom(frameworkId.get());
+  }
   event.mutable_apply_operation()->mutable_info()->CopyFrom(operation);
   event.mutable_apply_operation()
     ->mutable_operation_uuid()->CopyFrom(message.operation_uuid());
@@ -499,9 +524,12 @@ void ResourceProviderManagerProcess::applyOperation(
 
   if (!resourceProvider->http.send(event)) {
     LOG(WARNING) << "Failed to send operation '" << operation.id() << "' "
-                 << "(uuid: " << operationUUID << ") from framework "
-                 << frameworkId << " to resource provider "
-                 << resourceProviderId.get() << ": connection closed";
+                 << "(uuid: " << operationUUID << ") from "
+                 << (frameworkId.isSome()
+                       ? "framework " + stringify(frameworkId.get())
+                       : "an operator API call")
+                 << " to resource provider " << resourceProviderId.get()
+                 << ": connection closed";
   }
 }
 
@@ -565,18 +593,64 @@ void ResourceProviderManagerProcess::reconcileOperations(
       }
   };
 
-  // Construct events for individual resource providers.
+  // If the framework ID is set, then this reconciliation request originated
+  // from the framework; otherwise, it originated from the master. If it's from
+  // the framework, then we return an operation state based on whether or not
+  // the resource provider is known. If it's from the master, then we forward it
+  // to the resource provider to determine whether or not this operation was
+  // dropped en route to the agent.
+  Option<FrameworkID> frameworkId = message.has_framework_id()
+    ? message.framework_id()
+    : Option<FrameworkID>::none();
+
+  // Construct events for individual resource providers, or send appropriate
+  // operation status updates if resource providers are not subscribed.
   foreach (
       const ReconcileOperationsMessage::Operation& operation,
       message.operations()) {
     if (operation.has_resource_provider_id()) {
       if (!resourceProviders.subscribed.contains(
               operation.resource_provider_id())) {
-        LOG(WARNING) << "Dropping operation reconciliation message with"
-                     << " operation_uuid " << operation.operation_uuid()
-                     << " because resource provider "
-                     << operation.resource_provider_id()
-                     << " is not subscribed";
+        OperationState state;
+        if (resourceProviders.known.contains(
+                operation.resource_provider_id())) {
+          // The resource provider is not currently subscribed, but since it is
+          // known, we expect it to resubscribe. We send OPERATION_RECOVERING in
+          // this case.
+          state = OPERATION_RECOVERING;
+        } else {
+          // The resource provider is not known; we return OPERATION_UNKNOWN in
+          // this case.
+          state = OPERATION_UNKNOWN;
+        }
+
+        Option<id::UUID> uuid;
+        if (operation.has_operation_uuid()) {
+          Try<id::UUID> uuid_ =
+            id::UUID::fromBytes(operation.operation_uuid().value());
+          CHECK_SOME(uuid_);
+          uuid = uuid_.get();
+        }
+
+        // NOTE: the `SlaveID` is left blank in the operation status below
+        // because the RP manager does not have access to this information. The
+        // agent will inject the ID before forwarding to the master.
+        messages.put(
+            createUpdateOperationStatus(
+                protobuf::createOperationStatus(
+                    state,
+                    operation.has_operation_id()
+                      ? operation.operation_id()
+                      : Option<OperationID>::none(),
+                    None(),
+                    None(),
+                    uuid,
+                    None(),
+                    operation.resource_provider_id()),
+                uuid,
+                frameworkId,
+                None()));
+
         continue;
       }
 
@@ -601,6 +675,66 @@ void ResourceProviderManagerProcess::reconcileOperations(
 }
 
 
+Future<Nothing> ResourceProviderManagerProcess::removeResourceProvider(
+    const ResourceProviderID& resourceProviderId)
+{
+  LOG(INFO) << "Removing resource provider " << resourceProviderId;
+
+  Future<bool> removeResourceProvider = registrar
+    ->apply(Owned<mesos::resource_provider::Registrar::Operation>(
+        new RemoveResourceProvider(resourceProviderId)));
+
+  removeResourceProvider.onAny(
+      [resourceProviderId](const Future<bool>& removeResourceProvider) {
+        if (!removeResourceProvider.isReady()) {
+          LOG(ERROR) << "Not removing resource provider " << resourceProviderId
+                     << " as registry update did not succeed: "
+                     << removeResourceProvider;
+        }
+      });
+
+  return removeResourceProvider
+    .then(defer(
+        self(),
+        [this, resourceProviderId](const Future<bool>& removeResourceProvider) {
+          // TODO(bbannier): We should also on a best effort basis send a
+          // `TEARDOWN` event when a removed resource provider attempts to
+          // resubscribe. This can happen e.g., if the event gets lost on its
+          // way to the resource provider, or if the resource provider is not
+          // connected.
+          if (resourceProviders.subscribed.contains(resourceProviderId)) {
+            const Owned<ResourceProvider>& resourceProvider =
+              resourceProviders.subscribed.at(resourceProviderId);
+            Event event;
+            event.set_type(Event::TEARDOWN);
+
+            if (!resourceProvider->http.send(event)) {
+              LOG(WARNING)
+                << "Failed to send TEARDOWN event to resource provider "
+                << resourceProviderId << ": connection closed";
+            }
+          } else {
+            LOG(WARNING)
+              << "Failed to send TEARDOWN event to resource provider "
+              << resourceProviderId << ": resource provider not subscribed";
+          }
+
+          resourceProviders.known.erase(resourceProviderId);
+          resourceProviders.subscribed.erase(resourceProviderId);
+
+          ResourceProviderMessage::Remove remove{resourceProviderId};
+
+          ResourceProviderMessage message;
+          message.type = ResourceProviderMessage::Type::REMOVE;
+          message.remove = std::move(remove);
+
+          messages.put(std::move(message));
+
+          return Nothing();
+        }));
+}
+
+
 Future<Nothing> ResourceProviderManagerProcess::publishResources(
     const Resources& resources)
 {
@@ -633,12 +767,13 @@ Future<Nothing> ResourceProviderManagerProcess::publishResources(
   foreachpair (const ResourceProviderID& resourceProviderId,
                const Resources& resources,
                providedResources) {
-    UUID uuid = protobuf::createUUID();
+    id::UUID uuid = id::UUID::random();
 
     Event event;
     event.set_type(Event::PUBLISH_RESOURCES);
-    event.mutable_publish_resources()->mutable_uuid()->CopyFrom(uuid);
-    event.mutable_publish_resources()->mutable_resources()->CopyFrom(resources);
+    *event.mutable_publish_resources()->mutable_uuid() =
+      protobuf::createUUID(uuid);
+    *event.mutable_publish_resources()->mutable_resources() = resources;
 
     ResourceProvider* resourceProvider =
       resourceProviders.subscribed.at(resourceProviderId).get();
@@ -755,17 +890,6 @@ void ResourceProviderManagerProcess::_subscribe(
 
   const ResourceProviderID& resourceProviderId = resourceProvider->info.id();
 
-  Event event;
-  event.set_type(Event::SUBSCRIBED);
-  event.mutable_subscribed()->mutable_provider_id()
-    ->CopyFrom(resourceProviderId);
-
-  if (!resourceProvider->http.send(event)) {
-    LOG(WARNING) << "Failed to send SUBSCRIBED event to resource provider "
-                 << resourceProviderId << ": connection closed";
-    return;
-  }
-
   resourceProvider->http.closed()
     .onAny(defer(self(), [=](const Future<Nothing>& future) {
       // Iff the remote side closes the HTTP connection, the future will be
@@ -787,6 +911,8 @@ void ResourceProviderManagerProcess::_subscribe(
       message.disconnect = std::move(disconnect);
 
       messages.put(std::move(message));
+
+      ++metrics.disconnections;
     }));
 
   if (!resourceProviders.known.contains(resourceProviderId)) {
@@ -798,10 +924,37 @@ void ResourceProviderManagerProcess::_subscribe(
         std::move(resourceProvider_));
   }
 
+  ResourceProviderMessage::Subscribe subscribe{resourceProvider->info};
+
+  ResourceProviderMessage message;
+  message.type = ResourceProviderMessage::Type::SUBSCRIBE;
+  message.subscribe = std::move(subscribe);
+
+  // Store a pointer to the resource provider so that we can access it
+  // even after moving it into the list of subscribed providers below.
+  ResourceProvider* _resourceProvider = resourceProvider.get();
+
   // TODO(jieyu): Start heartbeat for the resource provider.
   resourceProviders.subscribed.put(
       resourceProviderId,
       std::move(resourceProvider));
+
+  messages.put(std::move(message));
+  ++metrics.subscriptions;
+
+  // Only send a `SUBSCRIBED` event after we have updated our internal
+  // bookkeeping so that e.g., metrics always reflect correct
+  // subscription state.
+  Event event;
+  event.set_type(Event::SUBSCRIBED);
+  event.mutable_subscribed()->mutable_provider_id()
+    ->CopyFrom(resourceProviderId);
+
+  if (!_resourceProvider->http.send(event)) {
+    LOG(WARNING) << "Failed to send SUBSCRIBED event to resource provider "
+                 << resourceProviderId << ": connection closed";
+    return;
+  }
 }
 
 
@@ -809,21 +962,62 @@ void ResourceProviderManagerProcess::updateOperationStatus(
     ResourceProvider* resourceProvider,
     const Call::UpdateOperationStatus& update)
 {
-  ResourceProviderMessage::UpdateOperationStatus body;
-  body.update.mutable_status()->CopyFrom(update.status());
-  body.update.mutable_operation_uuid()->CopyFrom(update.operation_uuid());
+  CHECK_EQ(update.status().resource_provider_id(), resourceProvider->info.id());
+
+  Option<FrameworkID> frameworkId;
   if (update.has_framework_id()) {
-    body.update.mutable_framework_id()->CopyFrom(update.framework_id());
+    frameworkId = update.framework_id();
   }
+
+  Option<OperationStatus> latestStatus;
   if (update.has_latest_status()) {
-    body.update.mutable_latest_status()->CopyFrom(update.latest_status());
+    CHECK_EQ(
+        update.latest_status().resource_provider_id(),
+        resourceProvider->info.id());
+
+    latestStatus = update.latest_status();
+  }
+
+  Try<id::UUID> uuid = id::UUID::fromBytes(update.operation_uuid().value());
+  CHECK_SOME(uuid);
+
+  messages.put(
+      createUpdateOperationStatus(
+          update.status(),
+          uuid.get(),
+          frameworkId,
+          latestStatus));
+}
+
+
+ResourceProviderMessage
+ResourceProviderManagerProcess::createUpdateOperationStatus(
+    const OperationStatus& operationStatus,
+    const Option<id::UUID>& operationUuid,
+    const Option<FrameworkID>& frameworkId,
+    const Option<OperationStatus>& latestStatus)
+{
+  ResourceProviderMessage::UpdateOperationStatus body;
+  body.update.mutable_status()->CopyFrom(operationStatus);
+
+  if (operationUuid.isSome()) {
+    body.update.mutable_operation_uuid()->CopyFrom(
+        protobuf::createUUID(operationUuid.get()));
+  }
+
+  if (frameworkId.isSome()) {
+    body.update.mutable_framework_id()->CopyFrom(frameworkId.get());
+  }
+
+  if (latestStatus.isSome()) {
+    body.update.mutable_latest_status()->CopyFrom(latestStatus.get());
   }
 
   ResourceProviderMessage message;
   message.type = ResourceProviderMessage::Type::UPDATE_OPERATION_STATUS;
   message.updateOperationStatus = std::move(body);
 
-  messages.put(std::move(message));
+  return message;
 }
 
 
@@ -848,7 +1042,7 @@ void ResourceProviderManagerProcess::updateState(
     << resourceProvider->info.id();
 
   ResourceProviderMessage::UpdateState updateState{
-      resourceProvider->info,
+      resourceProvider->info.id(),
       update.resource_version_uuid(),
       update.resources(),
       std::move(operations)};
@@ -865,32 +1059,39 @@ void ResourceProviderManagerProcess::updatePublishResourcesStatus(
     ResourceProvider* resourceProvider,
     const Call::UpdatePublishResourcesStatus& update)
 {
-  const UUID& uuid = update.uuid();
+  Try<id::UUID> uuid = id::UUID::fromBytes(update.uuid().value());
+
+  if (uuid.isError()) {
+    LOG(ERROR)
+      << "Ignoring UpdatePublishResourcesStatus from resource provider "
+      << resourceProvider->info.id() << ": " << uuid.error();
+
+    return;
+  }
+
+  if (!resourceProvider->publishes.contains(uuid.get())) {
+    LOG(ERROR)
+      << "Ignoring UpdatePublishResourcesStatus from resource provider "
+      << resourceProvider->info.id() << ": Unknown UUID " << uuid.get();
 
-  if (!resourceProvider->publishes.contains(uuid)) {
-    LOG(ERROR) << "Ignoring UpdatePublishResourcesStatus from resource"
-               << " provider " << resourceProvider->info.id()
-               << " because UUID " << uuid << " is unknown";
     return;
   }
 
   LOG(INFO)
     << "Received UPDATE_PUBLISH_RESOURCES_STATUS call for PUBLISH_RESOURCES"
-    << " event " << uuid << " with " << update.status()
+    << " event " << uuid.get() << " with " << update.status()
     << " status from resource provider " << resourceProvider->info.id();
 
   if (update.status() == Call::UpdatePublishResourcesStatus::OK) {
-    resourceProvider->publishes.at(uuid)->set(Nothing());
+    resourceProvider->publishes.at(uuid.get())->set(Nothing());
   } else {
     // TODO(jieyu): Consider to include an error message in
     // 'UpdatePublishResourcesStatus' and surface that to the caller.
-    resourceProvider->publishes.at(uuid)->fail(
-        "Failed to publish resources for resource provider " +
-        stringify(resourceProvider->info.id()) + ": Received " +
-        stringify(update.status()) + " status");
+    resourceProvider->publishes.at(uuid.get())->fail(
+        "Received " + stringify(update.status()) + " status");
   }
 
-  resourceProvider->publishes.erase(uuid);
+  resourceProvider->publishes.erase(uuid.get());
 }
 
 
@@ -911,16 +1112,22 @@ double ResourceProviderManagerProcess::gaugeSubscribed()
 ResourceProviderManagerProcess::Metrics::Metrics(
     const ResourceProviderManagerProcess& manager)
   : subscribed(
-      "resource_provider_manager/subscribed",
-      defer(manager, &ResourceProviderManagerProcess::gaugeSubscribed))
+        "resource_provider_manager/subscribed",
+        defer(manager, &ResourceProviderManagerProcess::gaugeSubscribed)),
+    subscriptions("resource_provider_manager/events/subscribe"),
+    disconnections("resource_provider_manager/events/disconnect")
 {
   process::metrics::add(subscribed);
+  process::metrics::add(subscriptions);
+  process::metrics::add(disconnections);
 }
 
 
 ResourceProviderManagerProcess::Metrics::~Metrics()
 {
   process::metrics::remove(subscribed);
+  process::metrics::remove(subscriptions);
+  process::metrics::remove(disconnections);
 }
 
 
@@ -980,6 +1187,15 @@ void ResourceProviderManager::reconcileOperations(
 }
 
 
+Future<Nothing> ResourceProviderManager::removeResourceProvider(
+    const ResourceProviderID& resourceProviderId) const
+{
+  return dispatch(
+      process.get(),
+      &ResourceProviderManagerProcess::removeResourceProvider,
+      resourceProviderId);
+}
+
 Future<Nothing> ResourceProviderManager::publishResources(
     const Resources& resources)
 {
