@@ -340,10 +340,13 @@ Future<Nothing> CgroupsIsolatorProcess::___recover(
   // TODO(haosdent): Use foreachkey once MESOS-5037 is resolved.
   foreach (const string& hierarchy, subsystems.keys()) {
     if (!cgroups::exists(hierarchy, cgroup)) {
-      // This may occur if the executor has exited and the isolator
-      // has destroyed the cgroup but the agent dies before noticing
-      // this. This will be detected when the containerizer tries to
-      // monitor the executor's pid.
+      // This may occur in two cases:
+      // 1. If the executor has exited and the isolator has destroyed
+      //    the cgroup but the agent dies before noticing this. This
+      //    will be detected when the containerizer tries to monitor
+      //    the executor's pid.
+      // 2. After the agent recovery/upgrade, new cgroup subsystems
+      //    are added to the agent cgroup isolation configuration.
       LOG(WARNING) << "Couldn't find the cgroup '" << cgroup << "' "
                    << "in hierarchy '" << hierarchy << "' "
                    << "for container " << containerId;
@@ -435,7 +438,8 @@ Future<Option<ContainerLaunchInfo>> CgroupsIsolatorProcess::prepare(
 
     Try<Nothing> create = cgroups::create(
         hierarchy,
-        infos[containerId]->cgroup);
+        infos[containerId]->cgroup,
+        true);
 
     if (create.isError()) {
       return Failure(
@@ -581,15 +585,13 @@ Future<Option<ContainerLaunchInfo>> CgroupsIsolatorProcess::__prepare(
   // cgroups mount, e.g.:
   //   mount --bind /sys/fs/cgroup/memory/mesos/<containerId> /sys/fs/cgroup/memory // NOLINT(whitespace/line_length)
   foreach (const string& hierarchy, subsystems.keys()) {
-    ContainerMountInfo* mount = launchInfo.add_mounts();
-    mount->set_source(path::join(hierarchy, infos[rootContainerId]->cgroup));
-
-    mount->set_target(path::join(
-        containerConfig.rootfs(),
-        "/sys/fs/cgroup",
-        Path(hierarchy).basename()));
-
-    mount->set_flags(MS_BIND | MS_REC);
+    *launchInfo.add_mounts() = protobuf::slave::createContainerMount(
+        path::join(hierarchy, infos[rootContainerId]->cgroup),
+        path::join(
+            containerConfig.rootfs(),
+            "/sys/fs/cgroup",
+            Path(hierarchy).basename()),
+        MS_BIND | MS_REC);
   }
 
   // Linux launcher will create freezer and systemd cgroups for the container,
@@ -603,38 +605,34 @@ Future<Option<ContainerLaunchInfo>> CgroupsIsolatorProcess::__prepare(
           hierarchy.error());
     }
 
-    ContainerMountInfo* mount = launchInfo.add_mounts();
-    mount->set_source(path::join(
-        hierarchy.get(),
-        flags.cgroups_root,
-        containerizer::paths::buildPath(
-            containerId,
-            CGROUP_SEPARATOR,
-            containerizer::paths::JOIN)));
-
-    mount->set_target(path::join(
-        containerConfig.rootfs(),
-        "/sys/fs/cgroup",
-        "freezer"));
-
-    mount->set_flags(MS_BIND | MS_REC);
+    *launchInfo.add_mounts() = protobuf::slave::createContainerMount(
+        path::join(
+            hierarchy.get(),
+            flags.cgroups_root,
+            containerizer::paths::buildPath(
+                containerId,
+                CGROUP_SEPARATOR,
+                containerizer::paths::JOIN)),
+        path::join(
+            containerConfig.rootfs(),
+            "/sys/fs/cgroup",
+            "freezer"),
+        MS_BIND | MS_REC);
 
     if (systemd::enabled()) {
-      mount = launchInfo.add_mounts();
-      mount->set_source(path::join(
-          systemd::hierarchy(),
-          flags.cgroups_root,
-          containerizer::paths::buildPath(
-              containerId,
-              CGROUP_SEPARATOR,
-              containerizer::paths::JOIN)));
-
-      mount->set_target(path::join(
-          containerConfig.rootfs(),
-          "/sys/fs/cgroup",
-          "systemd"));
-
-      mount->set_flags(MS_BIND | MS_REC);
+      *launchInfo.add_mounts() = protobuf::slave::createContainerMount(
+          path::join(
+              systemd::hierarchy(),
+              flags.cgroups_root,
+              containerizer::paths::buildPath(
+                  containerId,
+                  CGROUP_SEPARATOR,
+                  containerizer::paths::JOIN)),
+          path::join(
+              containerConfig.rootfs(),
+              "/sys/fs/cgroup",
+              "systemd"),
+          MS_BIND | MS_REC);
     }
   }
 
@@ -677,18 +675,33 @@ Future<Nothing> CgroupsIsolatorProcess::isolate(
     return Failure("Failed to isolate the container: Unknown root container");
   }
 
+  const string& cgroup = infos[rootContainerId]->cgroup;
+
   // TODO(haosdent): Use foreachkey once MESOS-5037 is resolved.
   foreach (const string& hierarchy, subsystems.keys()) {
+    // If new cgroup subsystems are added after the agent
+    // upgrade, the newly added cgroup subsystems do not
+    // exist on old container's cgroup hierarchy. So skip
+    // assigning the pid to this cgroup subsystem.
+    if (containerId.has_parent() && !cgroups::exists(hierarchy, cgroup)) {
+      LOG(INFO) << "Skipping assigning pid " << stringify(pid)
+                << " to cgroup at '" << path::join(hierarchy, cgroup)
+                << "' for container " << containerId
+                << " because its parent container " << containerId.parent()
+                << " does not have this cgroup hierarchy";
+      continue;
+    }
+
     Try<Nothing> assign = cgroups::assign(
         hierarchy,
-        infos[rootContainerId]->cgroup,
+        cgroup,
         pid);
 
     if (assign.isError()) {
       string message =
-        "Failed to assign pid " + stringify(pid) + " to cgroup at "
-        "'" + path::join(hierarchy, infos[rootContainerId]->cgroup) + "'"
-        ": " + assign.error();
+        "Failed to assign container " + stringify(containerId) +
+        " pid " + stringify(pid) + " to cgroup at '" +
+        path::join(hierarchy, cgroup) + "': " + assign.error();
 
       LOG(ERROR) << message;
 
