@@ -24,6 +24,8 @@
 #include <stout/os/realpath.hpp>
 #include <stout/os/which.hpp>
 
+#include "common/protobuf_utils.hpp"
+
 #include "linux/ns.hpp"
 
 #include "slave/flags.hpp"
@@ -266,8 +268,16 @@ Try<Nothing> DockerVolumeIsolatorProcess::_recover(
     paths::getVolumesPath(rootDir, containerId.value());
 
   if (!os::exists(volumesPath)) {
-    VLOG(1) << "The docker volumes checkpointed at '" << volumesPath
-            << "' for container " << containerId << " does not exist";
+    // This could happen if the slave died after creating the container
+    // directory but before it checkpointed anything in it.
+    LOG(WARNING) << "The docker volumes checkpointed at '" << volumesPath
+                 << "' for container " << containerId << " does not exist";
+
+    // Construct an info object with empty docker volumes since no docker
+    // volumes are mounted yet for this container, and this container will
+    // be cleaned up by containerizer (as known orphan container) or by
+    // `recover` (as unknown orphan container).
+    infos.put(containerId, Owned<Info>(new Info(hashset<DockerVolume>())));
 
     return Nothing();
   }
@@ -279,6 +289,21 @@ Try<Nothing> DockerVolumeIsolatorProcess::_recover(
         volumesPath + "': " + read.error());
   }
 
+  if (read->empty()) {
+    // This could happen if the slave is hard rebooted after the file is
+    // created but before the data is synced on disk.
+    LOG(WARNING) << "The docker volumes checkpointed at '" << volumesPath
+                 << "' for container " << containerId << " is empty";
+
+    // Construct an info object with empty docker volumes since no docker
+    // volumes are mounted yet for this container, and this container will
+    // be cleaned up by containerizer (as known orphan container) or by
+    // `recover` (as unknown orphan container).
+    infos.put(containerId, Owned<Info>(new Info(hashset<DockerVolume>())));
+
+    return Nothing();
+  }
+
   Try<JSON::Object> json = JSON::parse<JSON::Object>(read.get());
   if (json.isError()) {
     return Error("JSON parse failed: " + json.error());
@@ -555,17 +580,10 @@ Future<Option<ContainerLaunchInfo>> DockerVolumeIsolatorProcess::_prepare(
     LOG(INFO) << "Mounting docker volume mount point '" << source
               << "' to '" << target << "' for container " << containerId;
 
-    ContainerMountInfo* mount = launchInfo.add_mounts();
-    mount->set_source(source);
-    mount->set_target(target);
-    mount->set_flags(MS_BIND | MS_REC);
-
-    // If the mount needs to be read-only, do a remount.
-    if (volumeMode == Volume::RO) {
-      mount = launchInfo.add_mounts();
-      mount->set_target(target);
-      mount->set_flags(MS_BIND | MS_RDONLY | MS_REMOUNT);
-    }
+    *launchInfo.add_mounts() = protobuf::slave::createContainerMount(
+        source,
+        target,
+        MS_BIND | MS_REC | (volumeMode == Volume::RO ? MS_RDONLY : 0));
   }
 
   return launchInfo;
