@@ -27,7 +27,9 @@
 
 #include <stout/check.hpp>
 #include <stout/foreach.hpp>
+#include <stout/fs.hpp>
 #include <stout/os.hpp>
+#include <stout/unreachable.hpp>
 #include <stout/utils.hpp>
 
 #include <stout/os/stat.hpp>
@@ -37,6 +39,8 @@
 #include "slave/paths.hpp"
 
 using std::list;
+using std::make_pair;
+using std::pair;
 using std::string;
 using std::vector;
 
@@ -75,7 +79,35 @@ static Try<IntervalSet<prid_t>> getIntervalSet(
 }
 
 
-static Option<Bytes> getDiskResource(
+static Try<Nothing> isPathSupported(const string& path)
+{
+  if (!xfs::isPathXfs(path)) {
+    return Error("'" + path + "' is not an XFS filesystem");
+  }
+
+  Try<bool> enabled = xfs::isQuotaEnabled(path);
+  if (enabled.isError()) {
+    return Error(
+        "Failed to get quota status for '" + path + "': " + enabled.error());
+  }
+
+  if (!enabled.get()) {
+    return Error(
+        "XFS project quotas are not enabled on '" + path + "'");
+  }
+
+  return Nothing();
+}
+
+
+static bool isMountDisk(const Resource::DiskInfo& info)
+{
+  return info.has_source() &&
+    info.source().type() == Resource::DiskInfo::Source::MOUNT;
+}
+
+
+static Option<Bytes> getSandboxDisk(
     const Resources& resources)
 {
   Option<Bytes> bytes = None();
@@ -85,16 +117,7 @@ static Option<Bytes> getDiskResource(
       continue;
     }
 
-    // TODO(jpeach): Ignore persistent volume resources. The problem here is
-    // that we need to guarantee that we can track the removal of every
-    // directory for which we assign a project ID. Since destruction of
-    // persistent is not visible to the isolator, we don't want to risk
-    // leaking the project ID, or spuriously reusing it.
-    if (Resources::isPersistentVolume(resource)) {
-      continue;
-    }
-
-    if (resource.has_disk() && resource.disk().has_volume()) {
+    if (Resources::isPersistentVolume(resource) || resource.has_disk()) {
       continue;
     }
 
@@ -111,21 +134,32 @@ static Option<Bytes> getDiskResource(
 
 Try<Isolator*> XfsDiskIsolatorProcess::create(const Flags& flags)
 {
-  if (!xfs::isPathXfs(flags.work_dir)) {
-    return Error("'" + flags.work_dir + "' is not an XFS filesystem");
+  Try<Nothing> supported = isPathSupported(flags.work_dir);
+  if (supported.isError()) {
+    return Error(supported.error());
   }
 
-  Try<bool> enabled = xfs::isQuotaEnabled(flags.work_dir);
-  if (enabled.isError()) {
-    return Error(
-        "Failed to get quota status for '" +
-        flags.work_dir + "': " + enabled.error());
-  }
+  // In general, we can't be sure of the filesystem configuration of
+  // PATH disks. We verify prior to use so that we don't later have to
+  // emit a confusing error (like "inappropriate IOCTL") when attempting
+  // to sample the project ID.
+  vector<Resource> resources = CHECK_NOTERROR(Resources::fromString(
+        flags.resources.getOrElse(""), flags.default_role));
 
-  if (!enabled.get()) {
-    return Error(
-        "XFS project quotas are not enabled on '" +
-        flags.work_dir + "'");
+  foreach (const Resource& resource, resources) {
+    if (resource.name() != "disk" || !resource.has_disk()) {
+      continue;
+    }
+
+    const Resource::DiskInfo& diskInfo = resource.disk();
+
+    if (diskInfo.has_source() &&
+        diskInfo.source().type() == Resource::DiskInfo::Source::PATH) {
+      supported = isPathSupported(diskInfo.source().path().root());
+      if (supported.isError()) {
+        return Error(supported.error());
+      }
+    }
   }
 
   Result<uid_t> uid = os::getuid();
@@ -199,7 +233,8 @@ XfsDiskIsolatorProcess::XfsDiskIsolatorProcess(
   // At the beginning, the free project range is the same as the
   // configured project range.
 
-  LOG(INFO) << "Allocating XFS project IDs from the range " << totalProjectIds;
+  LOG(INFO) << "Allocating " << totalProjectIds.size()
+            << " XFS project IDs from the range " << totalProjectIds;
 
   metrics.project_ids_total = totalProjectIds.size();
   metrics.project_ids_free = totalProjectIds.size();
@@ -282,6 +317,53 @@ Future<Nothing> XfsDiskIsolatorProcess::recover(
     }
   }
 
+  Try<list<string>> volumes = paths::getPersistentVolumePaths(workDir);
+
+  if (volumes.isError()) {
+    return Failure(
+        "Failed to scan persistent volume directories: " + volumes.error());
+  }
+
+  // Track any project IDs that we have assigned to persistent volumes. Note
+  // that is is possible for operators to delete persistent volumes while
+  // the agent isn't running. If that happened, the quota record would be
+  // stale, but eventually the project ID would be re-used and the quota
+  // updated correctly.
+  foreach (const string& directory, volumes.get()) {
+    Result<prid_t> projectId = xfs::getProjectId(directory);
+    if (projectId.isError()) {
+      return Failure(projectId.error());
+    }
+
+    if (projectId.isNone()) {
+      continue;
+    }
+
+    // We should never have duplicate projects IDs assigned
+    // to persistent volumes, and since we haven't checked the
+    // recovered container resources yet, we can't already know
+    // about this persistent volume.
+    CHECK(!scheduledProjects.contains(projectId.get()))
+        << "Duplicate project ID " << projectId.get()
+        << " assigned to '" << directory << "'"
+        << " and '" << scheduledProjects.at(projectId.get()).second << "'";
+
+    freeProjectIds -= projectId.get();
+    if (totalProjectIds.contains(projectId.get())) {
+      --metrics.project_ids_free;
+    }
+
+    Try<string> devname = xfs::getDeviceForPath(directory);
+    if (devname.isError()) {
+      LOG(ERROR) << "Unable to schedule project ID " << projectId.get()
+                 << " for reclaimation: " << devname.error();
+      continue;
+    }
+
+    scheduledProjects.put(
+        projectId.get(), make_pair(devname.get(), directory));
+  }
+
   return Nothing();
 }
 
@@ -344,41 +426,27 @@ Future<ContainerLimitation> XfsDiskIsolatorProcess::watch(
 }
 
 
-Future<Nothing> XfsDiskIsolatorProcess::update(
-    const ContainerID& containerId,
-    const Resources& resources)
+static Try<xfs::QuotaInfo> applyProjectQuota(
+    const string& path,
+    prid_t projectId,
+    Bytes limit,
+    xfs::QuotaPolicy quotaPolicy)
 {
-  if (!infos.contains(containerId)) {
-    LOG(INFO) << "Ignoring update for unknown container " << containerId;
-    return Nothing();
-  }
-
-  const Owned<Info>& info = infos[containerId];
-  Option<Bytes> needed = getDiskResource(resources);
-
-  if (needed.isNone()) {
-    // TODO(jpeach) If there's no disk resource attached, we should set the
-    // minimum quota (1 block), since a zero quota would be unconstrained.
-    LOG(WARNING) << "Ignoring quota update with no disk resources";
-    return Nothing();
-  }
-
   switch (quotaPolicy) {
     case xfs::QuotaPolicy::ACCOUNTING: {
-      Try<Nothing> status = xfs::clearProjectQuota(
-          info->directory, info->projectId);
+      Try<Nothing> status = xfs::clearProjectQuota(path, projectId);
 
       if (status.isError()) {
-        return Failure("Failed to clear quota for project " +
-                       stringify(info->projectId) + ": " + status.error());
+        return Error("Failed to clear quota for project " +
+                     stringify(projectId) + ": " + status.error());
       }
 
-      break;
+      return xfs::QuotaInfo();
     }
 
     case xfs::QuotaPolicy::ENFORCING_ACTIVE:
     case xfs::QuotaPolicy::ENFORCING_PASSIVE: {
-      Bytes hardLimit = needed.get();
+      Bytes hardLimit = limit;
 
       // The purpose behind adding to the hard limit is so that the soft
       // limit can be exceeded thereby allowing us to check if the limit
@@ -389,22 +457,129 @@ Future<Nothing> XfsDiskIsolatorProcess::update(
       }
 
       Try<Nothing> status = xfs::setProjectQuota(
-          info->directory, info->projectId, needed.get(), hardLimit);
+          path, projectId, limit, hardLimit);
 
       if (status.isError()) {
-        return Failure("Failed to update quota for project " +
-                       stringify(info->projectId) + ": " + status.error());
+        return Error("Failed to update quota for project " +
+                     stringify(projectId) + ": " + status.error());
       }
 
-      LOG(INFO) << "Set quota on container " << containerId
-                << " for project " << info->projectId
-                << " to " << needed.get() << "/" << hardLimit;
+      return xfs::QuotaInfo{limit, hardLimit, 0};
+    }
+  }
 
-      break;
+  UNREACHABLE();
+}
+
+
+Future<Nothing> XfsDiskIsolatorProcess::update(
+    const ContainerID& containerId,
+    const Resources& resources)
+{
+  if (!infos.contains(containerId)) {
+    LOG(INFO) << "Ignoring update for unknown container " << containerId;
+    return Nothing();
+  }
+
+  const Owned<Info>& info = infos[containerId];
+
+  // First, apply the disk quota to the sandbox.
+  Option<Bytes> sandboxQuota = getSandboxDisk(resources);
+  if (sandboxQuota.isSome()) {
+    foreachpair (
+        const string& directory, Info::PathInfo& pathInfo, info->paths) {
+      if (pathInfo.disk.isNone()) {
+        pathInfo.quota = sandboxQuota.get();
+
+        Try<xfs::QuotaInfo> status = applyProjectQuota(
+            directory, pathInfo.projectId, sandboxQuota.get(), quotaPolicy);
+        if (status.isError()) {
+          return Failure(status.error());
+        }
+
+        LOG(INFO) << "Set quota on container " << containerId
+                  << " for project " << pathInfo.projectId
+                  << " to " << status->softLimit << "/" << status->hardLimit;
+        break;
+      }
+    }
+  }
+
+  // Make sure that we have project IDs assigned to all persistent volumes.
+  foreach (const Resource& resource, resources.persistentVolumes()) {
+    CHECK(resource.disk().has_volume());
+
+    const Bytes size = Megabytes(resource.scalar().value());
+    const string directory = paths::getPersistentVolumePath(workDir, resource);
+
+    // Don't apply project quotas to mount disks, since they are never
+    // subdivided and we can't guarantee that they are XFS filesystems. We
+    // still track the path for the container so that we can generate disk
+    // statistics correctly.
+    if (isMountDisk(resource.disk())) {
+      info->paths.put(directory, Info::PathInfo{size, 0, resource.disk()});
+      continue;
+    }
+
+    Result<prid_t> projectId = xfs::getProjectId(directory);
+    if (projectId.isError()) {
+      return Failure(projectId.error());
+    }
+
+    // If this volume already has a project ID, we must have assigned it
+    // here, or assigned or recovered it during project recovery.
+    if (projectId.isSome()) {
+      CHECK(scheduledProjects.contains(projectId.get()))
+        << "untracked project ID " << projectId.get()
+        << " for volume ID " << resource.disk().persistence().id()
+        << " on " << directory;
+    }
+
+    if (projectId.isNone()) {
+      Try<Nothing> supported = isPathSupported(directory);
+      if (supported.isError()) {
+        return Failure(supported.error());
+      }
+
+      projectId = nextProjectId();
+
+      Try<Nothing> status = xfs::setProjectId(directory, projectId.get());
+      if (status.isError()) {
+        return Failure(
+            "Failed to assign project " + stringify(projectId.get()) + ": " +
+            status.error());
+      }
+
+      LOG(INFO) << "Assigned project " << stringify(projectId.get()) << " to '"
+                << directory << "'";
+    }
+
+    Try<xfs::QuotaInfo> status =
+      applyProjectQuota(directory, projectId.get(), size, quotaPolicy);
+    if (status.isError()) {
+      return Failure(status.error());
+    }
+
+    info->paths.put(
+        directory, Info::PathInfo{size, projectId.get(), resource.disk()});
+
+    LOG(INFO) << "Set quota on volume " << resource.disk().persistence().id()
+              << " for project " << projectId.get()
+              << " to " << status->softLimit << "/" << status->hardLimit;
+
+    if (!scheduledProjects.contains(projectId.get())) {
+      Try<string> devname = xfs::getDeviceForPath(directory);
+      if (devname.isError()) {
+        LOG(ERROR) << "Unable to schedule project " << projectId.get()
+                   << " for reclaimation: " << devname.error();
+        continue;
+      }
+
+      scheduledProjects.put(
+          projectId.get(), make_pair(devname.get(), directory));
     }
   }
 
-  info->quota = needed.get();
   return Nothing();
 }
 
@@ -414,31 +589,51 @@ void XfsDiskIsolatorProcess::check()
   CHECK(quotaPolicy == xfs::QuotaPolicy::ENFORCING_ACTIVE);
 
   foreachpair(const ContainerID& containerId, const Owned<Info>& info, infos) {
-    Result<xfs::QuotaInfo> quotaInfo = xfs::getProjectQuota(
-        info->directory, info->projectId);
+    foreachpair(
+        const string& directory, const Info::PathInfo& pathInfo, info->paths) {
+      Result<xfs::QuotaInfo> quotaInfo = xfs::getProjectQuota(
+          directory, pathInfo.projectId);
 
-    if (quotaInfo.isError()) {
-      LOG(WARNING) << "Failed to check disk usage for container '"
-                   << containerId  << "': " << quotaInfo.error();
+      if (quotaInfo.isError()) {
+        LOG(WARNING)
+          << "Failed to check disk usage for container '"
+          << containerId  << "' in '" << directory << "': "
+          << quotaInfo.error();
 
-      continue;
-    }
+        continue;
+      }
 
-    // If the soft limit is exceeded the container should be killed.
-    if (quotaInfo->used > quotaInfo->softLimit) {
-      Resource resource;
-      resource.set_name("disk");
-      resource.set_type(Value::SCALAR);
-      resource.mutable_scalar()->set_value(
-        quotaInfo->used.bytes() / Bytes::MEGABYTES);
+      // If the soft limit is exceeded the container should be killed.
+      if (quotaInfo->used > quotaInfo->softLimit) {
+        Resource resource;
+        resource.set_name("disk");
+        resource.set_type(Value::SCALAR);
+        resource.mutable_scalar()->set_value(
+          quotaInfo->used.bytes() / Bytes::MEGABYTES);
 
-      info->limitation.set(
-          protobuf::slave::createContainerLimitation(
-              Resources(resource),
-              "Disk usage (" + stringify(quotaInfo->used) +
-              ") exceeds quota (" +
-              stringify(quotaInfo->softLimit) + ")",
-              TaskStatus::REASON_CONTAINER_LIMITATION_DISK));
+        string volumeInfo;
+
+        if (pathInfo.disk.isSome()) {
+          resource.mutable_disk()->CopyFrom(pathInfo.disk.get());
+          volumeInfo =
+            " for volume '" + pathInfo.disk->persistence().id() + "'";
+        }
+
+        LOG(INFO)
+          << "Container " << stringify(containerId)
+          << " disk usage " << stringify(quotaInfo->used)
+          << " exceeded quota " << stringify(quotaInfo->softLimit)
+          << volumeInfo
+          << " in '" << directory << "'";
+
+        info->limitation.set(
+            protobuf::slave::createContainerLimitation(
+                Resources(resource),
+                "Disk usage (" + stringify(quotaInfo->used) +
+                ") exceeded quota (" + stringify(quotaInfo->softLimit) + ")" +
+                volumeInfo,
+                TaskStatus::REASON_CONTAINER_LIMITATION_DISK));
+      }
     }
   }
 }
@@ -452,24 +647,59 @@ Future<ResourceStatistics> XfsDiskIsolatorProcess::usage(
     return ResourceStatistics();
   }
 
-  ResourceStatistics statistics;
   const Owned<Info>& info = infos[containerId];
+  ResourceStatistics statistics;
 
-  Result<xfs::QuotaInfo> quota = xfs::getProjectQuota(
-      info->directory, info->projectId);
+  foreachpair(
+      const string& directory, const Info::PathInfo& pathInfo, info->paths) {
+    DiskStatistics* disk = nullptr;
 
-  if (quota.isError()) {
-    return Failure(quota.error());
-  }
+    if (pathInfo.disk.isSome()) {
+      disk = statistics.add_disk_statistics();
+      disk->mutable_persistence()->CopyFrom(pathInfo.disk->persistence());
 
-  // If we didn't set the quota (ie. we are in ACCOUNTING mode),
-  // the quota limit will be 0. Since we are already tracking
-  // what the quota ought to be in the Info, we just always
-  // use that.
-  statistics.set_disk_limit_bytes(info->quota.bytes());
+      // Root disk volumes don't have a source.
+      if (pathInfo.disk->has_source()) {
+        disk->mutable_source()->CopyFrom(pathInfo.disk->source());
+      }
+    }
 
-  if (quota.isSome()) {
-    statistics.set_disk_used_bytes(quota->used.bytes());
+    // We don't require XFS on MOUNT disks, but we still want the isolator
+    // to publish usage for them to make them visible to operational tooling.
+    if (disk && isMountDisk(pathInfo.disk.get())) {
+      Try<Bytes> used = fs::used(directory);
+      if (used.isError()) {
+        return Failure("Failed to query disk usage for '" + directory +
+                      "': " + used.error());
+      }
+
+      disk->set_used_bytes(used->bytes());
+      continue;
+    }
+
+    Result<xfs::QuotaInfo> quotaInfo =
+      xfs::getProjectQuota(directory, pathInfo.projectId);
+    if (quotaInfo.isError()) {
+      return Failure(quotaInfo.error());
+    }
+
+    if (quotaInfo.isSome()) {
+      if (disk) {
+        disk->set_used_bytes(quotaInfo->used.bytes());
+      } else {
+        statistics.set_disk_used_bytes(quotaInfo->used.bytes());
+      }
+    }
+
+    // If we didn't set the quota (ie. we are in ACCOUNTING mode),
+    // the quota limit will be 0. Since we are already tracking
+    // what the quota ought to be in the Info, we just always
+    // use that.
+    if (disk) {
+      disk->set_limit_bytes(pathInfo.quota.bytes());
+    } else {
+      statistics.set_disk_limit_bytes(pathInfo.quota.bytes());
+    }
   }
 
   return statistics;
@@ -486,14 +716,9 @@ Future<Nothing> XfsDiskIsolatorProcess::cleanup(const ContainerID& containerId)
     return Nothing();
   }
 
-  // Take a copy of the Info we are removing so that we can use it
-  // to construct the Failure message if necessary.
-  const std::string directory = infos[containerId]->directory;
-  const prid_t projectId = infos[containerId]->projectId;
+  const Owned<Info>& info = infos[containerId];
 
-  infos.erase(containerId);
-
-  // Schedule the directory for project ID reclaiming.
+  // Schedule the directory for project ID reclaimation.
   //
   // We don't reclaim project ID here but wait until sandbox GC time.
   // This is because the sandbox can potentially contain symlinks,
@@ -501,24 +726,30 @@ Future<Nothing> XfsDiskIsolatorProcess::cleanup(const ContainerID& containerId)
   // limitations. Such symlinks would then contribute to disk usage
   // of another container if the project ID was reused causing small
   // inaccuracies in accounting.
-  scheduledProjects.put(projectId, directory);
+  //
+  // Fortunately, this behaviour also suffices to reclaim project IDs from
+  // persistent volumes. We can just leave the project quota in place until
+  // we determine that the persistent volume is no longer present.
+  foreachpair (
+      const string& directory, const Info::PathInfo& pathInfo, info->paths) {
+    // If we assigned a project ID to a persistent volume, it might
+    // already be scheduled for reclaimation.
+    if (scheduledProjects.contains(pathInfo.projectId)) {
+      continue;
+    }
 
-  LOG(INFO) << "Removing quota from project " << projectId
-            << " for '" << directory << "'";
+    Try<string> devname = xfs::getDeviceForPath(directory);
+    if (devname.isError()) {
+      LOG(ERROR) << "Unable to schedule project " << pathInfo.projectId
+                 << " for reclaimation: " << devname.error();
+      continue;
+    }
 
-  Try<Nothing> quotaStatus = xfs::clearProjectQuota(
-      directory, projectId);
-
-  // Note that if we failed to clear the quota, we will still eventually
-  // reclaim the project ID. If there is a persistent error will the quota
-  // system, then we would ultimately fail to re-use that project ID since
-  // the quota update would fail.
-  if (quotaStatus.isError()) {
-    LOG(ERROR) << "Failed to clear quota for '"
-               << directory << "': " << quotaStatus.error();
-    return Failure("Failed to cleanup '" + directory + "'");
+    scheduledProjects.put(
+        pathInfo.projectId, make_pair(devname.get(), directory));
   }
 
+  infos.erase(containerId);
   return Nothing();
 }
 
@@ -552,14 +783,28 @@ void XfsDiskIsolatorProcess::returnProjectId(
 
 void XfsDiskIsolatorProcess::reclaimProjectIds()
 {
+  // Note that we need both the directory we assigned the project ID to,
+  // and the device node for the block device hosting the directory. Since
+  // we can only reclaim the project ID if the former doesn't exist, we
+  // need the latter to make the corresponding quota record updates.
+
   foreachpair (
-      prid_t projectId, const string& dir, utils::copy(scheduledProjects)) {
-    if (!os::exists(dir)) {
-      returnProjectId(projectId);
-      scheduledProjects.erase(projectId);
-      LOG(INFO) << "Reclaimed project ID " << projectId
-                << " from '" << dir << "'";
+      prid_t projectId, const auto& dir, utils::copy(scheduledProjects)) {
+    if (os::exists(dir.second)) {
+      continue;
     }
+
+    Try<Nothing> status = xfs::clearProjectQuota(dir.first, projectId);
+    if (status.isError()) {
+      LOG(ERROR) << "Failed to clear quota for '"
+                 << dir.second << "': " << status.error();
+    }
+
+    returnProjectId(projectId);
+    scheduledProjects.erase(projectId);
+
+    LOG(INFO) << "Reclaimed project ID " << projectId
+              << " from '" << dir.second << "'";
   }
 }
 
