@@ -65,6 +65,10 @@
 #include "linux/ns.hpp"
 #endif
 
+#ifdef ENABLE_SECCOMP_ISOLATOR
+#include "linux/seccomp/seccomp.hpp"
+#endif
+
 #ifndef __WINDOWS__
 #include "posix/rlimits.hpp"
 #endif // __WINDOWS__
@@ -74,17 +78,22 @@
 
 using std::cerr;
 using std::endl;
-using std::list;
 using std::set;
 using std::string;
 using std::vector;
 
+using process::Owned;
+
 #ifdef __linux__
 using mesos::internal::capabilities::Capabilities;
 using mesos::internal::capabilities::Capability;
 using mesos::internal::capabilities::ProcessCapabilities;
 #endif // __linux__
 
+#ifdef ENABLE_SECCOMP_ISOLATOR
+using mesos::internal::seccomp::SeccompFilter;
+#endif
+
 using mesos::slave::ContainerLaunchInfo;
 using mesos::slave::ContainerMountInfo;
 
@@ -371,16 +380,13 @@ static Try<Nothing> prepareMounts(const ContainerLaunchInfo& launchInfo)
     // should consider forcing all isolators to use
     // `ContainerMountInfo` for volume mounts.
     if (hasSharedMount) {
-      Result<string> realTargetPath = os::realpath(mount.target());
-      if (!realTargetPath.isSome()) {
-        return Error(
-            "Failed to get the realpath of the mount target '" +
-            mount.target() + "': " +
-            (realTargetPath.isError() ? realTargetPath.error() : "Not found"));
+      Try<fs::MountInfoTable> table = fs::MountInfoTable::read();
+      if (table.isError()) {
+        return Error("Failed to read mount table: " + table.error());
       }
 
       Try<fs::MountInfoTable::Entry> entry =
-        fs::MountInfoTable::findByTarget(realTargetPath.get());
+        table->findByTarget(mount.target());
 
       if (entry.isError()) {
         return Error(
@@ -411,9 +417,14 @@ static Try<Nothing> prepareMounts(const ContainerLaunchInfo& launchInfo)
         }
       }
 
-      Try<Nothing> target = os::stat::isdir(mount.source())
-        ? os::mkdir(mount.target())
-        : os::touch(mount.target());
+      // The mount source could be a file, a directory, or a Linux
+      // pseudo-filesystem. In the last case, the target must be a
+      // directory, so if the source doesn't exist, we default to
+      // mounting on a directory.
+      Try<Nothing> target =
+        (!os::exists(mount.source()) || os::stat::isdir(mount.source()))
+          ? os::mkdir(mount.target())
+          : os::touch(mount.target());
 
       if (target.isError()) {
         return Error(
@@ -434,8 +445,6 @@ static Try<Nothing> prepareMounts(const ContainerLaunchInfo& launchInfo)
           "Failed to mount '" + stringify(JSON::protobuf(mount)) +
           "': " + mnt.error());
     }
-
-    cerr << "Prepared mount '" << JSON::protobuf(mount) << "'" << endl;
   }
 #endif // __linux__
 
@@ -463,37 +472,6 @@ static Try<Nothing> installResourceLimits(const RLimitInfo& limits)
 }
 
 
-static Try<Nothing> prepareChroot(const string& rootfs)
-{
-#ifdef __WINDOWS__
-  return Error("Changing rootfs is not supported on Windows");
-#else
-  // Verify that rootfs is an absolute path.
-  Result<string> realpath = os::realpath(rootfs);
-  if (realpath.isError()) {
-    return Error(
-        "Failed to determine if rootfs '" + rootfs +
-        "' is an absolute path: " + realpath.error());
-  } else if (realpath.isNone()) {
-    return Error("Rootfs path '" + rootfs + "' does not exist");
-  } else if (realpath.get() != rootfs) {
-    return Error("Rootfs path '" + rootfs + "' is not an absolute path");
-  }
-
-#ifdef __linux__
-  Try<Nothing> prepare = fs::chroot::prepare(rootfs);
-  if (prepare.isError()) {
-    return Error(
-        "Failed to prepare chroot '" + rootfs + "': " +
-        prepare.error());
-  }
-#endif // __linux__
-
-  return Nothing();
-#endif // __WINDOWS__
-}
-
-
 static Try<Nothing> enterChroot(const string& rootfs)
 {
 #ifdef __WINDOWS__
@@ -517,39 +495,44 @@ static Try<Nothing> enterChroot(const string& rootfs)
 }
 
 
-// On Windows all new processes create by Mesos go through the
-// `create_process` wrapper which with the completion of MESOS-8926
-// will prevent inadvertent leaks making this code unnecessary there.
-//
-// TODO(bbannier): Consider moving this to stout as e.g., `os::lsof`.
-#ifndef __WINDOWS__
-static Try<hashset<int_fd>> getOpenFileDescriptors()
+#ifdef __linux__
+static void calculateCapabilities(
+    const ContainerLaunchInfo& launchInfo,
+    ProcessCapabilities* capabilities)
 {
-  Try<list<string>> fds =
-#if defined(__linux__)
-    os::ls("/proc/self/fd");
-#elif defined(__APPLE__)
-    os::ls("/dev/fd");
-#endif
+  // If the task has any effective capabilities, grant them to all
+  // the capability sets.
+  if (launchInfo.has_effective_capabilities()) {
+    set<Capability> target =
+      capabilities::convert(launchInfo.effective_capabilities());
 
-  if (fds.isError()) {
-    return Error(fds.error());
+    capabilities->set(capabilities::AMBIENT, target);
+    capabilities->set(capabilities::EFFECTIVE, target);
+    capabilities->set(capabilities::PERMITTED, target);
+    capabilities->set(capabilities::INHERITABLE, target);
+    capabilities->set(capabilities::BOUNDING, target);
   }
 
-  hashset<int_fd> result;
-  foreach (const string& fd, fds.get()) {
-    Try<int_fd> fd_ = numify<int_fd>(fd);
+  // If we also have bounding capabilities, apply that in preference to
+  // the effective capabilities.
+  if (launchInfo.has_bounding_capabilities()) {
+    set<Capability> bounding =
+      capabilities::convert(launchInfo.bounding_capabilities());
 
-    if (fd_.isError()) {
-      return Error("Could not interpret file descriptor: " + fd_.error());
-    }
-
-    result.insert(fd_.get());
+    capabilities->set(capabilities::BOUNDING, bounding);
   }
 
-  return result;
+  // Force the inherited set to be the same as the bounding set. If we
+  // are root and capabilities have not been specified, then this is a
+  // no-op. If capabilities have been specified, then we need to clip the
+  // inherited set to prevent file-based capabilities granting privileges
+  // outside the bounding set.
+  capabilities->set(
+      capabilities::INHERITABLE,
+      capabilities->get(capabilities::BOUNDING));
 }
-#endif // __WINDOWS__
+#endif // __linux__
+
 
 int MesosContainerizerLaunch::execute()
 {
@@ -730,18 +713,29 @@ int MesosContainerizerLaunch::execute()
   }
 #endif // __linux__
 
-  // Prepare root to a new root, if provided. Make sure that we do this before
-  // processing the container mounts so that container mounts can be made on
-  // top of the rootfs template.
+  // Verify that the rootfs is an absolute path.
   if (launchInfo.has_rootfs()) {
-    cerr << "Preparing rootfs at " << launchInfo.rootfs() << endl;
+    const string& rootfs = launchInfo.rootfs();
 
-    Try<Nothing> prepare = prepareChroot(launchInfo.rootfs());
+    cerr << "Preparing rootfs at '" << rootfs << "'" << endl;
 
-    if (prepare.isError()) {
-      cerr << prepare.error() << endl;
+    Result<string> realpath = os::realpath(rootfs);
+    if (realpath.isError()) {
+      cerr << "Failed to determine if rootfs '" << rootfs
+           << "' is an absolute path: " << realpath.error() << endl;
+      exitWithStatus(EXIT_FAILURE);
+    } else if (realpath.isNone()) {
+      cerr << "Rootfs path '" << rootfs << "' does not exist" << endl;
+      exitWithStatus(EXIT_FAILURE);
+    } else if (realpath.get() != rootfs) {
+      cerr << "Rootfs path '" << rootfs << "' is not an absolute path" << endl;
       exitWithStatus(EXIT_FAILURE);
     }
+
+#ifdef __WINDOWS__
+    cerr << "Changing rootfs is not supported on Windows";
+    exitWithStatus(EXIT_FAILURE);
+#endif // __WINDOWS__
   }
 
   Try<Nothing> mount = prepareMounts(launchInfo);
@@ -830,6 +824,10 @@ int MesosContainerizerLaunch::execute()
         exitWithStatus(EXIT_FAILURE);
       }
 
+      foreach (uint32_t supplementaryGroup, launchInfo.supplementary_groups()) {
+        _gids->push_back(supplementaryGroup);
+      }
+
       uid = _uid.get();
       gid = _gid.get();
       gids = _gids.get();
@@ -845,9 +843,10 @@ int MesosContainerizerLaunch::execute()
 #ifdef __linux__
   // Initialize capabilities support if necessary.
   Option<Capabilities> capabilitiesManager = None();
+  const bool needSetCapabilities = launchInfo.has_effective_capabilities() ||
+                                   launchInfo.has_bounding_capabilities();
 
-  if (launchInfo.has_effective_capabilities() ||
-      launchInfo.has_bounding_capabilities()) {
+  if (needSetCapabilities || launchInfo.has_seccomp_profile()) {
     Try<Capabilities> _capabilitiesManager = Capabilities::create();
     if (_capabilitiesManager.isError()) {
       cerr << "Failed to initialize capabilities support: "
@@ -856,15 +855,15 @@ int MesosContainerizerLaunch::execute()
     }
 
     capabilitiesManager = _capabilitiesManager.get();
+  }
 
-    // Prevent clearing of capabilities on `setuid`.
-    if (uid.isSome()) {
-      Try<Nothing> keepCaps = capabilitiesManager->setKeepCaps();
-      if (keepCaps.isError()) {
-        cerr << "Failed to set process control for keeping capabilities "
-             << "on potential uid change: " << keepCaps.error() << endl;
-        exitWithStatus(EXIT_FAILURE);
-      }
+  // Prevent clearing of capabilities on `setuid`.
+  if (needSetCapabilities && uid.isSome()) {
+    Try<Nothing> keepCaps = capabilitiesManager->setKeepCaps();
+    if (keepCaps.isError()) {
+      cerr << "Failed to set process control for keeping capabilities "
+           << "on potential uid change: " << keepCaps.error() << endl;
+      exitWithStatus(EXIT_FAILURE);
     }
   }
 #else
@@ -947,6 +946,37 @@ int MesosContainerizerLaunch::execute()
     }
   }
 
+#ifdef ENABLE_SECCOMP_ISOLATOR
+  if (launchInfo.has_seccomp_profile()) {
+    CHECK_SOME(capabilitiesManager);
+
+    Try<ProcessCapabilities> capabilities = capabilitiesManager->get();
+    if (capabilities.isError()) {
+      cerr << "Failed to get capabilities for the current process: "
+           << capabilities.error() << endl;
+      exitWithStatus(EXIT_FAILURE);
+    }
+
+    calculateCapabilities(launchInfo, &capabilities.get());
+
+    Try<Owned<SeccompFilter>> seccompFilter = SeccompFilter::create(
+        launchInfo.seccomp_profile(),
+        capabilities.get());
+
+    if (seccompFilter.isError()) {
+      cerr << "Failed to create Seccomp filter: "
+           << seccompFilter.error() << endl;
+      exitWithStatus(EXIT_FAILURE);
+    }
+
+    Try<Nothing> load = seccompFilter.get()->load();
+    if (load.isError()) {
+      cerr << "Failed to load Seccomp filter: " << load.error() << endl;
+      exitWithStatus(EXIT_FAILURE);
+    }
+  }
+#endif // ENABLE_SECCOMP_ISOLATOR
+
 #ifndef __WINDOWS__
   // Change user if provided. Note that we do that after executing the
   // preparation commands so that those commands will be run with the
@@ -976,7 +1006,9 @@ int MesosContainerizerLaunch::execute()
 #endif // __WINDOWS__
 
 #ifdef __linux__
-  if (capabilitiesManager.isSome()) {
+  if (needSetCapabilities) {
+    CHECK_SOME(capabilitiesManager);
+
     Try<ProcessCapabilities> capabilities = capabilitiesManager->get();
     if (capabilities.isError()) {
       cerr << "Failed to get capabilities for the current process: "
@@ -997,36 +1029,7 @@ int MesosContainerizerLaunch::execute()
       exitWithStatus(EXIT_FAILURE);
     }
 
-    // If the task has any effective capabilities, grant them to all
-    // the capability sets.
-    if (launchInfo.has_effective_capabilities()) {
-      set<Capability> target =
-        capabilities::convert(launchInfo.effective_capabilities());
-
-      capabilities->set(capabilities::AMBIENT, target);
-      capabilities->set(capabilities::EFFECTIVE, target);
-      capabilities->set(capabilities::PERMITTED, target);
-      capabilities->set(capabilities::INHERITABLE, target);
-      capabilities->set(capabilities::BOUNDING, target);
-    }
-
-    // If we also have bounding capabilities, apply that in preference to
-    // the effective capabilities.
-    if (launchInfo.has_bounding_capabilities()) {
-      set<Capability> bounding =
-        capabilities::convert(launchInfo.bounding_capabilities());
-
-      capabilities->set(capabilities::BOUNDING, bounding);
-    }
-
-    // Force the inherited set to be the same as the bounding set. If we
-    // are root and capabilities have not been specified, then this is a
-    // no-op. If capabilities have been specified, then we need to clip the
-    // inherited set to prevent file-based capabilities granting privileges
-    // outside the bounding set.
-    capabilities->set(
-        capabilities::INHERITABLE,
-        capabilities->get(capabilities::BOUNDING));
+    calculateCapabilities(launchInfo, &capabilities.get());
 
     Try<Nothing> set = capabilitiesManager->set(capabilities.get());
     if (set.isError()) {
@@ -1097,17 +1100,13 @@ int MesosContainerizerLaunch::execute()
 
 #ifndef __WINDOWS__
   // Construct a set of file descriptors to close before `exec`'ing.
-  Try<hashset<int_fd>> fds = getOpenFileDescriptors();
+  //
+  // On Windows all new processes create by Mesos go through the
+  // `create_process` wrapper which with the completion of MESOS-8926
+  // will prevent inadvertent leaks making this code unnecessary there.
+  Try<vector<int_fd>> fds = os::lsof();
   CHECK_SOME(fds);
 
-  std::set<int_fd> whitelistedFds = {
-    STDIN_FILENO, STDOUT_FILENO, STDERR_FILENO};
-
-  // Exclude required file descriptors from closing.
-  foreach (int_fd fd, whitelistedFds) {
-    fds->erase(fd);
-  }
-
   // If we have `containerStatusFd` set, then we need to fork-exec the
   // command we are launching and checkpoint its status on exit. We
   // use fork-exec directly (as opposed to `process::subprocess()`) to
@@ -1181,9 +1180,17 @@ int MesosContainerizerLaunch::execute()
 
     // Avoid leaking not required file descriptors into the forked process.
     foreach (int_fd fd, fds.get()) {
-      // We use the unwrapped `::close` as opposed to `os::close`
-      // since the former is guaranteed to be async signal safe.
-      ::close(fd);
+      if (fd != STDIN_FILENO && fd != STDOUT_FILENO && fd != STDERR_FILENO) {
+        // NOTE: Set "FD_CLOEXEC" on the fd, instead of closing it
+        // because exec below might exec a memfd.
+        int flags = ::fcntl(fd, F_GETFD);
+        if (flags == -1) {
+          ABORT("Failed to get FD flags");
+        }
+        if (::fcntl(fd, F_SETFD, flags | FD_CLOEXEC) == -1) {
+          ABORT("Failed to set FD_CLOEXEC");
+        }
+      }
     }
   }
 #endif // __WINDOWS__
