@@ -528,11 +528,6 @@ Try<pid_t> LinuxLauncherProcess::fork(
 
   childHooks.push_back(Subprocess::ChildHook::SETSID());
 
-  // TODO(jpeach) libprocess should take care of this, see MESOS-9164.
-  foreach (int_fd fd, whitelistFds) {
-    childHooks.push_back(Subprocess::ChildHook::UNSET_CLOEXEC(fd));
-  }
-
   Try<Subprocess> child = subprocess(
       path,
       argv,
@@ -559,7 +554,8 @@ Try<pid_t> LinuxLauncherProcess::fork(
         }
       },
       parentHooks,
-      childHooks);
+      childHooks,
+      whitelistFds);
 
   if (child.isError()) {
     return Error("Failed to clone child process: " + child.error());
