@@ -441,6 +441,10 @@ Future<Nothing> ProvisionerProcess::recover(
             layers->paths().begin(),
             layers->paths().end(),
             std::back_inserter(info->layers.get()));
+
+        if (layers->has_config()) {
+          info->layers->push_back(layers->config());
+        }
       }
     }
 
@@ -555,6 +559,10 @@ Future<ProvisionInfo> ProvisionerProcess::_provision(
   infos[containerId]->rootfses[backend].insert(rootfsId);
   infos[containerId]->layers = imageInfo.layers;
 
+  if (imageInfo.config.isSome()) {
+    infos[containerId]->layers->push_back(imageInfo.config.get());
+  }
+
   string backendDir = provisioner::paths::getBackendDir(
       rootDir,
       containerId,
@@ -574,6 +582,10 @@ Future<ProvisionInfo> ProvisionerProcess::_provision(
         containerLayers.add_paths(layer);
       }
 
+      if (imageInfo.config.isSome()) {
+        containerLayers.set_config(imageInfo.config.get());
+      }
+
       Try<Nothing> checkpoint = slave::state::checkpoint(path, containerLayers);
       if (checkpoint.isError()) {
         return Failure(
