@@ -175,15 +175,18 @@ mesos::internal::slave::Flags::Flags()
       "auto image gc is enabled. This image gc config can be provided either\n"
       "as a path pointing to a local file, or as a JSON-formatted string.\n"
       "Please note that the image garbage collection only work with Mesos\n"
-      "Containerizer for now."
+      "Containerizer for now.\n"
       "\n"
       "See the ImageGcConfig message in `flags.proto` for the expected\n"
       "format.\n"
-      "Example:\n"
+      "\n"
+      "In the following example, image garbage collection is configured to\n"
+      "sample disk usage every hour, and will attempt to maintain at least\n"
+      "10 percent of free space on the container image filesystem:\n"
       "{\n"
       "  \"image_disk_headroom\": 0.1,\n"
       "  \"image_disk_watch_interval\": {\n"
-      "    \"nanoseconds\": 3600\n"
+      "    \"nanoseconds\": 3600000000000\n"
       "  },\n"
       "  \"excluded_images\": []\n"
       "}");
@@ -327,6 +330,15 @@ mesos::internal::slave::Flags::Flags()
       "NOTE: This feature is not yet supported on Windows agent, and\n"
       "therefore the flag currently does not exist on that platform.",
       true);
+
+  add(&Flags::volume_gid_range,
+      "volume_gid_range",
+      "When this flag is specified, if a task running as non-root user uses a\n"
+      "shared persistent volume or a PARENT type SANDBOX_PATH volume, the\n"
+      "volume will be owned by a gid allocated from this range and have the\n"
+      "`setgid` bit set, and the task process will be launched with the gid\n"
+      "as its supplementary group to make sure it can access the volume.\n"
+      "(Example: `[10000-20000]`)");
 #endif // __WINDOWS__
 
   add(&Flags::http_heartbeat_interval,
@@ -647,6 +659,12 @@ mesos::internal::slave::Flags::Flags()
       "no cgroup limits are set, they are inherited from the root mesos\n"
       "cgroup.");
 
+  add(&Flags::host_path_volume_force_creation,
+      "host_path_volume_force_creation",
+      "A colon-separated list of directories where descendant directories\n"
+      "are allowed to be created by the `volume/host_path` isolator,\n"
+      "if the directories do not exist.");
+
   add(&Flags::nvidia_gpu_devices,
       "nvidia_gpu_devices",
       "A comma-separated list of Nvidia GPU devices. When `gpus` is\n"
@@ -752,14 +770,16 @@ mesos::internal::slave::Flags::Flags()
   add(&Flags::agent_features,
       "agent_features",
       "JSON representation of agent features to whitelist. We always require\n"
-      "'MULTI_ROLE', 'HIERARCHICAL_ROLE', and 'RESERVATION_REFINEMENT'.\n"
+      "'MULTI_ROLE', 'HIERARCHICAL_ROLE', 'RESERVATION_REFINEMENT', and\n"
+      "'AGENT_OPERATION_FEEDBACK'.\n"
       "\n"
       "Example:\n"
       "{\n"
       "    \"capabilities\": [\n"
       "        {\"type\": \"MULTI_ROLE\"},\n"
       "        {\"type\": \"HIERARCHICAL_ROLE\"},\n"
-      "        {\"type\": \"RESERVATION_REFINEMENT\"}\n"
+      "        {\"type\": \"RESERVATION_REFINEMENT\"},\n"
+      "        {\"type\": \"AGENT_OPERATION_FEEDBACK\"}\n"
       "    ]\n"
       "}\n",
       [](const Option<SlaveCapabilities>& agentFeatures) -> Option<Error> {
@@ -770,16 +790,25 @@ mesos::internal::slave::Flags::Flags()
 
           if (!capabilities.multiRole ||
               !capabilities.hierarchicalRole ||
-              !capabilities.reservationRefinement) {
+              !capabilities.reservationRefinement ||
+              !capabilities.agentOperationFeedback) {
             return Error(
-                "At least the following agent features need to be enabled: "
-                "MULTI_ROLE, HIERARCHICAL_ROLE, RESERVATION_REFINEMENT");
+                "At least the following agent features need to be enabled:"
+                " MULTI_ROLE, HIERARCHICAL_ROLE, RESERVATION_REFINEMENT,"
+                " AGENT_OPERATION_FEEDBACK");
           }
 
           if (capabilities.resizeVolume && !capabilities.resourceProvider) {
             return Error(
                 "RESIZE_VOLUME feature requires RESOURCE_PROVIDER feature");
           }
+
+          if (capabilities.agentOperationFeedback &&
+              !capabilities.resourceProvider) {
+            return Error(
+                "AGENT_OPERATION_FEEDBACK feature"
+                " requires RESOURCE_PROVIDER feature");
+          }
         }
 
         return None();
@@ -1152,11 +1181,19 @@ mesos::internal::slave::Flags::Flags()
       "published by the agent's resources. Otherwise tasks are restricted\n"
       "to only listen on ports for which they have been assigned resources.",
       false);
+
   add(&Flags::enforce_container_ports,
       "enforce_container_ports",
       "Whether to enable port enforcement for containers. This flag\n"
-      "is used by `network/ports` isolator.",
+      "is used by the `network/ports` isolator.",
       false);
+
+  add(&Flags::container_ports_isolated_range,
+      "container_ports_isolated_range",
+      "When this flag is specified, the `network/ports` isolator will\n"
+      "only enforce port isolation for the specified range of ports.\n"
+      "(Example: `[0-35000]`)\n");
+
 #endif // ENABLE_NETWORK_PORTS_ISOLATOR
 
   add(&Flags::network_cni_plugins_dir,
@@ -1173,6 +1210,12 @@ mesos::internal::slave::Flags::Flags()
       "the operator should install a network configuration file in JSON\n"
       "format in the specified directory.");
 
+  add(&Flags::network_cni_root_dir_persist,
+      "network_cni_root_dir_persist",
+      "This setting controls whether the CNI root directory\n"
+      "persists across reboot or not.",
+      false);
+
   add(&Flags::network_cni_metrics,
       "network_cni_metrics",
       "This setting controls whether the networking metrics of the CNI\n"
@@ -1369,6 +1412,26 @@ mesos::internal::slave::Flags::Flags()
       false);
 #endif
 
+#if ENABLE_SECCOMP_ISOLATOR
+  add(&Flags::seccomp_config_dir,
+      "seccomp_config_dir",
+      "Directory path of the Seccomp profiles.\n"
+      "If a container is launched with a specified Seccomp profile name,\n"
+      "the `linux/seccomp` isolator will try to locate a Seccomp profile\n"
+      "in the specified directory.");
+
+  add(&Flags::seccomp_profile_name,
+      "seccomp_profile_name",
+      "Path of the default Seccomp profile relative to the"
+      "`seccomp_config_dir`.\n"
+      "If this flag is specified, the `linux/seccomp` isolator\n"
+      "applies the Seccomp profile by default when launching\n"
+      "a new Mesos container.\n"
+      "NOTE: A Seccomp profile must be compatible with the\n"
+      "Docker Seccomp profile format (e.g., https://github.com/moby/moby/"
+      "blob/master/profiles/seccomp/default.json).");
+#endif
+
   add(&Flags::http_command_executor,
       "http_command_executor",
       "The underlying executor library to be used for the command executor.\n"
