@@ -56,6 +56,7 @@
 #include "authentication/executor/jwt_secret_generator.hpp"
 #endif // USE_SSL_SOCKET
 
+#include "common/authorization.hpp"
 #include "common/build.hpp"
 #include "common/http.hpp"
 
@@ -202,7 +203,7 @@ static Try<Nothing> assignCgroups(const slave::Flags& flags)
 
       LOG(INFO) << "An agent (or child process) is still running, please"
                 << " consider checking the following process(es) listed in "
-                << path::join(hierarchy.get(), cgroup, "cgroups.proc")
+                << path::join(hierarchy.get(), cgroup, "cgroup.procs")
                 << ":\n" << strings::join("\n", infos);
     }
 
@@ -476,8 +477,27 @@ int main(int argc, char** argv)
         << "Failed to initialize secret resolver: " << secretResolver.error();
   }
 
-  Try<Containerizer*> containerizer =
-    Containerizer::create(flags, false, fetcher, gc, secretResolver.get());
+  VolumeGidManager* volumeGidManager = nullptr;
+
+#ifndef __WINDOWS__
+  if (flags.volume_gid_range.isSome()) {
+    Try<VolumeGidManager*> _volumeGidManager = VolumeGidManager::create(flags);
+    if (_volumeGidManager.isError()) {
+      EXIT(EXIT_FAILURE) << "Failed to initialize volume gid manager: "
+                         << _volumeGidManager.error();
+    }
+
+    volumeGidManager = _volumeGidManager.get();
+  }
+#endif // __WINDOWS__
+
+  Try<Containerizer*> containerizer = Containerizer::create(
+      flags,
+      false,
+      fetcher,
+      gc,
+      secretResolver.get(),
+      volumeGidManager);
 
   if (containerizer.isError()) {
     EXIT(EXIT_FAILURE)
@@ -527,7 +547,7 @@ int main(int argc, char** argv)
     // future it becomes possible to dynamically set the authorizer, this would
     // break.
     process::http::authorization::setCallbacks(
-        createAuthorizationCallbacks(authorizer_.get()));
+        mesos::authorization::createAuthorizationCallbacks(authorizer_.get()));
   }
 
   Files* files = new Files(READONLY_HTTP_AUTHENTICATION_REALM, authorizer_);
@@ -590,6 +610,7 @@ int main(int argc, char** argv)
       resourceEstimator.get(),
       qosController.get(),
       secretGenerator,
+      volumeGidManager,
       authorizer_);
 
   process::spawn(slave);
@@ -615,6 +636,12 @@ int main(int argc, char** argv)
 
   delete containerizer.get();
 
+#ifndef __WINDOWS__
+  delete volumeGidManager;
+#endif // __WINDOWS__
+
+  delete secretResolver.get();
+
   delete gc;
 
   delete fetcher;
