@@ -64,10 +64,12 @@ const char HTTP_MARKER_FILE[] = "http.marker";
 const char FORKED_PID_FILE[] = "forked.pid";
 const char TASK_INFO_FILE[] = "task.info";
 const char TASK_UPDATES_FILE[] = "task.updates";
+const char RESOURCE_STATE_FILE[] = "resources_and_operations.state";
 const char RESOURCES_INFO_FILE[] = "resources.info";
 const char RESOURCES_TARGET_FILE[] = "resources.target";
 const char RESOURCE_PROVIDER_STATE_FILE[] = "resource_provider.state";
 const char OPERATION_UPDATES_FILE[] = "operation.updates";
+const char VOLUME_GIDS_FILE[] = "volume_gids";
 
 
 const char CONTAINERS_DIR[] = "containers";
@@ -556,6 +558,14 @@ Try<list<string>> getOperationPaths(
 }
 
 
+Try<list<string>> getSlaveOperationPaths(
+    const string& metaDir,
+    const SlaveID& slaveId)
+{
+  return getOperationPaths(getSlavePath(metaDir, slaveId));
+}
+
+
 string getOperationPath(
     const string& rootDir,
     const id::UUID& operationUuid)
@@ -564,6 +574,15 @@ string getOperationPath(
 }
 
 
+string getSlaveOperationPath(
+    const string& metaDir,
+    const SlaveID& slaveId,
+    const id::UUID& operationUuid)
+{
+  return getOperationPath(getSlavePath(metaDir, slaveId), operationUuid);
+}
+
+
 Try<id::UUID> parseOperationPath(
     const string& rootDir,
     const string& dir)
@@ -591,6 +610,15 @@ Try<id::UUID> parseOperationPath(
 }
 
 
+Try<id::UUID> parseSlaveOperationPath(
+    const string& metaDir,
+    const SlaveID& slaveId,
+    const string& dir)
+{
+  return parseOperationPath(getSlavePath(metaDir, slaveId), dir);
+}
+
+
 string getOperationUpdatesPath(
     const string& rootDir,
     const id::UUID& operationUuid)
@@ -601,6 +629,21 @@ string getOperationUpdatesPath(
 }
 
 
+string getSlaveOperationUpdatesPath(
+    const string& metaDir,
+    const SlaveID& slaveId,
+    const id::UUID& operationUuid)
+{
+  return getOperationUpdatesPath(getSlavePath(metaDir, slaveId), operationUuid);
+}
+
+
+string getResourceStatePath(const string& rootDir)
+{
+  return path::join(rootDir, "resources", RESOURCE_STATE_FILE);
+}
+
+
 string getResourcesInfoPath(
     const string& rootDir)
 {
@@ -615,6 +658,13 @@ string getResourcesTargetPath(
 }
 
 
+Try<list<string>> getPersistentVolumePaths(
+    const std::string& workDir)
+{
+  return fs::list(path::join(workDir, "volumes", "roles", "*", "*"));
+}
+
+
 string getPersistentVolumePath(
     const string& workDir,
     const string& role,
@@ -722,6 +772,12 @@ string getPersistentVolumePath(
 }
 
 
+string getVolumeGidsPath(const string& rootDir)
+{
+  return path::join(rootDir, "volume_gid_manager", VOLUME_GIDS_FILE);
+}
+
+
 Try<string> createExecutorDirectory(
     const string& rootDir,
     const SlaveID& slaveId,
@@ -792,7 +848,7 @@ Try<Nothing> createSandboxDirectory(
   // Since this is a sandbox directory containing private task data,
   // we want to ensure that it is not accessible to "others".
   Try<Nothing> chmod = os::chmod(directory, 0750);
-  if (mkdir.isError()) {
+  if (chmod.isError()) {
     return Error("Failed to chmod directory: " + chmod.error());
   }
 
