@@ -22,13 +22,9 @@
 #include <list>
 #include <memory>
 #include <string>
+#include <utility>
 #include <vector>
 
-// See comments in `master.hpp` or MESOS-9177.
-#define BOOST_CB_DISABLE_DEBUG 1
-#define BOOST_CB_ENABLE_DEBUG 0
-#include <boost/circular_buffer.hpp>
-
 #include <mesos/attributes.hpp>
 #include <mesos/resources.hpp>
 #include <mesos/type_utils.hpp>
@@ -60,6 +56,7 @@
 
 #include <stout/boundedhashmap.hpp>
 #include <stout/bytes.hpp>
+#include <stout/circular_buffer.hpp>
 #include <stout/linkedhashmap.hpp>
 #include <stout/hashmap.hpp>
 #include <stout/hashset.hpp>
@@ -69,6 +66,7 @@
 #include <stout/recordio.hpp>
 #include <stout/uuid.hpp>
 
+#include "common/heartbeater.hpp"
 #include "common/http.hpp"
 #include "common/protobuf_utils.hpp"
 #include "common/recordio.hpp"
@@ -91,6 +89,8 @@
 #include "slave/paths.hpp"
 #include "slave/state.hpp"
 
+#include "status_update_manager/operation.hpp"
+
 // `REGISTERING` is used as an enum value, but it's actually defined as a
 // constant in the Windows SDK.
 #ifdef __WINDOWS__
@@ -112,7 +112,6 @@ class TaskStatusUpdateManager;
 class Executor;
 class Framework;
 
-struct HttpConnection;
 struct ResourceProvider;
 
 
@@ -129,6 +128,7 @@ public:
         mesos::slave::ResourceEstimator* resourceEstimator,
         mesos::slave::QoSController* qosController,
         mesos::SecretGenerator* secretGenerator,
+        VolumeGidManager* volumeGidManager,
         const Option<Authorizer*>& authorizer);
 
   ~Slave() override;
@@ -259,12 +259,14 @@ public:
   void updateFramework(
       const UpdateFrameworkMessage& message);
 
-  void checkpointResources(
-      std::vector<Resource> checkpointedResources,
+  void checkpointResourceState(const Resources& resources, bool changeTotal);
+
+  void checkpointResourceState(
+      std::vector<Resource> resources,
       bool changeTotal);
 
   void checkpointResourcesMessage(
-      const std::vector<Resource>& checkpointedResources);
+      const std::vector<Resource>& resources);
 
   void applyOperation(const ApplyOperationMessage& message);
 
@@ -275,7 +277,7 @@ public:
   void reconcileOperations(const ReconcileOperationsMessage& message);
 
   void subscribe(
-      HttpConnection http,
+      StreamingHttpConnection<v1::executor::Event> http,
       const executor::Call::Subscribe& subscribe,
       Framework* framework,
       Executor* executor);
@@ -352,6 +354,10 @@ public:
   // added to the update before forwarding.
   void forward(StatusUpdate update);
 
+  // This is called by the operation status update manager to forward operation
+  // status updates to the master.
+  void sendOperationStatusUpdate(const UpdateOperationStatusMessage& update);
+
   void statusUpdateAcknowledgement(
       const process::UPID& from,
       const SlaveID& slaveId,
@@ -428,16 +434,20 @@ public:
   void finalize() override;
   void exited(const process::UPID& pid) override;
 
-  process::Future<Secret> generateSecret(
+  // Generates a secret for executor authentication. Returns None if there is
+  // no secret generator.
+  process::Future<Option<Secret>> generateSecret(
       const FrameworkID& frameworkId,
       const ExecutorID& executorId,
       const ContainerID& containerId);
 
-  // If an executor is launched for a task group, `taskInfo` would not be set.
+  // `executorInfo` is a mutated executor info with some default fields and
+  // resources. If an executor is launched for a task group, `taskInfo` would
+  // not be set.
   void launchExecutor(
-      const Option<process::Future<Secret>>& future,
+      const process::Future<Option<Secret>>& authorizationToken,
       const FrameworkID& frameworkId,
-      const ExecutorID& executorId,
+      const ExecutorInfo& executorInfo,
       const Option<TaskInfo>& taskInfo);
 
   void fileAttached(const process::Future<Nothing>& result,
@@ -495,7 +505,8 @@ public:
   // not receive a ping.
   void pingTimeout(process::Future<Option<MasterInfo>> future);
 
-  void authenticate(Duration minTimeout, Duration maxTimeout);
+  // Made virtual for testing purpose.
+  virtual void authenticate(Duration minTimeout, Duration maxTimeout);
 
   // Helper routines to lookup a framework/executor.
   Framework* getFramework(const FrameworkID& frameworkId) const;
@@ -541,12 +552,33 @@ public:
   // executors. Otherwise, the slave attempts to shutdown/kill them.
   process::Future<Nothing> _recover();
 
-  // This is a helper to call recover() on the containerizer at the end of
-  // recover() and before __recover().
+  // This is a helper to call `recover()` on the volume gid manager.
+  process::Future<Nothing> _recoverVolumeGidManager(bool rebooted);
+
+  // This is a helper to call `recover()` on the task status update manager.
+  process::Future<Option<state::SlaveState>> _recoverTaskStatusUpdates(
+      const Option<state::SlaveState>& slaveState);
+
+  // This is a helper to call `recover()` on the containerizer at the end of
+  // `recover()` and before `__recover()`.
   // TODO(idownes): Remove this when we support defers to objects.
   process::Future<Nothing> _recoverContainerizer(
       const Option<state::SlaveState>& state);
 
+  // This is called after `_recoverContainerizer()`. It will add all
+  // checkpointed operations affecting agent default resources and call
+  // `OperationStatusUpdateManager::recover()`.
+  process::Future<Nothing> _recoverOperations(
+      const Option<state::SlaveState>& state);
+
+  // This is called after `OperationStatusUpdateManager::recover()`
+  // completes.
+  //
+  // If necessary it will add any missing operation status updates
+  // that couldn't be checkpointed before the agent failed over.
+  process::Future<Nothing> __recoverOperations(
+      const process::Future<OperationStatusUpdateManagerState>& state);
+
   // This is called when recovery finishes.
   // Made 'virtual' for Slave mocking.
   virtual void __recover(const process::Future<Nothing>& future);
@@ -611,9 +643,6 @@ private:
       const TaskInfo& task,
       const FrameworkInfo& frameworkInfo);
 
-  process::Future<bool> authorizeLogAccess(
-      const Option<process::http::authentication::Principal>& principal);
-
   process::Future<bool> authorizeSandboxAccess(
       const Option<process::http::authentication::Principal>& principal,
       const FrameworkID& frameworkId,
@@ -661,6 +690,9 @@ private:
 
   void removeOperation(Operation* operation);
 
+  process::Future<Nothing> markResourceProviderGone(
+      const ResourceProviderID& resourceProviderId) const;
+
   Operation* getOperation(const UUID& uuid) const;
 
   void addResourceProvider(ResourceProvider* resourceProvider);
@@ -668,13 +700,9 @@ private:
 
   void apply(Operation* operation);
 
-  // Publish all resources that are needed to run the current set of
-  // tasks and executors on the agent.
-  // NOTE: The `additionalResources` parameter is for publishing
-  // additional task resources when launching executors. Consider
-  // removing this parameter once we revisited MESOS-600.
+  // Prepare all resources to be consumed by the specified container.
   process::Future<Nothing> publishResources(
-      const Option<Resources>& additionalResources = None());
+      const ContainerID& containerId, const Resources& resources);
 
   // PullGauge methods.
   double _frameworks_active()
@@ -767,6 +795,8 @@ private:
 
   TaskStatusUpdateManager* taskStatusUpdateManager;
 
+  OperationStatusUpdateManager operationStatusUpdateManager;
+
   // Master detection future.
   process::Future<Option<MasterInfo>> detection;
 
@@ -814,6 +844,8 @@ private:
 
   mesos::SecretGenerator* secretGenerator;
 
+  VolumeGidManager* volumeGidManager;
+
   const Option<Authorizer*> authorizer;
 
   // The most recent estimate of the total amount of oversubscribed
@@ -839,46 +871,19 @@ private:
   // have already been invalidated.
   UUID resourceVersion;
 
-  // Keeps track of the following:
-  // (1) Pending operations for resources from the agent.
-  // (2) Pending operations or terminal operations that have
-  //     unacknowledged status updates for resource provider
-  //     provided resources.
+  // Keeps track of pending operations or terminal operations that
+  // have unacknowledged status updates. These operations may affect
+  // either agent default resources or resources offered by a resource
+  // provider.
   hashmap<UUID, Operation*> operations;
-};
 
+  // Maps framework-supplied operation IDs to the operation's internal UUID.
+  // This is used to satisfy some reconciliation requests which are forwarded
+  // from the master to the agent.
+  hashmap<std::pair<FrameworkID, OperationID>, UUID> operationIds;
 
-// Represents the streaming HTTP connection to an executor.
-struct HttpConnection
-{
-  HttpConnection(const process::http::Pipe::Writer& _writer,
-                 ContentType _contentType)
-    : writer(_writer),
-      contentType(_contentType),
-      encoder(lambda::bind(serialize, contentType, lambda::_1)) {}
-
-  // Converts the message to an Event before sending.
-  template <typename Message>
-  bool send(const Message& message)
-  {
-    // We need to evolve the internal 'message' into a
-    // 'v1::executor::Event'.
-    return writer.write(encoder.encode(evolve(message)));
-  }
-
-  bool close()
-  {
-    return writer.close();
-  }
-
-  process::Future<Nothing> closed() const
-  {
-    return writer.readerClosed();
-  }
-
-  process::http::Pipe::Writer writer;
-  ContentType contentType;
-  ::recordio::Encoder<v1::executor::Event> encoder;
+  // Operations that are checkpointed by the agent.
+  hashmap<UUID, Operation> checkpointedOperations;
 };
 
 
@@ -1018,7 +1023,10 @@ public:
   //           *       REGISTERING       None       None   Not known yet
   //           *                 *       None       Some      Libprocess
   //           *                 *       Some       None            HTTP
-  Option<HttpConnection> http;
+  Option<StreamingHttpConnection<v1::executor::Event>> http;
+  process::Owned<ResponseHeartbeater<executor::Event, v1::executor::Event>>
+    heartbeater;
+
   Option<process::UPID> pid;
 
   // Tasks can be found in one of the following four data structures:
@@ -1043,9 +1051,9 @@ public:
 
   // Terminated and updates acked.
   // NOTE: We use a shared pointer for Task because clang doesn't like
-  // Boost's implementation of circular_buffer with Task (Boost
-  // attempts to do some memset's which are unsafe).
-  boost::circular_buffer<std::shared_ptr<Task>> completedTasks;
+  // stout's implementation of circular_buffer with Task (the Boost code
+  // used internally by stout attempts to do some memset's which are unsafe).
+  circular_buffer<std::shared_ptr<Task>> completedTasks;
 
   // When the slave initiates a destroy of the container, we expect a
   // termination to occur. The 'pendingTermation' indicates why the
@@ -1173,7 +1181,7 @@ public:
   // Current running executors.
   hashmap<ExecutorID, Executor*> executors;
 
-  boost::circular_buffer<process::Owned<Executor>> completedExecutors;
+  circular_buffer<process::Owned<Executor>> completedExecutors;
 
 private:
   Framework(const Framework&) = delete;
@@ -1186,7 +1194,7 @@ struct ResourceProvider
   ResourceProvider(
       const ResourceProviderInfo& _info,
       const Resources& _totalResources,
-      const UUID& _resourceVersion)
+      const Option<UUID>& _resourceVersion)
     : info(_info),
       totalResources(_totalResources),
       resourceVersion(_resourceVersion) {}
@@ -1208,7 +1216,7 @@ struct ResourceProvider
   // different resource version UUID than that it maintains, because
   // this means the operation is operating on resources that might
   // have already been invalidated.
-  UUID resourceVersion;
+  Option<UUID> resourceVersion;
 
   // Pending operations or terminal operations that have
   // unacknowledged status updates.
