@@ -70,6 +70,7 @@
 
 #include "authorizer/local/authorizer.hpp"
 
+#include "common/authorization.hpp"
 #include "common/http.hpp"
 
 #include "files/files.hpp"
@@ -96,6 +97,8 @@
 #include "slave/containerizer/containerizer.hpp"
 #include "slave/containerizer/fetcher.hpp"
 
+#include "slave/volume_gid_manager/volume_gid_manager.hpp"
+
 #include "tests/cluster.hpp"
 #include "tests/mock_registrar.hpp"
 
@@ -395,7 +398,7 @@ MasterInfo Master::getMasterInfo()
 void Master::setAuthorizationCallbacks(Authorizer* authorizer)
 {
   process::http::authorization::setCallbacks(
-      mesos::createAuthorizationCallbacks(authorizer));
+      authorization::createAuthorizationCallbacks(authorizer));
 
   authorizationCallbacksSet = true;
 }
@@ -426,6 +429,23 @@ Try<process::Owned<Slave>> Slave::create(
     slave->gc.reset(new slave::GarbageCollector(flags.work_dir));
   }
 
+  // If the flag `--volume_gid_range` is specified, create a volume gid manager.
+  slave::VolumeGidManager* volumeGidManager = nullptr;
+
+#ifndef __WINDOWS__
+  if (flags.volume_gid_range.isSome()) {
+    Try<slave::VolumeGidManager*> _volumeGidManager =
+      slave::VolumeGidManager::create(flags);
+
+    if (_volumeGidManager.isError()) {
+      return Error(
+          "Failed to create volume gid manager: " + _volumeGidManager.error());
+    }
+
+    volumeGidManager = _volumeGidManager.get();
+  }
+#endif // __WINDOWS__
+
   // If the containerizer is not provided, create a default one.
   if (containerizer.isSome()) {
     slave->containerizer = containerizer.get();
@@ -435,7 +455,12 @@ Try<process::Owned<Slave>> Slave::create(
 
     Try<slave::Containerizer*> _containerizer =
       slave::Containerizer::create(
-          flags, true, slave->fetcher.get(), gc.getOrElse(slave->gc.get()));
+          flags,
+          true,
+          slave->fetcher.get(),
+          gc.getOrElse(slave->gc.get()),
+          nullptr,
+          volumeGidManager);
 
     if (_containerizer.isError()) {
       return Error("Failed to create containerizer: " + _containerizer.error());
@@ -535,7 +560,7 @@ Try<process::Owned<Slave>> Slave::create(
     slave->qosController.reset(_qosController.get());
   }
 
-  // If the QoS controller is not provided, create a default one.
+  // If the secret generator is not provided, create a default one.
   if (secretGenerator.isNone()) {
     SecretGenerator* _secretGenerator = nullptr;
 
@@ -587,6 +612,7 @@ Try<process::Owned<Slave>> Slave::create(
         resourceEstimator.getOrElse(slave->resourceEstimator.get()),
         qosController.getOrElse(slave->qosController.get()),
         secretGenerator.getOrElse(slave->secretGenerator.get()),
+        volumeGidManager,
         authorizer));
   } else {
     slave->slave.reset(new slave::Slave(
@@ -600,6 +626,7 @@ Try<process::Owned<Slave>> Slave::create(
         resourceEstimator.getOrElse(slave->resourceEstimator.get()),
         qosController.getOrElse(slave->qosController.get()),
         secretGenerator.getOrElse(slave->secretGenerator.get()),
+        volumeGidManager,
         authorizer));
   }
 
@@ -687,8 +714,19 @@ Slave::~Slave()
       }
     }
 
-    if (paused) {
-      process::Clock::pause();
+    // When using the composing containerizer, the assertion checking
+    // `containers->empty()` below is racy, since the containers are
+    // not yet removed from that containerizer's internal data structures
+    // when the future becomes ready. Instead, an internal function to clean
+    // up these internal data structures is dispatched when the future
+    // becomes ready.
+    // To work around this, we wait for the clock to settle here. This can
+    // be removed once MESOS-9413 is resolved.
+    process::Clock::pause();
+    process::Clock::settle();
+
+    if (!paused) {
+      process::Clock::resume();
     }
 
     containers = containerizer->containers();
@@ -703,7 +741,7 @@ Slave::~Slave()
 void Slave::setAuthorizationCallbacks(Authorizer* authorizer)
 {
   process::http::authorization::setCallbacks(
-      mesos::createAuthorizationCallbacks(authorizer));
+      authorization::createAuthorizationCallbacks(authorizer));
 
   authorizationCallbacksSet = true;
 }
