@@ -320,6 +320,48 @@ TEST_F(FilesTest, ResolveTest)
 }
 
 
+// Tests paths with percent-encoded sequences in the HTTP request
+// query. Very specifically, this checks that a percent-encoded symbol
+// is not "double decoded." That is to say, say we have a literal path
+// `foo%3Abar` because we couldn't use `:` literally on Windows, and
+// so instead encoded it literally in the file path on Windows. This
+// demonstrated a bug in libprocess where the encoding `%3A` was
+// decoded too many times, and so the query was for `foo:bar` instead
+// of literally `foo%3Abar`.
+TEST_F(FilesTest, QueryWithEncodedSequence)
+{
+  Files files;
+  process::UPID upid("files", process::address());
+
+  // This path has the ASCII escape sequence `%3A` literally instead
+  // of `:` because the latter is a reserved character on Windows.
+  //
+  // NOTE: This is not just an arbitrary character such as `+`, it is
+  // explicitly a percent-encoded sequence, but it could be e.g. `+`
+  // percent-encoded as `%2B`. Hence the assertion that this could be
+  // decoded again.
+  const string filename = "foo%3Abar";
+  ASSERT_SOME_EQ("foo:bar", process::http::decode(filename));
+
+  ASSERT_SOME(os::write(filename, "body"));
+  ASSERT_SOME_EQ("body", os::read(filename));
+  AWAIT_EXPECT_READY(files.attach(sandbox.get(), "/"));
+
+  // NOTE: The query here has to be encoded because it is a `string`
+  // and not a `hashmap<string, string>`. The latter is automatically
+  // encoded, but the former is not.
+  Future<Response> response = process::http::get(
+      upid, "read", "path=/" + process::http::encode(filename) + "&offset=0");
+
+  JSON::Object expected;
+  expected.values["offset"] = 0;
+  expected.values["data"] = "body";
+
+  AWAIT_EXPECT_RESPONSE_STATUS_EQ(OK().status, response);
+  AWAIT_EXPECT_RESPONSE_BODY_EQ(stringify(expected), response);
+}
+
+
 TEST_F(FilesTest, BrowseTest)
 {
   Files files;
