@@ -914,8 +914,14 @@ TEST_F(GarbageCollectorIntegrationTest, LongLivedDefaultExecutorRestart)
   // Turn on GC of nested container sandboxes by default.
   flags.gc_non_executor_container_sandboxes = true;
 
+  // Start the slave with a static process ID. This allows the executor to
+  // reconnect with the slave upon a process restart.
+  const string id(process::ID::generate("agent"));
+
   Owned<MasterDetector> detector = master.get()->createDetector();
-  Try<Owned<cluster::Slave>> slave = StartSlave(detector.get(), flags);
+  Try<Owned<cluster::Slave>> slave =
+    StartSlave(detector.get(), id, flags, false);
+
   ASSERT_SOME(slave);
 
   auto scheduler = std::make_shared<v1::MockHTTPScheduler>();
@@ -1093,7 +1099,7 @@ TEST_F(GarbageCollectorIntegrationTest, LongLivedDefaultExecutorRestart)
 
   // The agent should reregister once recovery is complete, which also means
   // that any finished tasks metadata/sandboxes should be rescheduled for GC.
-  slave = StartSlave(detector.get(), flags);
+  slave = StartSlave(detector.get(), id, flags, false);
   ASSERT_SOME(slave);
   AWAIT_READY(slaveReregisteredMessage);
 
