@@ -25,6 +25,7 @@
 
 #include <mesos/quota/quota.hpp>
 
+#include <process/clock.hpp>
 #include <process/future.hpp>
 #include <process/http.hpp>
 #include <process/id.hpp>
@@ -50,6 +51,8 @@ using google::protobuf::RepeatedPtrField;
 
 using mesos::internal::master::Master;
 
+using mesos::internal::master::allocator::MesosAllocatorProcess;
+
 using mesos::internal::slave::Slave;
 
 using mesos::master::detector::MasterDetector;
@@ -58,6 +61,7 @@ using mesos::quota::QuotaInfo;
 using mesos::quota::QuotaRequest;
 using mesos::quota::QuotaStatus;
 
+using process::Clock;
 using process::Future;
 using process::Owned;
 using process::PID;
@@ -423,8 +427,10 @@ TEST_F(MasterQuotaTest, SetExistingQuota)
 // endpoint via a DELETE request against /quota.
 TEST_F(MasterQuotaTest, RemoveSingleQuota)
 {
+  Clock::pause();
+
   TestAllocator<> allocator;
-  EXPECT_CALL(allocator, initialize(_, _, _, _, _, _, _, _));
+  EXPECT_CALL(allocator, initialize(_, _, _));
 
   Try<Owned<cluster::Master>> master = StartMaster(&allocator);
   ASSERT_SOME(master);
@@ -468,9 +474,13 @@ TEST_F(MasterQuotaTest, RemoveSingleQuota)
 
     AWAIT_EXPECT_RESPONSE_STATUS_EQ(OK().status, response);
 
+    // Ensure metrics update is finished.
+    Clock::settle();
+
     const string metricKey =
       "allocator/mesos/quota/roles/" + ROLE1 + "/resources/cpus/guarantee";
 
+
     JSON::Object metrics = Metrics();
 
     EXPECT_EQ(1, metrics.values[metricKey]);
@@ -488,6 +498,9 @@ TEST_F(MasterQuotaTest, RemoveSingleQuota)
     // Ensure that the quota remove request has reached the allocator.
     AWAIT_READY(receivedRemoveRequest);
 
+    // Ensure metrics update is finished.
+    Clock::settle();
+
     metrics = Metrics();
 
     ASSERT_NONE(metrics.at<JSON::Number>(metricKey));
@@ -599,7 +612,7 @@ TEST_F(MasterQuotaTest, Status)
 TEST_F(MasterQuotaTest, InsufficientResourcesSingleAgent)
 {
   TestAllocator<> allocator;
-  EXPECT_CALL(allocator, initialize(_, _, _, _, _, _, _, _));
+  EXPECT_CALL(allocator, initialize(_, _, _));
 
   Try<Owned<cluster::Master>> master = StartMaster(&allocator);
   ASSERT_SOME(master);
@@ -638,6 +651,11 @@ TEST_F(MasterQuotaTest, InsufficientResourcesSingleAgent)
         createRequestBody(ROLE1, quotaResources));
 
     AWAIT_EXPECT_RESPONSE_STATUS_EQ(Conflict().status, response);
+    EXPECT_EQ("Quota guarantees overcommit the cluster"
+              " (use 'force' to bypass this check): "
+              "Total quota guarantees 'cpus:3; mem:2048'"
+              " exceed cluster capacity 'cpus:2; disk:1024; mem:1024'",
+              response->body);
   }
 
   // Force flag should override the `capacityHeuristic` check and make the
@@ -659,7 +677,7 @@ TEST_F(MasterQuotaTest, InsufficientResourcesSingleAgent)
 TEST_F(MasterQuotaTest, InsufficientResourcesMultipleAgents)
 {
   TestAllocator<> allocator;
-  EXPECT_CALL(allocator, initialize(_, _, _, _, _, _, _, _));
+  EXPECT_CALL(allocator, initialize(_, _, _));
 
   Try<Owned<cluster::Master>> master = StartMaster(&allocator);
   ASSERT_SOME(master);
@@ -713,6 +731,11 @@ TEST_F(MasterQuotaTest, InsufficientResourcesMultipleAgents)
         createRequestBody(ROLE1, quotaResources));
 
     AWAIT_EXPECT_RESPONSE_STATUS_EQ(Conflict().status, response);
+    EXPECT_EQ("Quota guarantees overcommit the cluster"
+              " (use 'force' to bypass this check):"
+              " Total quota guarantees 'cpus:5; mem:3072' exceed"
+              " cluster capacity 'cpus:4; disk:2048; mem:2048'",
+              response->body);
   }
 
   // Force flag should override the `capacityHeuristic` check and make the
@@ -734,7 +757,7 @@ TEST_F(MasterQuotaTest, InsufficientResourcesMultipleAgents)
 TEST_F(MasterQuotaTest, AvailableResourcesSingleAgent)
 {
   TestAllocator<> allocator;
-  EXPECT_CALL(allocator, initialize(_, _, _, _, _, _, _, _));
+  EXPECT_CALL(allocator, initialize(_, _, _));
 
   Try<Owned<cluster::Master>> master = StartMaster(&allocator);
   ASSERT_SOME(master);
@@ -779,12 +802,132 @@ TEST_F(MasterQuotaTest, AvailableResourcesSingleAgent)
 }
 
 
+// Checks that the quota capacity heuristic includes
+// reservations. This means quota is potentially not
+// satisfiable! E.g. if all resources are reserved to
+// some role without any quota guarantee, the role with
+// quota guarantee won't be satisfied.
+TEST_F(MasterQuotaTest, AvailableResourcesSingleReservedAgent)
+{
+  TestAllocator<> allocator;
+  EXPECT_CALL(allocator, initialize(_, _, _));
+
+  Try<Owned<cluster::Master>> master = StartMaster(&allocator);
+  ASSERT_SOME(master);
+
+  // Start an agent and wait until its resources are available.
+  Future<Resources> agentTotalResources;
+  EXPECT_CALL(allocator, addSlave(_, _, _, _, _, _))
+    .WillOnce(DoAll(InvokeAddSlave(&allocator),
+                    FutureArg<4>(&agentTotalResources)));
+
+  slave::Flags flags = CreateSlaveFlags();
+  flags.resources = "cpus(other-role):1;mem(other-role):512";
+
+  Owned<MasterDetector> detector = master.get()->createDetector();
+  Try<Owned<cluster::Slave>> slave = StartSlave(detector.get(), flags);
+  ASSERT_SOME(slave);
+
+  AWAIT_READY(agentTotalResources);
+
+  // We request all resources on the agent, including reservations.
+  Resources quotaResources = Resources::parse("cpus:1;mem:512").get();
+
+  // Send a quota request for the specified role.
+  Future<Quota> receivedQuotaRequest;
+  EXPECT_CALL(allocator, setQuota(Eq(ROLE1), _))
+    .WillOnce(DoAll(InvokeSetQuota(&allocator),
+                    FutureArg<1>(&receivedQuotaRequest)));
+
+  Future<Response> response = process::http::post(
+      master.get()->pid,
+      "quota",
+      createBasicAuthHeaders(DEFAULT_CREDENTIAL),
+      createRequestBody(ROLE1, quotaResources));
+
+  AWAIT_EXPECT_RESPONSE_STATUS_EQ(OK().status, response);
+
+  // Quota request is granted and reached the allocator. Make sure nothing
+  // got lost in-between.
+  AWAIT_READY(receivedQuotaRequest);
+
+  EXPECT_EQ(ROLE1, receivedQuotaRequest->info.role());
+  EXPECT_EQ(quotaResources, receivedQuotaRequest->info.guarantee());
+}
+
+
+// Checks that an operator can request quota when enough resources are
+// available on single disconnected agent.
+TEST_F(MasterQuotaTest, AvailableResourcesSingleDisconnectedAgent)
+{
+  TestAllocator<> allocator;
+  EXPECT_CALL(allocator, initialize(_, _, _));
+
+  Try<Owned<cluster::Master>> master = StartMaster(&allocator);
+  ASSERT_SOME(master);
+
+  // Start an agent and wait until its resources are available.
+  Future<Resources> agentTotalResources;
+  EXPECT_CALL(allocator, addSlave(_, _, _, _, _, _))
+    .WillOnce(DoAll(InvokeAddSlave(&allocator),
+                    FutureArg<4>(&agentTotalResources)));
+
+  Owned<MasterDetector> detector = master.get()->createDetector();
+  Try<Owned<cluster::Slave>> slave = StartSlave(detector.get());
+  ASSERT_SOME(slave);
+
+  AWAIT_READY(agentTotalResources);
+  EXPECT_EQ(defaultAgentResources, agentTotalResources.get());
+
+  // Make sure there are no agent registration retries in flight.
+  // Pausing also ensures the master does not continue to ping
+  // the agent once we spoof disconnection (which would trigger
+  // the agent to re-register).
+  Clock::pause();
+  Clock::settle();
+
+  // Spoof the agent disconnecting from the master.
+  Future<Nothing> deactivateSlave =
+    FUTURE_DISPATCH(_, &MesosAllocatorProcess::deactivateSlave);
+
+  process::inject::exited(slave.get()->pid, master.get()->pid);
+
+  AWAIT_READY(deactivateSlave);
+
+  // We request quota for a portion of resources available
+  // on the disconnected agent.
+  Resources quotaResources = Resources::parse("cpus:1;mem:512").get();
+  EXPECT_TRUE(agentTotalResources->contains(quotaResources));
+
+  // Send a quota request for the specified role.
+  Future<Quota> receivedQuotaRequest;
+  EXPECT_CALL(allocator, setQuota(Eq(ROLE1), _))
+    .WillOnce(DoAll(InvokeSetQuota(&allocator),
+                    FutureArg<1>(&receivedQuotaRequest)));
+
+  Future<Response> response = process::http::post(
+      master.get()->pid,
+      "quota",
+      createBasicAuthHeaders(DEFAULT_CREDENTIAL),
+      createRequestBody(ROLE1, quotaResources));
+
+  AWAIT_EXPECT_RESPONSE_STATUS_EQ(OK().status, response);
+
+  // Quota request is granted and reached the allocator. Make sure nothing
+  // got lost in-between.
+  AWAIT_READY(receivedQuotaRequest);
+
+  EXPECT_EQ(ROLE1, receivedQuotaRequest->info.role());
+  EXPECT_EQ(quotaResources, receivedQuotaRequest->info.guarantee());
+}
+
+
 // Checks that an operator can request quota when enough resources are
 // available in the cluster, but not on a single agent.
 TEST_F(MasterQuotaTest, AvailableResourcesMultipleAgents)
 {
   TestAllocator<> allocator;
-  EXPECT_CALL(allocator, initialize(_, _, _, _, _, _, _, _));
+  EXPECT_CALL(allocator, initialize(_, _, _));
 
   Try<Owned<cluster::Master>> master = StartMaster(&allocator);
   ASSERT_SOME(master);
@@ -853,7 +996,7 @@ TEST_F(MasterQuotaTest, AvailableResourcesMultipleAgents)
 TEST_F(MasterQuotaTest, AvailableResourcesAfterRescinding)
 {
   TestAllocator<> allocator;
-  EXPECT_CALL(allocator, initialize(_, _, _, _, _, _, _, _));
+  EXPECT_CALL(allocator, initialize(_, _, _));
 
   Try<Owned<cluster::Master>> master = StartMaster(&allocator);
   ASSERT_SOME(master);
@@ -1088,7 +1231,7 @@ TEST_F(MasterQuotaTest, RecoverQuotaEmptyCluster)
   }
 
   TestAllocator<> allocator;
-  EXPECT_CALL(allocator, initialize(_, _, _, _, _, _, _, _));
+  EXPECT_CALL(allocator, initialize(_, _, _));
 
   // Restart the master; configured quota should be recovered from the registry.
   master->reset();
@@ -1121,7 +1264,7 @@ TEST_F(MasterQuotaTest, RecoverQuotaEmptyCluster)
 TEST_F(MasterQuotaTest, NoAuthenticationNoAuthorization)
 {
   TestAllocator<> allocator;
-  EXPECT_CALL(allocator, initialize(_, _, _, _, _, _, _, _));
+  EXPECT_CALL(allocator, initialize(_, _, _));
 
   // Disable http_readwrite authentication and authorization.
   // TODO(alexr): Setting master `--acls` flag to `ACLs()` or `None()` seems
@@ -1227,7 +1370,7 @@ TEST_F(MasterQuotaTest, UnauthenticatedQuotaRequest)
 TEST_F(MasterQuotaTest, AuthorizeGetUpdateQuotaRequests)
 {
   TestAllocator<> allocator;
-  EXPECT_CALL(allocator, initialize(_, _, _, _, _, _, _, _));
+  EXPECT_CALL(allocator, initialize(_, _, _));
 
   // Setup ACLs so that only the default principal can modify quotas
   // for `ROLE1` and read status.
@@ -1775,7 +1918,7 @@ TEST_F(MasterQuotaTest, DISABLED_ChildRoleDeleteParentQuota)
 TEST_F(MasterQuotaTest, DISABLED_ClusterCapacityWithNestedRoles)
 {
   TestAllocator<> allocator;
-  EXPECT_CALL(allocator, initialize(_, _, _, _, _, _, _, _));
+  EXPECT_CALL(allocator, initialize(_, _, _));
 
   Try<Owned<cluster::Master>> master = StartMaster(&allocator);
   ASSERT_SOME(master);
