@@ -14,6 +14,7 @@
 // See the License for the specific language governing permissions and
 // limitations under the License.
 
+#include <utility>
 #include <vector>
 
 #include <gmock/gmock.h>
@@ -497,6 +498,13 @@ TEST_F(
         updateOperationStatus->mutable_operation_uuid()->CopyFrom(
             operationUuid);
 
+        ASSERT_TRUE(resourceProvider->info.has_id())
+          << "Asked to reconcile before subscription was finished";
+
+        updateOperationStatus->mutable_status()
+          ->mutable_resource_provider_id()
+          ->CopyFrom(resourceProvider->info.id());
+
         resourceProvider->send(call);
       }
     };
@@ -516,7 +524,7 @@ TEST_F(
 
   ContentType contentType = ContentType::PROTOBUF;
 
-  resourceProvider->start(endpointDetector, contentType);
+  resourceProvider->start(std::move(endpointDetector), contentType);
 
   // Wait until the agent's resources have been updated to include the
   // resource provider resources.
@@ -577,7 +585,7 @@ TEST_F(
   operationId.set_value("operation");
 
   mesos.send(v1::createCallAccept(
-      frameworkId, offer, {v1::RESERVE(reserved, operationId.value())}));
+      frameworkId, offer, {v1::RESERVE(reserved, operationId)}));
 
   AWAIT_READY(applyOperationMessage);
 
@@ -597,6 +605,7 @@ TEST_F(
 
   EXPECT_EQ(operationId, operationDroppedUpdate->status().operation_id());
   EXPECT_EQ(v1::OPERATION_DROPPED, operationDroppedUpdate->status().state());
+  EXPECT_TRUE(metricEquals("master/operations/dropped", 1));
 }
 
 // This test verifies that the master reconciles tasks that are
