@@ -35,6 +35,8 @@
 #include <stout/interval.hpp>
 #include <stout/strings.hpp>
 
+#include "common/validation.hpp"
+
 using std::max;
 using std::min;
 using std::ostream;
@@ -97,12 +99,30 @@ bool operator==(const Value::Scalar& left, const Value::Scalar& right)
 }
 
 
+bool operator<(const Value::Scalar& left, const Value::Scalar& right)
+{
+  return convertToFixed(left.value()) < convertToFixed(right.value());
+}
+
+
 bool operator<=(const Value::Scalar& left, const Value::Scalar& right)
 {
   return convertToFixed(left.value()) <= convertToFixed(right.value());
 }
 
 
+bool operator>(const Value::Scalar& left, const Value::Scalar& right)
+{
+  return convertToFixed(left.value()) > convertToFixed(right.value());
+}
+
+
+bool operator>=(const Value::Scalar& left, const Value::Scalar& right)
+{
+  return convertToFixed(left.value()) >= convertToFixed(right.value());
+}
+
+
 Value::Scalar operator+(const Value::Scalar& left, const Value::Scalar& right)
 {
   Value::Scalar result = left;
@@ -735,6 +755,13 @@ Try<Value> parse(const string& text)
     } else if (index == string::npos) {
       Try<double> value_ = numify<double>(temp);
       if (!value_.isError()) {
+        Option<Error> error =
+          mesos::internal::common::validation::validateInputScalarValue(
+              *value_);
+        if (error.isSome()) {
+          return Error("Invalid scalar value '" + temp + "':" + error->message);
+        }
+
         // This is a scalar.
         Value::Scalar* scalar = value.mutable_scalar();
         value.set_type(Value::SCALAR);
